
#ifndef ARRAYFEATURETREE_H
#define ARRAYFEATURETREE_H

#include "FeatureTree.h"
#include "FBinaryArray.h"
#include "utils.h"

class ArrayFeatureIterator : public FeatureIterator
{ 
  private:
    FBinaryArray* array;
    int index;
  public:
    ArrayFeatureIterator(FBinaryArray* a, int i) 
      : array(a), index(i) { isMapType = false; }
    Feat* operator*() { return array->index(index); }
    bool operator==(const ArrayFeatureIterator& other) const
      { return array == other.array && index == other.index; }
    bool isEqualTo(const FeatureIterator* other) const
      { return !other->isMapType && *this == *(ArrayFeatureIterator*)other; }
    Feat* operator++() { return array->index(++index); }
};

class ArrayFeatureTreeIterator : public FeatureTreeIterator
{ 
  private:
    FTreeBinaryArray* array;
    int index;
  public:
    ArrayFeatureTreeIterator(FTreeBinaryArray* a, int i)
      : array(a), index(i) { isMapType = false; }
    FeatureTree* operator*();
    bool operator==(const ArrayFeatureTreeIterator& other) const
      { return array == other.array && index == other.index; }
    bool isEqualTo(const FeatureTreeIterator* other) const
     { return !other->isMapType && *this == *(ArrayFeatureTreeIterator*)other; }
    FeatureTree* operator++();
};

class ArrayFeatureTree : public FeatureTree
{
public:
  ArrayFeatureTree() : FeatureTree() {  }
  ArrayFeatureTree(int i) : FeatureTree(i) {  }
  ArrayFeatureTree(int i, FeatureTree* b) : FeatureTree(i, b) {  }
  ArrayFeatureTree(istream& is) : FeatureTree() 
    { init(); ind_ = ROOTIND; topLevelRead(is); }

  static void init (const string& path);
  friend ostream&  operator<<(ostream& os, const ArrayFeatureTree& ft);

private:
  FBinaryArray feats;
  FTreeBinaryArray subtree;

 public:
  void init() { subtree.set(99); }
  FeatureTree* getNextFeatureTree (int index, int arrayPos, 
      						FeatureTree* parent=NULL)
  { 
    ArrayFeatureTree* retval = &subtree.array_[arrayPos];
    retval->ind_ = index;
    return retval;
  }
  Feat& getNextFeat (int index, int arrayPos)
  { 
    Feat& retval = feats.array_[arrayPos];
    retval.ind() = index;
    return retval;
  }
  void adjustSize (int numFeats, int numTrees)
  { 
    if (numFeats > 0) feats.set(numFeats);
    if (numTrees > 0) subtree.set(numTrees);
  }
  void performAuxNdVoodoo(int ac)
  { 
    /* auxNds point back not to the node the are auxes of, but to its pred */
    ArrayFeatureTree* tmp = new ArrayFeatureTree(AUXIND,this->back);
    if(ac > 0) tmp->subtree.set(ac);
    auxNd = tmp;
  }
  void setFeatureInt(int n)
  {
  }

  int getNumChildren() const { return subtree.size(); }
  int getNumFeats() const { return feats.size(); }

  Feat* getFeat(int id) { return feats[id]; }
  void incFeat(int id) { myassert (0, "should never be called"); }
  FeatIterator featsBegin()
    { return FeatIterator(new ArrayFeatureIterator(&feats, 0)); }
  FeatIterator featsEnd()
    { return FeatIterator(new ArrayFeatureIterator(&feats, feats.size())); }

  FeatureTree* getChild(int id) { return subtree[id]; }
  FeatTreeIterator childrenBegin()
    { return FeatTreeIterator(new ArrayFeatureTreeIterator(&subtree, 0)); }
  FeatTreeIterator childrenEnd()
    { return FeatTreeIterator(new ArrayFeatureTreeIterator(&subtree, subtree.size())); }

 friend class FeatTreeIter;
 friend class FeatIter;

 private:
  static FeatureTree* roots_[NUMCALCS];
  void read(istream& is, FTypeTree* ftt);
  void othReadFeatureTree(istream& is, FTypeTree* ftt, int cnt);
};

inline FeatureTree* 
ArrayFeatureTreeIterator::operator*() { return array->index(index); }
inline FeatureTree* 
ArrayFeatureTreeIterator::operator++() { return array->index(++index); }

#endif /* ! ARRAYFEATURETREE_H */
