
#ifndef BCHART_H
#define BCHART_H

#include <string>
#include "ChartBase.h"
#include "EdgeHeap.h"
#include "Wrd.h"
#include "FullHist.h"

using namespace std;

#define Termstar const Term*

#define UBPRETERMS 80
#define UBNONTERMS 80
// MAXSENTLEN defined in ChartBase

struct Wwegt  
{
  int t;
  char e[2];
  float p;
};

class Bchart;
extern int globalDebugLevel;

class           Bchart : public ChartBase
{
public:
    Bchart(SentRep& sentence);
    virtual ~Bchart();
    virtual double  parse();
    bool    edgeFromIEdge(istream& str);
    static int&      printDebug() { return globalDebugLevel; }
    static bool      printDebug(int val) { return val < printDebug(); }
    static void     readTermProbs(string& path);
    static int      wtoInt(string& str);
    int     extraTime; //if no parse is found on regular time;
    static  Item*    dummyItem;
    static float    denomProbs[MAXSENTLEN];  //Restricts to sentences of len < 1000;
    void            check();
    static void     setPosStarts();
protected:
    void            add_reg_item(Item * itm);
    void            addFinishedEdge(Edge* newEdge);
    void            add_starter_edges(Item* itm);
    float           meEdgeProb(const Term* trm, Edge* edge, int whichInt);
    float           meFHProb(const Term* trm, FullHist& fh, int whichInt);
    //static int printDebug_;

    void            extend_rule(Edge* rule, Item * itm, int right);
    void            already_there_extention(int i, int start, int right,
					    Edge* edge);
    void            add_edge(Edge* rli, int left);
    void            put_in_reg(Item * item);
    void            addWordsToKeylist( );
    Item           *in_chart(const Wrd* hd, const Term * trm,
			     int start, int finish);
//    int             fastCompatCheck(int par,int chi,int right)
//      { return right ? compat_right_[par][chi] : compat_left_[par][chi]; }

    void            redoP(Edge* edge, double probRatio);
    void            redoP(Item *item, double probDiff);

    float           computeMerit(Edge* edge, int whichCalc);

    void  initDenom();
    list<float>& wordPlist(Wrd* word, int word_num);
    double  psktt(Wrd* shU, int t);
    double  pCapgt(const Wrd* shU, int t);
    float   pHst(int w, int t);
    double  psutt(const Wrd* shU, int t);
    float   pegt(string& sh, int t);
    void    getpHst(const string& hd, int t);
    double pHypgt(const string& shU, int t);
    static float&  pHcapgt(int i) { return pHcapgt_[i]; }
    static float&  pHhypgt(int i) { return pHhypgt_[i]; }
    static float&  pHugt(int i) { return pHugt_[i]; }
    static float& pT(int tInt) { return pT_[tInt-Term::lastTagInt()-1];}
    int     bucket(float val);
    int    greaterThan(Wwegt& wwegt, char e[2], int t);
    float  pHegt(string& es, int t);
    float  computepTgT(int t1,int t2);
    void   addToDemerits(Edge* edge);

    Wrd*   add_word(int i, string wrdStr);
    void   readItem(istream& str, int& b, int& e, string& trm);
    Item*  add_item(int b, int e, string trmNm, Wrd* wrd);
    Edge*  add_edge(Item* lhs, Items& rhs);
    int    headPosFromItems(Item* lhs, Items& rhs);

  static Item*    stops[MAXSENTLEN];
    EdgeHeap*       heap;
    int             alreadyPopedNum;
    Edge*           alreadyPoped[250000];
    static int&     posStarts(int i, int j);
    static int      posStarts_[UBPRETERMS][30];
//    static int     compat_left_[30][80];
//    static int     compat_right_[30][80];
    static int     curDemerits_[MAXSENTLEN][MAXSENTLEN];

  static int egtSize_;
  static float bucketLims[14];
  static float pT_[UBNONTERMS];
  static float pHcapgt_[UBPRETERMS];
  static float pHhypgt_[UBPRETERMS];
  static float pHugt_[UBPRETERMS];
  //static float unigrams_[80];
  //static float bigrams_[50][50];

  static Wwegt* pHegt_;
  list<float> wordPlists[MAXSENTLEN];
  static map< string, int, less<string> > wordMap;
};

#endif	/* ! BCHART_H */

