// CLRInfo.h

#ifndef CLRINFO_H
#define CLRINFO_H

#include <iostream>
#include <string>
#include <vector>
#include <map>
#include <set>
#include "WordInfo.h"
#include "Term.h"
#include "Pst.h"
#include "utils.h"
class Hist;

template<class A, class B, class C>
struct triple
{ 
  triple() {  }
  triple(A a, B b, C c) : first(a), second(b), third(c) {  }
  A first;
  B second;
  C third;
  bool operator< (const triple& other) const
  {
    if (first != other.first)
      return first < other.first;
    else if (second != other.second)
      return second < other.second;
    else if (third != other.third)
      return third < other.third;
    else return false;
  }
  bool operator== (const triple& other) const
  { return 
         first == other.first
      && second == other.second
      && third == other.third;
  }
};
template <class A, class B, class C>
ostream& operator<< (ostream& os, const triple<A,B,C>& tr)
{ 
  os << "<" << tr.first << ", " << tr.second << ", " << tr.third << ">";
}

class InputTree;
class SentRep;
struct CLRPattern;
//typedef const WordInfo* wordtype;
typedef string wordtype;
typedef int termtype;
//typedef const Term* termtype;
//typedef string termtype;
//typedef pair<wordtype, bool> Wbpair;
//typedef map<termtype, Wbpair> TWbmap;
typedef triple<termtype, wordtype, bool> twbtriple;
typedef triple<wordtype, termtype, wordtype> vfhtriple;

#define string_to_wordtype(X) (X)
//#define string_to_wordtype(X) (Pst::get(toLower(X)))
//#define string_to_termtype(X) (X)
//#define string_to_termtype(X) (Term::get(X))
#define string_to_termtype(X) (Term::get(X)->toInt())
#define int_to_termtype(X) (X)

class CLRInfo
{ 
    static map<wordtype, map<wordtype, vector<vector<wordtype> > > > contextMap;

    static wordtype stringFilter (string str) 
      { return lemmatise(str); }

  public:
    static void addCLRPattern(InputTree*);
    static bool isCLRContext(const SentRep* sr, const Hist*);
    static void init(string data);
    static void init(istream& os);
    static void dump(ostream& os);
    static int getSize() { return contextMap.size(); }
};


#endif
