#include "Cluster.h"
#include <sys/types.h>
#include <dirent.h>
#include <set>
#include <vector>
#include <map>
#include <algorithm>

std::vector<std::map<int, int, std::less<int> > > Cluster::wordVec_;


void
Cluster::
init(std::string& prefix)
{
  std::string	dirName(prefix);
  dirName += "clusters/";
  DIR* clDir = opendir(dirName.c_str());
  struct dirent* dirEntry;
  set<std::string, less<std::string> > termSet;

  if (!clDir)
  {
    cerr << "Can't open clusters directory " << dirName << endl;
    return;
  }
  while (dirEntry = readdir(clDir)) 
  {
    char* suffix = strstr(dirEntry->d_name, ".txt");
    if (!suffix) continue;
    std::string termName(dirEntry->d_name, suffix - dirEntry->d_name);
    termSet.insert(termName);
  }
  closedir(clDir);

  for (set<std::string, less<std::string> >::iterator it = termSet.begin();
       it != termSet.end();
       ++it)
  {
    const std::string& termStr = *it;
    const Term* term = Term::get(termStr);
    if (!term) continue;
    std::string fileName(dirName);
    fileName += termStr + ".txt";
    ifstream  stream(fileName.c_str(), ios::in);
    if (!stream)
    {
      cerr << "Can't open clusters file " << fileName << endl;
      continue;
    }

    int wordInt, cluster;
    while (stream >> wordInt)
    {
      stream >> cluster;
      if (wordInt >= wordVec_.size())
      {
	//cout << "Increasing wordvec to size " << wordInt << endl;
	//wordVec_.resize(wordInt + 1);
	// stupid gay stl doesn't have resize
	wordVec_.reserve(wordInt + 1);
	wordVec_.insert(wordVec_.end(), wordInt - wordVec_.size() + 1, map<int, int, less<int> >());
	assert(wordVec_.size() == wordInt + 1);
      }

      wordVec_[wordInt][term->toInt()] = cluster;
    }
  }
}
