
#include <vector>
#include <fstream>
#include <iostream>
#include <string>
#include "Feature.h"
#include "utils.h"
#include "expat.h"

using namespace std;

SubFeature* SubFeature::array_[NUMCALCS][MAXNUMFS];
int SubFeature::total[NUMCALCS];
int (*SubFeature::Funs[MAXNUMFS])(Hist*);
int      SubFeature::ufArray[NUMCALCS][MAXNUMFS];
int      SubFeature::splitPts[NUMCALCS][MAXNUMFS];
Feature* Feature::array_[NUMCALCS][MAXNUMFS];
float*     Feature::lambdas_[NUMCALCS][MAXNUMFS];
int Feature::total[NUMCALCS];
int Feature::conditionedFeatureInt[NUMCALCS];
int  Feature::whichInt;
int Feature::assumedFeatVal;
int (*Feature::conditionedEvent)(Hist*);
int (*Feature::assumedSubFeat)(Hist*);
FTypeTree Feature::ftTree[NUMCALCS];
FTypeTree* Feature::ftTreeFromInt[NUMCALCS][MAXNUMFS];
int globalDebugLevel = 0;

inline static int debugLevel() 
{
  return globalDebugLevel;
}

#define LAMBDABUCKETS 15

void
Feature::
assignCalc(string conditioned)
{
  if(conditioned == "h") whichInt = HCALC;
  else if(conditioned == "u") whichInt = UCALC;
  else if(conditioned == "r") whichInt = RCALC;
  else if(conditioned == "ru") whichInt = RUCALC;
  else if(conditioned == "rm") whichInt = RMCALC;
  else if(conditioned == "tt") whichInt = TTCALC;
  else if(conditioned == "l") whichInt = LCALC;
  else if(conditioned == "lm") whichInt = LMCALC;
  else if(conditioned == "f") whichInt = FCALC;
  else if(conditioned == "gr") whichInt = GRCALC;
  else if(conditioned == "grH") whichInt = GRHCALC;
  else if(conditioned == "ff") whichInt = FFCALC;
  else if(conditioned == "ffH") whichInt = FFHCALC;
  else if(conditioned == "tp") whichInt = TPCALC;
  else if(conditioned == "tpH") whichInt = TPHCALC;
  else if(conditioned == "mf") whichInt = MFCALC;
  else if(conditioned == "mfH") whichInt = MFHCALC;
  else if(conditioned == "HC") whichInt = HCCALC;
  else
    {
      assert(conditioned == "m");
      whichInt = MCALC;
    }
}


void
Feature::
init(const string& path)
{
}

void
Feature::
init(const string& path, const string& conditioned)
{
  assignCalc(conditioned);
  int f;
  for(f = 0 ; f < MAXNUMFS ; f++)
    {
      float* vec = new float[LAMBDABUCKETS];
      lambdas_[whichInt][f] = vec;
      for(int k = 0 ; k < LAMBDABUCKETS ; k++) vec[k] = 0.0;
    }

  Feature::ftTreeFromInt[whichInt][0] = &(Feature::ftTree[whichInt]);

  string dataString(path);
  dataString += "featureTree.";
  dataString += conditioned;
  ifstream dataStream(dataString.c_str());
  if (dataStream)
  {
    if (debugLevel() > 40)
      cerr << "Feature::init: reading XML feature tree for "
	<< conditioned << endl;
    readTreeXML(dataStream);
  }
  else
  {
    dataString = path;
    dataString += "featInfo.";
    dataString += conditioned;
    ifstream dataStream(dataString.c_str());
    myassert(dataStream, dataString);

    if (debugLevel() > 40)
      cerr << "Feature::init: reading old-style feature tree for "
	<< conditioned << endl;
    readTree(dataStream);
  }

  int num;
  /* set the universal function num on feats from their subfeats */
  for(num = 0 ; num < Feature::total[whichInt] ; num++)
    {
      Feature* f = array_[whichInt][num];
      f->usubFeat = SubFeature::fromInt(f->subFeat,whichInt)->usf;
    }
  /* set up the table from universal subfeat nums to subfeat nums */
  for(num = 0 ; num < MAXNUMFS ; num++)
    SubFeature::ufArray[whichInt][num] = -1;
  for(num = 0 ; num < SubFeature::total[whichInt] ; num++)
    {
      SubFeature* sf = SubFeature::fromInt(num,whichInt);
      SubFeature::ufArray[whichInt][sf->usf] = num;
    }
}

void Feature::
readTree(ifstream& dataStrm)
{
  int auxCnts[MAXNUMFS];
  for(int i = 0 ; i < MAXNUMFS ; i++) auxCnts[i] = 0;

  int conditionedInt;
  dataStrm >> conditionedInt;
  if (debugLevel() > 15)
    cerr << "setting condint: " << whichInt << ", " << conditionedInt << endl;
  conditionedFeatureInt[whichInt] = conditionedInt;

  int num;
  for(num = 0 ;  ; num++)
    {
      int n, subf, pos, cpr;
      string nm;
      string tmp;
      dataStrm >> tmp;
      if(tmp == "--") break;
      n = atoi(tmp.c_str());
      dataStrm >> nm;
      dataStrm >> subf;
      dataStrm >> pos;
      dataStrm >> tmp;
      
      if(tmp == "|")
	cpr = -1;
      else
	{
	  cpr = atoi(tmp.c_str());
	  dataStrm >> tmp;
	  assert(tmp == "|");
	}
      if (debugLevel() > 15)
        cerr << "newing Feature: " << whichInt << ", " << n << ", "
	  	<< nm << ", " << subf << ", " << pos << ", " << cpr << endl;
      array_[whichInt][n-1] = new Feature(n, nm, subf, pos);
      array_[whichInt][n-1]->auxCnt = auxCnts[pos];
      auxCnts[pos]++;

      createFTypeTree(Feature::ftTreeFromInt[whichInt][pos], n, whichInt);
    }
  if (debugLevel() > 15)
    cerr << "setting f total: " << whichInt << ", " << num << endl;
  Feature::total[whichInt] = num;
  for(num = 0 ;  ; num++)
    {
      int n, fn;
      string nm;
      string tmp;
      dataStrm >> tmp;
      if(tmp == "--") break;
      n = atoi(tmp.c_str());
      dataStrm >> nm;
      dataStrm >> fn;
      for( ; ; )
	{
	  dataStrm >> tmp;
	  if(tmp == "|") break;
	}
      if (debugLevel() > 15)
        cerr << "newing SubFeature: " << n << ", " << whichInt << ", "
	  	<< nm << ", " << fn << endl;
      SubFeature::fromInt(n, whichInt)
	= new SubFeature(n, nm, fn);
      assert(SubFeature::fromInt(n, whichInt));
    }
  if (debugLevel() > 15)
    cerr << "setting sf total: " << whichInt << ", " << num << endl;
  SubFeature::total[whichInt] = num;
}

struct TreeData
{
  int n;
  vector<Feature*> path;
  int auxCnts[MAXNUMFS];
};

/** This procedure gets called every time a start tag is read.
  * @param userData a structure (in this case, TreeData) that contains
  *	any information I want to persist between invocations 
  * @param nm the name of the tag ('feature', 'target')
  * @param atts an array of char*, containing attribute-value pairs,
  *	terminated by a null pointer */
void Feature::
readStartXMLElement(void *userData, const char *nm, const char **atts)
{
  vector<Feature*>& pathToHere = ((TreeData*)userData)->path;
  int& n = ((TreeData*)userData)->n;
  int* auxCnts = ((TreeData*)userData)->auxCnts;
  string name(nm);
  int id = -1;
  string featureName;

  // First, sort out the attribute-value pairs that we're interested in
  for (int i = 0; atts[i] != 0; i+=2)
  {
    string key(atts[i]);
    if (key == "id")
      id = atoi(atts[i+1]);
    else if (key == "name")
      featureName = atts[i+1];
  }

  if (name == "target") 
  {
    /* The target tag should be the first thing in the file, and 
       identifies the index of the conditioned feature (as well as its
       name, which we ignore). */
    assert (id >= 0);
    if (debugLevel() > 15)
      cerr << "setting condint: " << whichInt << ", " << id << endl;
    conditionedFeatureInt[whichInt] = id;
  }
  else if (name == "feature")
  {
    /* The feature tag is everything else; anything that depends on
       a given feature is contained inside that feature tag.  To keep
       track of the tree, we have a vector of features representing the
       path from the root of the tree to where we are now. */
    assert (id >= 0);
    int index = n++;
    int parentNum = pathToHere.size()? pathToHere.back()->toInt(): 0;
    if (debugLevel() > 15)
      cerr << "newing Feature: " << whichInt << ", " << n << ", "
	      << featureName << ", " << index << ", " << parentNum
	      <<endl;
    Feature* newFeature = new Feature(n, featureName, index, parentNum);
    array_[whichInt][index] = newFeature;
    newFeature->auxCnt = auxCnts[parentNum];
    auxCnts[parentNum]++;

    createFTypeTree(Feature::ftTreeFromInt[whichInt][parentNum], n, whichInt);

    if (debugLevel() > 15)
      cerr << "newing SubFeature: " << index << ", " << whichInt << ", "
	      << featureName << ", " << id << endl;
    SubFeature::fromInt(index, whichInt)
	= new SubFeature(index, featureName, id);
    assert(SubFeature::fromInt(index, whichInt));

    if (debugLevel() > 15)
      cerr << "Adding feature " + featureName + " under " << parentNum << endl;
    pathToHere.push_back(newFeature);
  }
}

void Feature::
readEndXMLElement(void *userData, const char* nm)
{
  vector<Feature*>& pathToHere = ((TreeData*)userData)->path;
  string name(nm);

  if (name == "feature") 
  {
    pathToHere.pop_back();
    if (debugLevel() > 15)
      cerr << "Ending feature\n";
  }
}

void Feature::
readTreeXML (ifstream& dataStream)
{
  XML_Parser parser = XML_ParserCreate(NULL);
  TreeData userdata;
  userdata.n = 0;
  for(int i = 0 ; i < MAXNUMFS ; i++) userdata.auxCnts[i] = 0;
  XML_SetUserData(parser, &userdata);
  XML_SetElementHandler(parser, readStartXMLElement, readEndXMLElement);
  string buffer;
  while (dataStream >> buffer) {
    buffer += ' ';
    if (!XML_Parse(parser, buffer.c_str(), buffer.length(), 0)) 
      error (string("Invalid XML in feature tree file ") + whichInt);
  }
  if (debugLevel() > 15)
    cerr << "setting total: " << whichInt << ", " << userdata.n << endl;
  Feature::total[whichInt] = userdata.n;
  SubFeature::total[whichInt] = userdata.n;
  XML_Parse(parser, "", 0, 1);
  XML_ParserFree(parser);
}

void
Feature::
readLam(int which, string tmp, string path)
{
  string ftstr(path);
  ftstr += tmp;
  ftstr += ".lambdas";
  ifstream fts(ftstr.c_str());
  assert(fts);
  int b,f;
  for(b = 1; b < LAMBDABUCKETS ; b++)
     {
       int bb;
       assert(fts);
       fts >> bb ;
       //cerr << bb << endl;
       myassert(bb == b, ftstr);
       for(f = 2; f <= Feature::total[which] ; f++)
	 {
	   float lam;
	   assert(fts);
	   fts >> lam;
	   //cerr << which << " " << f << " " << b << " " << lam << endl;
	   Feature::setLambda(which,f,b,lam);
	 }
     }
} 

void
Feature::
createFTypeTree(FTypeTree* posftTree, int n, int which)
{
  assert(posftTree);
  if(!posftTree->left)
    {
      posftTree->left = new FTypeTree(n);
      Feature::ftTreeFromInt[which][n] = posftTree->left;
    }
  else if(!posftTree->right)
    {
      posftTree->right = new FTypeTree(AUXIND);
      createFTypeTree(posftTree->right, n, which);
    }
  else createFTypeTree(posftTree->right, n, which);
}     
