#ifndef FEATUREFUNCTION_H
#define FEATUREFUNCTION_H

#include <string>
#include <list>
#include <fstream>
#include "utils.h"

#define MAXNUMFS (46 /*Used also in treeHistSf.C and fhSubFns.C; in sel too*/)

enum {
  RCALC=0,
  HCALC,
  UCALC,
  MCALC,
  LCALC,
  LMCALC,
  RUCALC,
  RMCALC,
  TTCALC,
  FCALC,
  GRCALC, //10
  FFCALC,
  TPCALC,
  MFCALC,
  HCCALC,
  GRHCALC,
  FFHCALC,
  TPHCALC,
  MFHCALC,
  NUMCALCS
};

#define AUXIND -9
#define ROOTIND -99

class FTypeTree;

class FTypeTree
{
 public:
  FTypeTree() : back(NULL),left(NULL), right(NULL), n(-1) {}
  FTypeTree(int fi) : back(NULL), left(NULL), right(NULL), n(fi) {}
  int n;
  FTypeTree* left;
  FTypeTree* right;
  FTypeTree* back;
};

class Hist;

  /*  Currently what goes in Funs.
    0 t  tree_term 0 |
    1 l  tree_parent_term
    2 u  tree_pos
    3 h  tree_head
    4 ph tree_parent_head
    */

/*
  num  name  function
  0    t      0
  1    l      1
  2    u      2
  */
class SubFeature
{
 public:
  SubFeature(int i, std::string& nm, int fnn)
    : num(i), name(nm), usf(fnn), fun(Funs[fnn])
    {}
  static SubFeature*& fromInt(int i, int which) { return array_[which][i]; }
  int num;
  std::string name;
  int (*fun)(Hist*);
  int usf;
  static int total[NUMCALCS];
  static int (*Funs[MAXNUMFS])(Hist*);
  static int      ufArray[NUMCALCS][MAXNUMFS];
  static int      splitPts[NUMCALCS][MAXNUMFS];
 private:
  static SubFeature* array_[NUMCALCS][MAXNUMFS];
};

/*
  num name ff startpos
  1   rt   0  0
  2   rtl  1  1
  3   rtu  2  1
  */
class Feature
{
 public:
  Feature(int i, std::string& nm, int ff, int pos)
    : num(i), name(nm), subFeat(ff), startPos(pos), auxCnt(0),
      condPR(-1) {}
  static Feature*& fromInt(int i, int which)
    {
      myassert(i > 0, i);
      return array_[which][i-1];
    }
  int toInt() const { return num; }
  int num;
  std::string name;
  int subFeat;
  int usubFeat;
  int startPos;
  int condPR;
  int auxCnt;
  static void assignCalc(std::string conditioned);
  static int total[NUMCALCS];
  static int conditionedFeatureInt[NUMCALCS];
  static void init(const std::string& path);
  static void init(const std::string& path, const std::string& conditioned);
  static void readLam(int which, std::string tmp, std::string path);
  static int whichInt;
  static int assumedFeatVal;
  static int (*conditionedEvent)(Hist*);
  static int (*assumedSubFeat)(Hist*);
  //e.g., when processing rules for NP, it would be 55;
  static FTypeTree ftTree[NUMCALCS];
  static FTypeTree* ftTreeFromInt[NUMCALCS][MAXNUMFS];
  static void  createFTypeTree(FTypeTree* ft, int n, int which);
  static float getLambda(int wi, int featInt, int bucketInt)
    {return lambdas_[wi][featInt-1][bucketInt];}
  static void setLambda(int wi, int featInt, int bucketInt, float val)
    { lambdas_[wi][featInt-1][bucketInt] = val;}
 private:
  static void readTree(std::ifstream& dataStream);
  static void readTreeXML(std::ifstream& dataStream);
  static void
    readStartXMLElement(void *userData, const char *nm, const char **atts);
  static void readEndXMLElement(void *userData, const char* nm);
 private:
  static Feature* array_[NUMCALCS][MAXNUMFS];
  static float* lambdas_[NUMCALCS][MAXNUMFS];
};

#endif				/* ! FEATUREFUNCTION_H */
