#include "Term.h"
#include "FuncInfo.h"
#include "Feature.h"
#include "Hist.h"
//#include "Edge.h"
#include "headFinder.h"
//#include "Bchart.h"
#include "ccInd.h"
#include "Cluster.h"
#include "Pst.h"
#include "CLRInfo.h"
#include "SentRep.h"
#include "Rule.h"

#define MAXFEATS (MAXNUMFS /*defined in Feature.h*/)

typedef set<int> intset;
typedef set<int>::iterator intsetiter;

static bool set_has_intersection(intsetiter first1, intsetiter last1,
			intsetiter first2, intsetiter last2);

SentRep* CurrentSentence;


int
node_label(Hist* hist)
{
  return hist->getLabel();
}

int
node_parent_label(Hist* hist)
{
  static int S1 = 0;
  if(!S1) { S1 = Term::get(string("S1"))->toInt(); }

  const Hist* parent = hist->getParent();
  if(!parent) return S1;
  else return  parent->getLabel();
}

int
node_parent_head_POS(Hist* hist)
{
  static int STOP = 0;
  if(!STOP) STOP = Term::get(string("STOP"))->toInt();

  const Hist* parent = hist->getParent();
  if(!parent) return STOP;
  return parent->getHeadPOS();
}

int
node_left_sibling_label(Hist* hist)
{
  static int STOP = 0;
  if(!STOP) STOP = Term::get(string("STOP"))->toInt();

  const Hist* parent = hist->getParent();
  if(!parent) return STOP;
  list<Hist*>::const_iterator it = parent->getChildren().begin();
  for( ; it != parent->getChildren().end() ; it++ )
    {
      const Hist* sibling = *it;
      if(sibling != hist) continue;
      if(it == parent->getChildren().begin()) return STOP;
      --it;
      sibling = *it;
      assert(sibling);
      return sibling->getLabel();
    }
  error("node_left_sibling_label: parent had no children");
  return -1;
}

int
node_right_sibling_label(Hist* hist)
{
  static int STOP = 0;
  if(!STOP) STOP = Term::get(string("STOP"))->toInt();

  const Hist* parent = hist->getParent();
  if(!parent) return STOP;
  list<Hist*>::const_iterator it = parent->getChildren().begin();
  for( ; it != parent->getChildren().end() ; it++ )
    {
      const Hist* sibling = *it;
      if(sibling != hist) continue;
      it++;
      if(it == parent->getChildren().end()) return STOP;
      sibling = *it;
      assert(sibling);
      return sibling->getLabel();
    }
  error("node_right_sibling_label: parent had no children");
  return -1;
}

int
node_head_POS(Hist* hist)
{
  int retval = hist->getHeadPOS();
  static int STOP = -1;
  if (STOP < 0) STOP = Term::get(string("STOP"))->toInt();
  assert(retval != STOP);
  assert(retval != 47);
  return retval;
}

int
node_head(Hist* hist)
{
  int ans = hist->getHead();
  assert(ans >= -1);
  return ans;
}

int
node_parent_head(Hist* hist)
{
  static int dummyWordInt = -1;
  if(dummyWordInt < 0) dummyWordInt = Pst::get(string("^^"))->toInt();
  assert (dummyWordInt >= 0);
  
  const Hist* parent = hist->getParent();
  if(!parent)
  {
    //assert(!((TreeHist*)hist)->tree->parent());
    return dummyWordInt;
  }
  int ans = parent->getHead();
  assert(ans >= -1);
  return ans;
}

int
node_grandparent_head(Hist* hist)
{
  static int dummyWordInt = -1;
  if(dummyWordInt < 0) dummyWordInt = Pst::get(string("^^"))->toInt();

  const Hist* parent = hist->getParent();
  if(!parent) return dummyWordInt;
  const Hist* grandparent = parent->getParent();
  if(!grandparent) return dummyWordInt;
  
  int ans = grandparent->getHead();
  assert(ans >= -1);
  return ans;
}

int
node_grandparent_label(Hist* hist)
{
  static int S1 = 0;
  if(!S1) S1 = Term::get(string("S1"))->toInt();

  const Hist* parent = hist->getParent();
  if (!parent) return S1;
  const Hist* grandparent = parent->getParent();
  if(!grandparent) return S1;
  else return grandparent->getLabel();
}

int
node_grandparent_head_POS(Hist* hist)
{
  static int STOP = 0;
  if(!STOP) STOP = Term::get(string("STOP"))->toInt();

  const Hist* parent = hist->getParent();
  if(!parent) return STOP;
  const Hist* grandparent = parent->getParent();
  if(!grandparent) return STOP;
  return grandparent->getHeadPOS();
}

int
node_parent_cc_label(Hist* hist)
{
  static int S1 = 0;
  if(!S1) S1 = Term::get(string("S1"))->toInt();

  const Hist* parent = hist->getParent();
  if(!parent) return S1;
  int parentLabel = parent->getLabel();
  //assert (parentLabel != 47); //STOP
  if(parentLabel != hist->getLabel()) return parentLabel;
  int cc_parentLabel = ltocc(parentLabel, parent->isConjunctivePhrase());
  //assert (cc_parentLabel != 47); //STOP
  return cc_parentLabel;
}

int
node_true(Hist* h) {return 1;}


/** Returns the nth node from the current one in either direction.
  * @param hist the current node
  * @param howFar the distance from the current node
  * @param direction which way (+1 or -1)
  * @return the label of the node at that spot */
int
node_ngram(Hist* hist, int howFar, int direction)
{
  static int STOP = -1;
  if(STOP < 0) STOP = Term::get(string("STOP"))->toInt();

  int position = hist->getPosition();
  int headPosition = hist->getHeadPosition(); //???;
  int targetPosition = position + (howFar * direction);

  if(targetPosition < 0) return STOP;
  if(targetPosition > headPosition && direction > 0) return STOP;
  if(targetPosition >= hist->getNumChildren()) return STOP;

  return hist->getChildLabel(targetPosition);
}

int
node_left0(Hist* hist)
{
  return node_ngram(hist, 0, 0);
}

int
node_left1(Hist* hist)
{
  return node_ngram(hist, 1, 1);
}

int
node_left2(Hist* hist)
{
  return node_ngram(hist, 2, 1);
}

int
node_left3(Hist* hist)
{
  return node_ngram(hist, 3, 1);
}

int
node_right1(Hist* hist)
{
  return node_ngram(hist, 1, -1);
}

int
node_right2(Hist* hist)
{
  return node_ngram(hist, 2, -1);
}

int
node_right3(Hist* hist)
{
  return node_ngram(hist, 3, -1);
}

int
node_noopenQr(Hist* hist)
{
  static int LDQ = -1;
  if (LDQ < 0) LDQ = Term::get(string("``"))->toInt();
  static int RDQ = -1;
  if (RDQ < 0) RDQ = Term::get(string("''"))->toInt();

  int position = hist->getPosition();
  bool sawOpen = false;
  int i;
  list<Hist*>::const_iterator it;

  for (i = 0; i < hist->getNumChildren(); ++i)
  {
    if (i == position) break;
    // Note: this is O(n^2) for TreeHists: FIX THIS.
    if (hist->getChildLabel(i) == LDQ) sawOpen = true;
    else if (hist->getChildLabel(i) == RDQ) sawOpen = false;
  }
 
  if(sawOpen) return 0;
  else return 1;
}

int
node_noopenQl(Hist* hist)
{
  static int LDQ = -1;
  if (LDQ < 0) LDQ = Term::get(string("``"))->toInt();
  static int RDQ = -1;
  if (RDQ < 0) RDQ = Term::get(string("''"))->toInt();

  int position = hist->getPosition();
  int headPosition = hist->getHeadPosition();
  bool sawOpen = false;
  int i;
  list<Hist*>::const_reverse_iterator it;

  for (i = hist->getNumChildren() - 1; i > position; --i)
  {
    if (i > headPosition) continue;
    assert (i >= 0);
    // Note: this is O(n^2) for TreeHists: FIX THIS.
    if (hist->getChildLabel(i) == RDQ) sawOpen = true;
    else if (hist->getChildLabel(i) == LDQ) sawOpen = false;
  }

  if(sawOpen) return 0;
  else return 1;
}

InputTree* sentence[256];
int endPos;

static int
node_B(Hist* hist, int blInd)
{
  int pos = hist->getPosition();
  int hpos = hist->getHeadPosition();
  //cerr << "tb1 " << pos << " " << hpos << " " << *tree << endl;
  int sz = hist->getNumChildren();
  int wpos;
  assert(pos <= sz);
  //cerr << "tb " << pos << " " << hpos << " " << sz << endl;
  if(pos < 0) wpos = hist->getStart()-1;
  else if(sz == 0) wpos = hist->getStart()-1;
  else if(pos == sz) wpos = hist->getFinish();
  else
    {
      int i = 0;
      for (list<Hist*>::const_iterator it = hist->getChildren().begin();
	   it != hist->getChildren().end();
	   ++it)
	{
	  if(i < pos) {++i; continue;}
	  Hist* child = *it;
	  if(pos < hpos) wpos = child->getStart()-1;
	  else if(pos > hpos) wpos = child->getFinish();
	  else if(blInd) wpos = child->getStart()-1;
	  else wpos = child->getFinish();
	  break;
	}
    }
  //cerr << "tb2 " << wpos << endl;
  assert(wpos <= endPos);
  if(wpos < 0 || wpos == endPos) return Term::stopTerm->toInt();
  else return Term::get(sentence[wpos]->term())->toInt();
}

int
node_Bl(Hist* hist)
{
  // shouldn't be called on fullhists, only treehists
  return node_B(hist, 1);
}

int
node_Br(Hist* hist)
{
  // shouldn't be called on fullhists, only treehists
  return node_B(hist, 0);
}

static set<int, less<int> >
buildAdjuncts()
{
  set<int, less<int> > retval;
  retval.insert(FuncInfo::get("ADV")->toInt());
  retval.insert(FuncInfo::get("VOC")->toInt());
  retval.insert(FuncInfo::get("BNF")->toInt());
  retval.insert(FuncInfo::get("DIR")->toInt());
  retval.insert(FuncInfo::get("EXT")->toInt());
  retval.insert(FuncInfo::get("LOC")->toInt());
  retval.insert(FuncInfo::get("MNR")->toInt());
  retval.insert(FuncInfo::get("TMP")->toInt());
  retval.insert(FuncInfo::get("CLR")->toInt());
  retval.insert(FuncInfo::get("PRP")->toInt());
  return retval;
}

int
node_complement(Hist* hist)
{
  /* Uses the heuristics from Collins97: Complements are either
     1. non terminal that is either
     	1. NP, SBAR[Q], or S, whose parent is S (or SINV or SQ)
	2. NP, SBAR[Q], S, or VP, whose parent is VP
	3. S whose parent is SBAR[Q]
	and
     2. non terminal cannot have func tags ADV, VOC, BNF, DIR,
     	EXT, LOC, MNR, TMP, CLR, or PRP.
     OR
     the first child after the head of a prepositional phrase (PP) */
  static int S = Term::get("S")->toInt();
  static int SQ = Term::get("SQ")->toInt();
  static int SINV = Term::get("SINV")->toInt();
  static int SBAR = Term::get("SBAR")->toInt();
  static int SBARQ = Term::get("SBARQ")->toInt();
  static int NP = Term::get("NP")->toInt();
  static int VP = Term::get("VP")->toInt();
  static int PP = Term::get("PP")->toInt();
  static int STOP = Term::get("STOP")->toInt();

  const Hist* parent = hist->getParent();
  if(!parent) return false;
  int parentLabel = parent->getLabel();

  if (parentLabel == PP)
    return (node_left_sibling_label(hist) != STOP);

  /* If any of the adjunct function tags are on the current tree node,
   * then we must return false.  We perform an intersection between the
   * adjunct function tag set and the set of function tags on this node;
   * if it is not empty, return false. */
  static set<int> adjunctFuncTags = buildAdjuncts();
  set<int> treeFuncTags;
  for (int i = 0; i < 4; ++i)
  {
    int fnum = ((TreeHist*)hist)->tree->getFuncTag(i);
    if (fnum != NOFUNC_NUM)
      treeFuncTags.insert (fnum);
  }
  if (set_has_intersection (adjunctFuncTags.begin(), adjunctFuncTags.end(),
      		    	  treeFuncTags.begin(), treeFuncTags.end()))
    return false;

  int thisLabel = hist->getLabel();

  if (parentLabel == S || parentLabel == SQ || parentLabel == SINV)
      return (thisLabel == NP || thisLabel == SBAR || thisLabel == SBARQ
	   || thisLabel == S || thisLabel == SQ || thisLabel == SINV);

  else if (parentLabel == VP)
      return (thisLabel == NP || thisLabel == SBAR || thisLabel == SBARQ
	   || thisLabel == S || thisLabel == SQ || thisLabel == SINV
	   || thisLabel == VP);

  else if (parentLabel == SBAR || parentLabel == SBARQ)
      return (thisLabel == S || thisLabel == SQ || thisLabel == SINV);

  else
      return false;
}


int
node_grammatical_function(Hist* hist)
{
  int retval = hist->getFunctionTag(FuncInfo::GRAMM);
  return retval;
}

int
node_form_function(Hist* hist)
{
  int retval = hist->getFunctionTag(FuncInfo::FORMFUNC);
  //static int PP = 0;
  //if (!PP) PP = Term::get(string("PP"))->toInt();
  //if (retval == NOFUNC_VAL && fh->back && fh->back->term == PP)
    //retval = fh->back->funcTag[FuncInfo::FORMFUNC];
  return retval;
}

int
node_topicalised(Hist* hist)
{
  int retval = hist->getFunctionTag(FuncInfo::TPC);
  return retval;
}

int
node_misc_function(Hist* hist)
{
  int retval = hist->getFunctionTag(FuncInfo::MISC);
  return retval;
}

int
node_alt_head_POS(Hist* hist)
{
  // Currently only defined for PPs, the "alt head" is the head of the object
  static int PP = 0;
  if(!PP) PP = Term::get(string("PP"))->toInt();

  if (hist->getLabel() != PP)
    return false;
  if (hist->getChildren().size() < 2)
    return false;

  Hist* secondChild = *++hist->getChildren().begin();
  const Term* label = Term::fromInt(secondChild->getHeadPOS());
  assert(label);
  if(!label->terminal_p())
    {
      cerr << "Bad head Part of Speech: " << *label << " in " <<endl;
      //cerr << *hist << endl;
      assert(label->terminal_p());
    }
  return label->toInt();
}

int
node_alt_head(Hist* hist)
{
  // Currently only defined for PPs, the "alt head" is the head of the object
  static int PP = 0;
  if(!PP) PP = Term::get(string("PP"))->toInt();

  if (hist->getLabel() != PP)
    return false;
  if (hist->getChildren().size() < 2)
    return false;


  Hist* secondChild = *++hist->getChildren().begin();
  return secondChild->getHead();
}


int
node_is_head_containing_child(Hist* hist)
{
  const Hist* parent = hist->getParent();
  if (!parent) return 0;

  int headPosition = parent->getHeadPosition();
  Hist* headofparent = ith(headPosition, parent->getChildren());
  return (headofparent == hist);
}

int
node_parent_form_function(Hist* hist)
{
  if (!hist->getParent()) return 0;
  int retval = hist->getParent()->getFunctionTag(FuncInfo::FORMFUNC);
  return retval;
}

int
node_label_backoff(Hist* hist)
{
  return Term::fromInt(hist->getLabel())->getSupertag()->toInt();
}

int
node_parent_label_backoff(Hist* hist)
{
  static int S1 = 0;
  if(!S1) S1 = Term::get(string("S1"))->toInt();

  if (hist->getParent())
    return Term::fromInt(hist->getParent()->getLabel())->getSupertag()->toInt();
  else return S1;
}

int
node_grandparent_label_backoff(Hist* hist)
{
  static int S1 = 0;
  if(!S1) S1 = Term::get(string("S1"))->toInt();

  if (hist->getParent() && hist->getParent()->getParent())
    return Term::fromInt(hist->getParent()->getParent()->getLabel())
      		->getSupertag()->toInt();
  else return S1;
}

int
node_head_cluster(Hist* hist)
{
  int head = node_head(hist);
  if (head == -1) return 0;
  int head_POS = node_head_POS(hist);
  int cluster = Cluster::of(head_POS, head);
  return cluster;
}

int
node_alt_head_cluster(Hist* hist)
{
  int head = node_alt_head(hist);
  if (head == -1) return 0;
  int head_POS = node_alt_head_POS(hist);
  int cluster = Cluster::of(head_POS, head);
  return cluster;
}

int
node_parent_head_cluster(Hist* hist)
{
  int head = node_parent_head(hist);
  if (head == -1) return 0;
  int head_POS = node_parent_head_POS(hist);
  int cluster = Cluster::of(head_POS, head);
  return cluster;
}

int
node_grandparent_head_cluster(Hist* hist)
{
  int head = node_grandparent_head(hist);
  if (head == -1) return 0;
  int head_POS = node_grandparent_head_POS(hist);
  int cluster = Cluster::of(head_POS, head);
  return cluster;
}

int
node_head_POS_backoff(Hist* hist)
{
  int head_POS = node_head_POS(hist);
  int supertag = Term::fromInt(head_POS)->getSupertag()->toInt();
  return supertag;
}

int
node_alt_head_POS_backoff(Hist* hist)
{
  int head_POS = node_alt_head_POS(hist);
  int supertag = Term::fromInt(head_POS)->getSupertag()->toInt();
  return supertag;
}

int
node_parent_head_POS_backoff(Hist* hist)
{
  int head_POS = node_parent_head_POS(hist);
  int supertag = Term::fromInt(head_POS)->getSupertag()->toInt();
  return supertag;
}

int
node_grandparent_head_POS_backoff(Hist* hist)
{
  int head_POS = node_grandparent_head_POS(hist);
  int supertag = Term::fromInt(head_POS)->getSupertag()->toInt();
  return supertag;
}

int
node_has_CLR_context(Hist* hist)
{
  bool clr = CLRInfo::isCLRContext(CurrentSentence, hist);
  //if (clr) cout << "Is CLR! ----------------------------------------\n";
  return clr;
}

int
node_parent_rule(Hist* hist)
{
  const Hist* parent = hist->getParent();
  if (!parent) return -1;
  int parLabel = parent->getLabel();
  vector<int> childLabels;
  for (list<Hist*>::const_iterator it = parent->getChildren().begin();
       it != parent->getChildren().end();
       ++it)
  {
    childLabels.push_back((*it)->getLabel());
  }

  const Rule* rule = Rule::getRule(parLabel, childLabels);
  if (!rule) return -1;
  else return rule->toInt();
}

void
addSubFeatureFunctions()
{
  /*
    0 t  fh_term
    1 l  fh_parent_term
    2 u  fh_pos
    3 h  fh_head
    4 i  fh_parent_head
    5 T  fh_true
    6 v  fh_parent_pos
    7 b  fh_term_before
    8 a  fh_term_after
    9 m  fh_grandparent_term
    10 w fh_grandparent_pos
    11 j fh_grandparent_head
    12 c fh_ccparent_term
    13 L1 fh_left1
    14 L1 fh_left2
    15 R1 fh_right1
    16 R1 fh_right2
    17 Qr fh_noopenQr
    18 L0 fh_left0;
    19 L3 fh_left3
    20 R3 fh_right3
    21 Qr fh_noopenQl
    22 Bl fh_Bl
    23 Br fh_Br
    24 C  fh_complement
    25 g  tree_grammatical_function
    26 f  tree_form_function
    27 TP  tree_topicalised
    28 M  tree_misc_function
    29 U  tree_alt_pos
    30 H  tree_alt_head
    31 HC tree_head_containing_child
    32 pf tree_parent_form_function
    33 T  tree_term_backoff
    34 L  tree_parent_term_backoff
    35 M  tree_grandparent_term_backoff
    36 x  tree_head_cluster
    37 X  tree_alt_head_cluster
    38 y  tree_parent_head_cluster
    39 z  tree_grandparent_head_cluster
    40    head POS backoff
    41    alt head POS backoff
    42    parent's head POS backoff
    43    gp's head POS backoff
    44    has CLR context
    45    parent's rule expansion
    */
  int (*funs[MAXFEATS])(Hist*)
    = {node_label, node_parent_label, node_head_POS, node_head,
       node_parent_head, node_true, node_parent_head_POS,
       node_left_sibling_label, node_right_sibling_label,
       node_grandparent_label, node_grandparent_head_POS, node_grandparent_head,
       node_parent_cc_label, node_left1, node_left2, node_right1, node_right2,
       node_noopenQr, node_left0, node_left3, node_right3, node_noopenQl,
       node_Bl, node_Br,
       node_complement, node_grammatical_function, node_form_function,
       node_topicalised, node_misc_function,
       node_alt_head_POS, node_alt_head,
       node_is_head_containing_child,
       node_parent_form_function,
       node_label_backoff, node_parent_label_backoff, 
       node_grandparent_label_backoff,
       node_head_cluster, node_alt_head_cluster, node_parent_head_cluster,
       node_grandparent_head_cluster ,
       node_head_POS_backoff, node_alt_head_POS_backoff, 
       node_parent_head_POS_backoff, node_grandparent_head_POS_backoff,
       node_has_CLR_context, node_parent_rule
    };
  int i;
  for(i = 0 ; i < MAXFEATS ; i++) SubFeature::Funs[i] = funs[i];
}

static
bool set_has_intersection(intsetiter first1, intsetiter last1,
			intsetiter first2, intsetiter last2) {
    while (first1 != last1 && first2 != last2)
	if (*first1 < *first2)
	    ++first1;
	else if (*first2 < *first1)
	    ++first2;
	else {
	  return true;
	}
    return false;
}

