#include "FeatureTree.h"
#include <string>
#include <set>
#include <cmath>
#include "Feature.h"

FeatureTree* FeatureTree::roots_[NUMCALCS];
extern int globalDebugLevel;

FeatureTree* 
FeatureTree::
follow(int val, int auxCnt)
{
  if(!auxCnt) return getChild(val);
  if(!auxNd) return NULL;
  return auxNd->follow(val, auxCnt-1);
}

/* basic format
   assumedNum //e.g., 55 (np)
        rule# count
        ...
	--
	48 //this would be np under adjp
             rule# count
	     ...
        --  //end  of rules
        -- //end of 48
	49
	...
	--  //end of 49
	...
	--  // end of g(rtl)s
   Pringint is brown into 2 procedures.  The first takes a position
   in the tree and determines if the data there supports it (i.e.,
   there is enough for any one conditioned value).  if so, it
   prints out the link value, and the conditioned values. It then
   calls f2 on itself.
*/

void FeatureTree::
topLevelRead(istream& is)
{
  init();
  int done = 0;
  int c = 0;
  while(is && !done)
    {
      int val = readOneLevel0(is,c);
      if(val == -1) done = 1;
      c++;
    }
  roots_[Feature::whichInt] = this;
}
 
int
FeatureTree::
readOneLevel0(istream& is, int arrayPos)
{
  int nextInd;
  string nextIndStr;
  is >> nextIndStr;
  if(!is) return -1;
  if(nextIndStr == "Selected") return -1;
  nextInd = atoi(nextIndStr.c_str());
  FeatureTree* nft = getNextFeatureTree(nextInd, arrayPos);
  nft->read(is,Feature::ftTree[Feature::whichInt].left);
  nft->back = this; 
  return nextInd;
}

void
FeatureTree::
read(istream& is, FTypeTree* ftt)
{
  string indStr;
  int indI, cfeats, ctrees, cf;
  is >> count;
  is >> cfeats;
  is >> ctrees;
  if (globalDebugLevel > 60)
    cerr << "Reading tree with count " << count
      << ", cfeats " << cfeats
      << ", ctrees " << ctrees << endl;
  adjustSize(cfeats, ctrees);
  for(cf = 0 ; cf < cfeats ; cf++)
  {
    is >> indI;
    Feat& nf = getNextFeat(indI, cf);
    if(Feat::Usage != PARSE)
    {
      int v;
      is >> v;
      nf.cnt() = v;
    }
    else
    {
      float v;
      is >> v;
      nf.g() = v;
    }
  }
  othReadFeatureTree(is, ftt, ctrees);
}

void
FeatureTree::
othReadFeatureTree(istream& is, FTypeTree* ftt, int ctrees)
{
  setFeatureInt(ftt->n);
  string indStr;
  int indI, c;
  for(c = 0 ; c < ctrees ; c++)
    {
      is >> indI;
      FeatureTree* ntr = getNextFeatureTree(indI, c, this);
      assert(ftt->left);
      ntr->read(is, ftt->left);
    }
  if(!ftt->right) return;
  assert(!auxNd);
  is >> indStr;
  if(indStr != "A")
    {
      cerr << "fi = " << ftt->n << " " << ctrees << " " << indStr
	   << " " << getInd() << " " << count << endl;
      for(int i = 0 ; i < 5 ; i++)
	{
	  string tmp;
	  is >> tmp;
	  cerr << tmp << " ";
	}
      cerr << endl;
      cerr << ftt->right->n << endl;
      assert(indStr == "A");
    }
  int ac;
  is >> ac;  
  performAuxNdVoodoo(ac);
  auxNd->othReadFeatureTree(is, ftt->right, ac);
}
