
#ifndef FULLHIST_H
#define FULLHIST_H

#include <string>
#include <list>
#include "Item.h"
#include "Edge.h"
#include "Wrd.h"
#include "Hist.h"

using namespace std;

class FullHist: public Hist
{
public:
  FullHist() : e(NULL), back(NULL) 
	{ memset (funcTag, 0, 4 * sizeof(*funcTag) ); }
  FullHist(Edge* e1) : e(e1), back(NULL)	
	{ memset (funcTag, 0, 4 * sizeof(*funcTag) ); }
  FullHist(int tint, FullHist* fh) : term(tint), back(fh), pos(-1),hd(NULL)
	{ memset (funcTag, 0, 4 * sizeof(*funcTag) ); }
  FullHist* extendByEdge(Edge* e1);
  FullHist* extendBySubConstit();
  FullHist* retractByEdge();
  friend ostream& operator<<(ostream& os, const FullHist& fh);
  int hpos;
  int pos;
  int term;
  int preTerm;
  Edge* e;
  const Wrd* hd;
  int      funcTag[4];
  FullHist* back;
  list<Hist*> subtrees;  
  list<Hist*>::iterator sti;

 public:
  int getHeadPosition() const { return hpos; }
  int getPosition() const { return pos; }
  int getLabel() const { return term; }
  int getStart() const { return e->start(); }
  int getFinish() const { return e->loc(); }
  int getHeadPOS() const { return preTerm; }
  bool isHeadKnown() const { return (bool)hd; }
  int getHead() const { return hd->toInt(); }
  string getHeadString() const { return hd->lexeme(); }
  int getFunctionTag(int type) const { return funcTag[type]; }

  void setFunctionTag(int type, int val) { funcTag[type] = val; }

  const Hist* getParent() const { return back; }
  const list<Hist*>& getChildren() const { return subtrees; }
  int getNumChildren() const;
  int getChildLabel (int i) const;

  bool isConjunctivePhrase() const { return e->ccInd(); }
};

#endif
