
#include "FuncInfo.h"
#include "InputTree.h"

//#define NPOS -1

FuncInfo*  FuncInfo::array_[400];
map<string, FuncInfo*> FuncInfo::funcMap_;
int    FuncInfo::lastFuncInfoInt_ = 0;
set<int> FuncInfo::functionalTags[4];

FuncInfo::
FuncInfo()
: name_( "" ), num_( -1 )
{}

FuncInfo::
FuncInfo(const string s, int num )
: name_( s ), num_( num )
{}

FuncInfo::
FuncInfo( const FuncInfo& src )
: name_( src.name() ), 
  num_(src.toInt())
{}

ostream&
operator<< ( ostream& os, const FuncInfo& f )
{
    os << f.name();
    return os;
}

int
FuncInfo::
operator== ( const FuncInfo& rhs ) const
{
    if( this == &rhs || name_ == rhs.name() )
	return 1;
    return 0;
}

void
FuncInfo::
init(string & prefix)
{
  string          fileName(prefix);
  fileName += "funcs.txt";
  ifstream           strm(fileName.c_str(), ios::in);
  if (!strm)
    {
      cerr << "Can't open function tags file " << fileName << endl;;
      return;
    }
  
  string          funcName;
  int n;
  array_[NOFUNC_NUM] = new FuncInfo("NONE", NOFUNC_NUM);
  n = 1;
  while (strm >> funcName)
    {
      FuncInfo* nextFuncInfo = new FuncInfo(funcName, n);
      funcMap_[nextFuncInfo->name()] = nextFuncInfo;
      array_[n] = nextFuncInfo;
      n++;
      assert(n < 400);
    }
  lastFuncInfoInt_ = n-1;
  strm.close();

  functionalTags[GRAMM].insert(get("DTV")->toInt());
  functionalTags[GRAMM].insert(get("LGS")->toInt());
  functionalTags[GRAMM].insert(get("PRD")->toInt());
  functionalTags[GRAMM].insert(get("PUT")->toInt());
  functionalTags[GRAMM].insert(get("SBJ")->toInt());
  functionalTags[GRAMM].insert(get("VOC")->toInt());

  functionalTags[FORMFUNC].insert(get("NOM")->toInt());
  functionalTags[FORMFUNC].insert(get("ADV")->toInt());
  functionalTags[FORMFUNC].insert(get("BNF")->toInt());
  functionalTags[FORMFUNC].insert(get("DIR")->toInt());
  functionalTags[FORMFUNC].insert(get("EXT")->toInt());
  functionalTags[FORMFUNC].insert(get("LOC")->toInt());
  functionalTags[FORMFUNC].insert(get("MNR")->toInt());
  functionalTags[FORMFUNC].insert(get("PRP")->toInt());
  functionalTags[FORMFUNC].insert(get("TMP")->toInt());

  functionalTags[TPC].insert(get("TPC")->toInt());

  //functionalTags[MISC].insert(get("CLR")->toInt());
  functionalTags[MISC].insert(get("CLF")->toInt());
  functionalTags[MISC].insert(get("HLN")->toInt());
  functionalTags[MISC].insert(get("TTL")->toInt());
}

const FuncInfo*
FuncInfo::
get(const string& nm)
{
  if (funcMap_.count(nm))
    return funcMap_[nm];
  else 
    return NULL;
}

const FuncInfo*
FuncInfo::
get (const string& str, int funcType)
{
  string getname = str;
  int hyphpos = getname.find("-");
  if (hyphpos == string::npos || isdigit(getname[hyphpos+1]))
    return NOFUNC_NUM;

  do {
    getname = getname.substr (hyphpos + 1);
    hyphpos = getname.find_first_of("-=");
    const FuncInfo* fi = get(getname.substr(0, hyphpos));
    if (fi)
    {
      int fint = fi->toInt();
      if (functionalTags[funcType].count(fint))
        return array_[fint];
    }
  } while (hyphpos != string::npos);
  return NULL;
}
int
FuncInfo::
getEqNum(const string& getname)
{
  int eqpos = getname.find("=");
  if (eqpos == string::npos)
    return -1;
  else 
  {
    eqpos ++;
    int endpos = getname.find("-", eqpos);
    	// if 2nd arg is npos, substr returns remainder of string
    string numstr = getname.substr(eqpos, endpos);
    return atoi(numstr.c_str());
  }
}

int
FuncInfo::
getHyphNum(const string& getname)
{
  int hyphpos = getname.find_last_of("-");
  if (hyphpos == string::npos)
    return -1;
  else 
  {
    hyphpos ++;
    if (!isdigit(getname[hyphpos]))
      return -1;
    else
      return atoi(getname.substr(hyphpos).c_str());
  }
}

int
FuncInfo::
getFuncInt (const string& str, int funcType)
{
  string getname = str;
  int hyphpos = getname.find("-");
  if (hyphpos == string::npos || isdigit(getname[hyphpos+1]))
    return NOFUNC_NUM;

  do {
    getname = getname.substr (hyphpos + 1);
    hyphpos = getname.find_first_of("-=");
    const FuncInfo* fi = get(getname.substr(0, hyphpos));
    if (fi)
    {
      int fint = fi->toInt();
      if (functionalTags[funcType].count(fint))
        return fint;
    }
  } while (hyphpos != string::npos);
  return NOFUNC_NUM;
}


int
FuncInfo::
getFirstFuncInt (string getname)
{
  int hyphpos = getname.find("-");
  if (hyphpos == string::npos || hyphpos == 0 || isdigit(getname[hyphpos+1]))
    return NOFUNC_NUM;

  getname = getname.substr (0, hyphpos + 1);
  hyphpos = getname.find_first_of("-=");
  if (hyphpos == string::npos)
    return get(getname)->toInt();
  else
    return get(getname.substr(0, hyphpos))->toInt();
}

set<int>
FuncInfo::
getAllFuncInts(string getname)
{
  int eqpos = getname.find("=");
  if (eqpos != string::npos) 
    getname = getname.substr(0, eqpos);

  set<int> retval;
  int hyphpos = getname.find("-");
  if (hyphpos == string::npos)
    return retval;
  getname = getname.substr(hyphpos + 1);
  while ((hyphpos = getname.find("-")) != string::npos)
  {
    retval.insert(get(getname.substr(0, hyphpos))->toInt());
    getname = getname.substr(hyphpos + 1);
  }
  if (!isdigit(getname[0]))
  {
    const FuncInfo* func = get(getname);
    if (func == NULL)
      warn ("FuncInfo::getAllFuncInts: Bad function tag " + getname);
    else
      retval.insert(func->toInt());
  }

  return retval;
}

vector<const FuncInfo*> 
//FuncInfo::
getAllFuncs(const string& getname)
{
  vector<const FuncInfo*> retval;

  return retval;
}
