
#ifndef FUNCINFO_H
#define FUNCINFO_H

#include <list>
#include <map>
#include <set>
#include <iostream>
#include <fstream>
#include <vector>
#include <string>
#include "utils.h"

using namespace std;

#define NOFUNC_NUM 0
#define NOFUNC_VAL NULL

class InputTree;

///
class FuncInfo
{
public:
    ///
    FuncInfo();			// provided only for maps.
    ///
    FuncInfo( const string s, int n );
    ///
    FuncInfo( const FuncInfo& src );
    ///
    int              toInt() const { return num_; }
    ///
    const string& name() const {  return name_;  }
    ///
    friend ostream& operator<< ( ostream& os, const FuncInfo& t );
    ///
    friend ostream& operator>> ( istream& os, const FuncInfo& t );
    ///
    int		operator== (const FuncInfo& rhs ) const;
    ///
    static const FuncInfo* get(const string& getname);
    ///
    static const FuncInfo* get(const string& getname, int funcType);
    ///
    static int getEqNum(const string& getname);
    ///
    static int getHyphNum(const string& getname);
    ///
    static int getFuncInt(const string& getname, int funcType);
    ///
    static int getFirstFuncInt(string getname);
    ///
    static set<int> getAllFuncInts(string getname);
    ///
    static void  init(string& prefix);
    ///
    static const FuncInfo* fromInt(int i) 
      { myassert(i >= 0, i);
	myassert(i <= lastFuncInfoInt_, i); 
	myassert(array_[i], i); 
	return array_[i]; }
    ///
    static int  lastFuncInfoInt() { return lastFuncInfoInt_; }
    ///
    enum { GRAMM, FORMFUNC, TPC, MISC };
    ///
    static set<int> functionalTags[4];

    static const int NONE = 0;
private:
    ///
    string* namePtr() { return (string*)&name_; }
    ///
    int		num_;
    ///
    const string name_;
    ///
    static FuncInfo*  array_[400];
    ///
    static map<string, FuncInfo*> funcMap_ ;
    ///
    static int    lastFuncInfoInt_;

};

/// a pair holding a function tag (in int form) and its probability
typedef pair<double,int> FunctionTagPair;

#endif /* ! FUNCINFO_H */

