
#include "MeChart.h"
#include "edgeSubFns.h"
#include "GotIter.h"

#include "TreeHist.h"

#define FEAT_BRANCH -2
#define FEAT_HAS_NO_VAL -3

static int depth = 0;

static void prDp()
{
  for(int i = 0 ; i < depth ; i++)
    cerr << " ";
}

static inline int printDebug ()
{
  return MeChart::printDebug();
}

static float bucketLims[14] =
  {0, .003, .01, .033, .09, .33, 1.01, 2.01, 5.1, 12, 30, 80, 200, 600};

static int 
bucket(float val)
{
  for(int i = 0 ; i < 14 ; i++)
    if(val <= bucketLims[i])
      return i;
  return 14;
}

float
ITProb(int cVal, TreeHist* th, int whichInt)
{
  if(printDebug() > 38)
    {
      prDp();
      //cerr << "meP" << whichInt << "(" << cVal << " | " << *th << ")" <<endl;
    }
  int subfVals[MAXNUMFS];
  FeatureTree* ginfo[MAXNUMFS];  
  ginfo[0] = FeatureTree::roots(whichInt);
  float smoothedPs[MAXNUMFS];

  float ans = 1;

  if (printDebug() > 60)
    cerr << "Out of " << Feature::total[whichInt] << endl;
  for(int i = 1 ; i <= Feature::total[whichInt] ; i++)
    {
      if (printDebug() > 40)
        cerr << "This: " << i << "...";
      ginfo[i] = NULL;
      Feature* feat = Feature::fromInt(i, whichInt); 
      /* e.g., g(rtlu) starts from where g(rtl) left off (after tl)*/
      int searchStartInd = feat->startPos;

      FeatureTree* strt = ginfo[searchStartInd];
      if(!strt)
	{
	  continue;
	}
      SubFeature* sf = SubFeature::fromInt(feat->subFeat, whichInt);
      int nfeatV = (*(sf->fun))(th);
      if (printDebug() > 40) 
        cerr << "nfeatv: " << nfeatV << " thnum: " << sf->num << "...";
      FeatureTree* histPt = strt->follow(nfeatV, feat->auxCnt); 
      ginfo[i] = histPt;
      if(i == 1)
	{
	  smoothedPs[0] = 1;
	  Feat* f =histPt->getFeat(cVal);
	  if(!f)
	    {
	      if(printDebug() > 30)
		{
		  prDp();
		  cerr << "Zero p" << feat->name << " " << nfeatV << endl;
		}
	      return 0.0;
	    }
	  smoothedPs[1] = f->g();
	  if(printDebug() > 38)
	    {
	      prDp();
	      cerr << i << " " << nfeatV << " " << smoothedPs[1] << endl;
	    }
	  for(int j = 2; j <= Feature::total[whichInt] ; j++)
	    smoothedPs[j] = 0;
	  ans = smoothedPs[1];
	  continue;
	}
      if(!histPt)
	{
	  continue;
	}
      float estm = histPt->getCount() * smoothedPs[1];
      int b = bucket(estm);

      Feat* ft = histPt->getFeat(cVal);
      float unsmoothedVal;
      if(!ft) unsmoothedVal = 0;
      else unsmoothedVal = ft->g();
      float lam = Feature::getLambda(whichInt, i, b);
      float uspathprob = lam*unsmoothedVal;
      float osmoothedVal = smoothedPs[searchStartInd];
      //float osmoothedVal = smoothedPs[i-1]; //for deleted interp.
      float smpathprob = (1-lam)*osmoothedVal;
      float nsmoothedVal = uspathprob+smpathprob;
      if(printDebug() > 38)
	{
	  prDp();
	  cerr << i << " " << nfeatV << " "
	       << estm << " " << b <<" "<<unsmoothedVal << " " << lam << " " 
	       << nsmoothedVal <<  endl;
	}
      smoothedPs[i] = nsmoothedVal;
      ans *= (nsmoothedVal/osmoothedVal);
    }
  if(whichInt == HCALC) ans *= 600;
  if(printDebug() > 30)
    {
      prDp();
      cerr<<"p"<<whichInt<< "(" << cVal << "|" << /* *h  << */ ") = " << ans << endl;
    }
  return ans;
}

static void
printFeatTree(ostream& os, int whichInt, int* featvals, int numfeats)
{
  static char funcName[][4] = {"gr", "ff", "tp", "mf"};
  os << funcName[whichInt - GRCALC];
  for (int i = 0; i < numfeats; ++i)
  {
    if (featvals[i] == FEAT_BRANCH) 
      os << " A";
    else
      os << " " << featvals[i];
  }
  os << endl;
}

/* this feature tree dump is for use in the outside ITProb 
   utility, which can only handle one path at a time; it will output
   one line per leaf; branches are noted using A notation, that is,
   at a branch, zero 'A's signify the leftmost branch, one 'A' signifies
   the next branch over, and so on. */
void
printFeatTrees(ostream& os, TreeHist* th, int target)
{
  int featPath[MAXNUMFS];
  int pathIndexCache[MAXNUMFS];
  int curPathIndex = -1;
  int parCacheIndex;

  FeatureTree* cache[MAXNUMFS];  
  cache[0] = FeatureTree::roots(target);

  /* Iterates through feature tree in depth-first order */
  for(int i = 1 ; i <= Feature::total[target] ; i++)
    {
      cache[i] = NULL;

      /* Get ith feature in tree (DFS order) */
      Feature* feat = Feature::fromInt(i, target); 

      /* e.g., g(rtlu) starts from where g(rtl) left off (after tl)*/
      parCacheIndex = feat->startPos;

      /* If the parent of this feature isn't the previous feature, then
	 the previous feature was a leaf, ergo we need to print a line */
      if (parCacheIndex != i - 1)
      {
	printFeatTree(os, target, featPath, curPathIndex + 1);

	/* Now we need to insert an 'A' at the appropriate point in the
	   feature path.  The parCacheIndex notes where the parent is
	   in the DF traversal (and thus the cache); we need to subtract 
	   from that an offset. */
	curPathIndex = ++pathIndexCache[parCacheIndex];
	featPath[curPathIndex++] = FEAT_BRANCH;
      }
      else
	++curPathIndex;
      /* curPathIndex now points to wherever the current item will go
	 in the path. */

      FeatureTree* featParent = cache[parCacheIndex];
      if(!featParent)
      {
	cache[i] = NULL;
        pathIndexCache[i] = curPathIndex;
	featPath[curPathIndex] = FEAT_HAS_NO_VAL;
      }
      else
      {
	/* Back in the day, each feature could have multiple subfeatures;
	   now they each have exactly one.  The function pointer is kept
	   there, though, so it sticks around for hysterical reasons. */
	SubFeature* sf = SubFeature::fromInt(feat->subFeat, target);
	int curFeatVal = (*(sf->fun))(th);

	/* Now that we have the value of the current feature, we know
	   which subtree to take (ultimately doesn't matter for printing
	   out the tree, except that the tree might peter out for lack
	   of training data).  This value could be NULL if no subtree
	   of the correct value exists. */
	cache[i] = featParent->follow(curFeatVal, feat->auxCnt); 
	pathIndexCache[i] = curPathIndex;
	featPath[curPathIndex] = curFeatVal;
	if (printDebug() > 40) 
	  cerr << "nfeatv: " << curFeatVal << " thnum: " << sf->num << "...";
      }
    }
  printFeatTree(os, target, featPath, curPathIndex + 1);
}
