#ifndef _INPUTPARSETREE_H_
#define _INPUTPARSETREE_H_

#include "GenericTree.h"
#include "InputTree.h"

class InputParseTree : public GenericTree
{
  public:
    InputParseTree (InputTree* intr, InputParseTree* p=NULL);

    virtual GenericTree* getParent() const { return parent; }
    virtual const vector<GenericTree*>& getChildren() const { return children; }
    virtual const string& getLabel() const { return tree->getLabel(); }
    virtual const string& getAltLabel() const { return altlabel; }
    virtual const vector<string>& getTags() const { return tags; }
    virtual bool isLeaf() const { return false; }

    virtual void linePrint (ostream& os) const;

  private:
    InputTree* tree;
    string altlabel;
    InputParseTree* parent;
    vector<GenericTree*> children; //Really InputParseTrees. Stupid C++!
    vector<string> tags;
};

class InputParseTreeLeaf : public InputParseTree
{ 
  public:
    InputParseTreeLeaf (string w, InputParseTree* p) 
      : word(w), InputParseTree(NULL, p)
      {  }

    virtual const string& getLabel() const { return word; }
    virtual bool isLeaf() const { return true; }

  private:
    string word;
};


inline
InputParseTree::InputParseTree (InputTree* intr, InputParseTree* p) 
: tree(intr), altlabel(""), parent(p), children(), tags()
{ 
  if (intr == NULL) return;
  if (intr->getSubTrees().empty())
  { 
    children.push_back(new InputParseTreeLeaf(intr->word(), this));
  }
  else
  {
    for (InputTreesIter it = intr->getSubTrees().begin();
	it != intr->getSubTrees().end();
	++it)
    { 
      children.push_back(new InputParseTree(*it, this));
    }
  }
  for (int i = 0; i < 4; ++i)
  {
    int tag = intr->getFuncTag(i);
    if (tag != 0) tags.push_back(FuncInfo::fromInt(tag)->name());
  }
}

#endif
