#include "MapFeatureTree.h"
#include <string>
#include "math.h"

int MapFeatureTree::totParams = 0;
int MapFeatureTree::minCount = 1;

MapFeatureTree*
MapFeatureTree::
next(int val, int auxCnt)
{
  if(!auxCnt)
    {
      MapFeatureTree* ans = subtree[val];
      if(ans) return ans;
      ans = new MapFeatureTree(val,this);
      subtree[val] = ans;
      return ans;
    }
  if(!auxNd) auxNd = new MapFeatureTree(AUXIND,this);
  return ((MapFeatureTree*) auxNd)->next(val, auxCnt-1);
}

/* basic format
   assumedNum //e.g., 55 (np)
        rule# count
        ...
	--
	48 //this would be np under adjp
             rule# count
	     ...
	     --
	49
	...
	--  //end of 49
	...
	--  // end of g(rtl)s
	A  //indicates that there is another group of features here.
	2  //np's with head's pos->toInt() == 2;
   Pringint is brown into 2 procedures.  The first takes a position
   in the tree and determines if the data there supports it (i.e.,
   there is enough for any one conditioned value).  if so, it
   prints out the link value, and the conditioned values. It then
   calls f2 on itself.
*/

static int
markedSize(FTreeMap& sub)
{
  if(Feat::Usage != SEL) return sub.size();
  int i = 0;
  FTreeMap::iterator subFeatIter = sub.begin();
  MapFeatureTree* subT;
  for( ; subFeatIter != sub.end() ; subFeatIter++ )
    {
      subT = (*subFeatIter).second;
      if(subT->getMarked() >= 0) i++;
    }
  return i;
}

void
MapFeatureTree::
printFTree(int asVal, ostream& os)
{
  printFfCounts(asVal, 0, os, Feature::ftTree[Feature::whichInt].left);
}

void
MapFeatureTree::
printFfCounts(int asVal, int depth, ostream& os, FTypeTree* ftt)
{
  //cerr << "pff " << " " << depth << endl;
  FeatMap::iterator conditionedIter = feats.begin();
  int i;
  if(count < minCount) return;
  if(Feat::Usage == SEL && marked < 0) return;
  for(i = 0 ; i < depth ; i++)
    os << "\t";
  os << getInd() << "\t" ;
  os << count;
  os << "\t" << feats.size();
  os << "\t" << markedSize(subtree);
  os << "\n"; 

  /* now print out the counts for the features */
  conditionedIter = feats.begin();
        
  for(; conditionedIter != feats.end() ; conditionedIter++)
    {
      Feat* ft = &((*conditionedIter).second);
      int cnt = ft->cnt();
      int fval = (*conditionedIter).first;
      for(i = 0 ; i <= depth ; i++)
	os << "\t";
      os << fval << "\t" ;
      // during iterative scaling we print out feature tree to look;
      // at gammas;
      if(Feat::Usage == ISCALE)
	{
	  float gval = (*conditionedIter).second.g();
	  os << gval;
	}
      else
	os << cnt;
      os << endl;
      totParams++;
    }
  //for(i = 0 ; i <= depth ; i++) os << "\t";
  //os << "--\n"; //This indicates end of feature/count pairs;
  os << "\n";
  printFfCounts2(asVal, depth, os, ftt);
}

void
MapFeatureTree::
printFfCounts2(int asVal, int depth, ostream& os, FTypeTree* ftt)
{
  //cerr << "pff " << asVal << endl;
  FTreeMap::iterator subFeatIter = subtree.begin();
  MapFeatureTree* subT;
  for( ; subFeatIter != subtree.end() ; subFeatIter++ )
    {
      subT = (*subFeatIter).second;
      subT->printFfCounts(asVal, depth+1, os, ftt->left);
    }
  if(ftt->right)
    {
      for(int i = 0 ; i < depth ; i++)
	os << "\t";
      os << "A\t";
      if(!auxNd)
	{
	  os << "0\n";
	  return;
	}
      MapFeatureTree* mapAuxNd = (MapFeatureTree*) auxNd;
      os << markedSize(mapAuxNd->subtree);
      os << "\n";
      //cerr << "Ap " << auxNd->subtree.size() << endl;
      mapAuxNd->printFfCounts2(asVal, depth, os, ftt->right); //???;
    }
}
  
ostream&
operator<< ( ostream& os, const MapFeatureTree& t )
{
  int temp[10];
  int i = 0;
  const FeatureTree* ft = &t;
  for( ; ft ; ft = ft->getParent() )
    if(ft->getInd() >= -1) temp[i++] = ft->getInd();
  os << "<";
  for(int j = i-1 ; j >= 0 ; j--)
    {
      os  << temp[j] ;
      if(j > 0) os << ",";
    }
  os << ">";
  return os;
}

      
    


