
#ifndef MAPFEATURETREE_H
#define MAPFEATURETREE_H

#include "FeatureTree.h"
#include <map>
#include <set>
#include "utils.h"

class MapFeatureTree;
typedef std::map<int,MapFeatureTree*, std::less<int> > FTreeMap;
typedef std::map<int, int, std::less<int> > IntIntMap;
typedef std::map<int, Feat, std::less<int> > FeatMap;
typedef set<int, std::less<int> > IntSet;
typedef std::map<int,IntSet, std::less<int> > IntSetMap;

class MapFeatureIterator : public FeatureIterator
{ 
  private:
    FeatMap::iterator it;
  public:
    MapFeatureIterator(const FeatMap::iterator& i) 
      : it(i) { isMapType = true; }
    Feat* operator*() { return &(*it).second; }
    bool operator==(const MapFeatureIterator& other) const { return it == other.it; }
    bool isEqualTo(const FeatureIterator* other) const
      { return other->isMapType && *this == *(MapFeatureIterator*)other; }
    Feat* operator++() { return &(*++it).second; }
};

class MapFeatureTreeIterator : public FeatureTreeIterator
{ 
  private:
    FTreeMap::iterator it;
  public:
    MapFeatureTreeIterator(const FTreeMap::iterator& i) 
      : it(i) { isMapType = true; }
    FeatureTree* operator*();
    bool operator==(const MapFeatureTreeIterator& other) const  
      { return it == other.it; }
    bool isEqualTo(const FeatureTreeIterator* other) const
      { return other->isMapType && *this == *(MapFeatureTreeIterator*)other; }
    FeatureTree* operator++();
};


class MapFeatureTree : public FeatureTree
{
public:
  MapFeatureTree() : FeatureTree() { init(); }
  MapFeatureTree(int i) : FeatureTree(i) { init(); }
  MapFeatureTree(int i, FeatureTree* b) : FeatureTree(i, b) { init(); }
  MapFeatureTree(istream& is) : FeatureTree() 
    { init(); ind_ = ROOTIND; topLevelRead(is); }

  MapFeatureTree* next(int val, int auxCnt);
  static MapFeatureTree*& root() 
    { return (MapFeatureTree*&) roots_[Feature::whichInt]; }
  void   printFTree(int asVal, ostream& os);

  friend ostream&  operator<<(ostream& os, const MapFeatureTree& ft);

private:
  int featureInt;
  int specFeatures;
  float marked;

  static int       totParams;
  static int       minCount;

  FeatMap feats;
  FTreeMap subtree;

 public:
  void init() { specFeatures = 0; marked = -1; featureInt = -1; }
  FeatureTree* getNextFeatureTree (int index, int arrayPos, 
      					FeatureTree* parent=NULL)
  { 
    MapFeatureTree* retval = subtree[index];
    if (!retval)
    {
      retval = new MapFeatureTree(index, parent);
      subtree[index] = retval;
    }
    return retval;
  }
  Feat& getNextFeat (int index, int arrayPos)
  { 
    Feat& retval = feats[index];
    retval.ind() = index;
    retval.setTree(this);
    return retval;
  }
  void adjustSize (int numFeats, int numTrees)
  { 
  }
  void performAuxNdVoodoo(int ac)
  { 
    auxNd = new MapFeatureTree(AUXIND,this);
  }
  void setFeatureInt(int n) 
  { 
    featureInt = n;
    if (featureInt <= 0 && featureInt != AUXIND)
      myassert(featureInt > 0, featureInt);
  }

  int getFeatureInt() const { return featureInt; }
  int getSpecFeatures() const { return specFeatures; }
  float getMarked() const { return marked; }
  void setMarked(float x) { marked = x; }

  static void setMinCount(int n) { minCount = n; }
  static int getTotParams() { return totParams; }

  int getNumChildren() const { return subtree.size(); }
  int getNumFeats() const { return feats.size(); }

  Feat* getFeat(int id)
    { FeatMap::iterator retval = feats.find(id);
      return (retval == feats.end())? NULL: &(*retval).second; }
  void incFeat(int id) { feats[id].cnt()++; }
  FeatIterator featsBegin() 
    { return FeatIterator(new MapFeatureIterator(feats.begin())); }
  FeatIterator featsEnd() 
    { return FeatIterator(new MapFeatureIterator(feats.end())); }

  FeatureTree* getChild(int id)
    { FTreeMap::iterator retval = subtree.find(id);
      return (retval == subtree.end())? NULL: (*retval).second; }
  FeatTreeIterator childrenBegin() 
    { return FeatTreeIterator(new MapFeatureTreeIterator(subtree.begin())); }
  FeatTreeIterator childrenEnd() 
    { return FeatTreeIterator(new MapFeatureTreeIterator(subtree.end())); }

 friend class FeatTreeIter;
 friend class FeatIter;

 private:
  void read(istream& is, FTypeTree* ftt);
  void othReadFeatureTree(istream& is, FTypeTree* ftt, int cnt);
  void         printFfCounts(int asVal, int depth,
			     ostream& os, FTypeTree* ftt);
  void printFfCounts2(int asVal, int depth, ostream& os, FTypeTree* ftt);
};

inline FeatureTree* 
MapFeatureTreeIterator::operator*() { return (*it).second; }
inline FeatureTree* 
MapFeatureTreeIterator::operator++() { return (*++it).second; }

#endif /* ! MAPFEATURETREE_H */
