
#include "MeChart.h"
#include "FeatureFunctions.h"
#include "edgeSubFns.h"
#include "GotIter.h"
#include "CntxArray.h"
#include "ArrayFeatureTree.h"
#include "parseIntegration.h"

#define FUNC_POSTPROC

typedef set<int, less<int> > intset;
typedef set<int, less<int> >::iterator intsetiter;

bool sufficiently_likely(Edge* edge);
bool sufficiently_likely(const Item* itm);

extern float endFactor;
extern float midFactor;
LeftRightGotIter* globalGi = NULL;

static int depth = 0;

void prDp()
{
  for(int i = 0 ; i < depth ; i++)
    cerr << " ";
}
  
void
MeChart::
init(string path)
{
  //assert(CntxArray::sz == (Feature::total[UCALC] -1));
  CntxArray::sz = Feature::total[UCALC] - 1;
} 

AnswerTree*
MeChart::
findMapParse()
{
  if(printDebug() > 8)
    {
      prDp();
      cerr << "In findMapParse" << endl;
    }
  Item* s = topS();
  assert(s);
  fillInHeads();
  int s1Int = s->term()->toInt();
  FullHist s1Fh(s1Int,NULL);
  AnswerTree* ans = bestParse(s, &s1Fh).second;
  //assert(ans);
  if (!ans)
  {
    cerr << "ans is null!  Voici the current state of the chart: " << endl;
    cerr << *this << endl;
    assert(ans);
  }
  return ans;
}

AnswerTreePair  // when I memoize again this should be a ref
MeChart::
bestParse(Item* itm, FullHist* h)
{
  if(printDebug() > 10)
    {
      prDp();
      cerr << "bestParse(" << *itm << ", ...)" << endl;
    }

  int bestFf = -1;
  double bestFfP = 0;
  AnswerTree* bestFfAT = NULL;

#ifdef INTEGRATED_FUNCTAG
  // calculate the possible function tags
  vector<FunctionTagPair> funcTags = tagsForNode(INTEGRATED_FUNCTAG, itm, h, false);

  vector<FunctionTagPair>::iterator fti = funcTags.begin();
  /* Cycle through the possible func tags; for each, set the 
     fullhist's functag, then go through the usual spiel, recursing
     downward.  If it was the best we've seen, keep it. */
  for (; fti != funcTags.end(); ++fti)
  {
    double ffProb = fti->first;
    int ffTag = fti->second;
    h->funcTag[INTEGRATED_FUNCTAG] = ffTag;
#else
    double ffProb = 1.0;
    int ffTag = 0;
#endif

    double bestP = 0;
    AnswerTree* bestAT = NULL;
    string bestW;
    int bestPos = -1;

    /* NB: recordedBP returns an lvalue; thus the value assigned to atp 
       is itself the memo. */
    AnswerTreePair& atp = recordedBP(itm, h);  // memo
    if(atp.first >= 0)
      {
	if(printDebug() > 19)
	  {
	    prDp();
	    cerr << "already known bestParse(" << *itm << ", ...) has p = "
		 << atp.first << endl;
	  }
	bestP = atp.first;
	bestAT = atp.second;
	//return atp; // if we've memoized this itm/hist pair, return stored
      }
    else
    {
      atp.first = 0;

      int itermInt = itm->term()->toInt();
      /* For all possible head POSs, we have a list of all possible heads;
	 for each of those we have a list of edges that might lead to that. */
      PosMap& pm = itm->posAndheads(); 
      PosIter pi = pm.begin();
      for( ; pi != pm.end() ; pi++ )  // step through head-POSs
	{
	  int posInt = (*pi).first;
	  if(printDebug() > 16)
	    {
	      prDp();
	      cerr << "consider Pos(" << *itm << ") = " << posInt << endl;
	    }
	  HeadMap& hm = (*pi).second;
	  /* we are using collected counts for p(u|t) */
	  float hposprob = 1;
	  float chposprob = 1;
	  /* if we have reached a preterminal, then termInt == posInt
	     and p(posInt|termInt) == 1 */
	  if( itermInt != posInt)
	    {
	      hposprob = meProb(posInt, h, UCALC); 
	      if(printDebug() > 16)
		{
		  prDp();
		  cerr <<  "p(pos) = " <<  hposprob << endl;
		}
	    }
	  h->preTerm = posInt;
	  HeadIter hi = hm.begin(); 
	  for( ;hi != hm.end();hi++) // step through heads
	    {
	      const Wrd& subhw = (*hi).first;
	      int wrdInt = subhw.toInt();
	      string subh = subhw.lexeme();
	      if(printDebug() > 16)
		{
		  prDp();
		  cerr << "consider head(" << *itm << ") = " << subh << endl;
		}
	      float hprob;
	      if(wrdInt < 0)
		{
		  hprob = psutt(&subhw,posInt);
		}
	      else
		{
		  hprob = pCapgt(&subhw,posInt); 
		  hprob *= (1 - pHugt(posInt)); 
		  float hprob2 = meHeadProb(wrdInt, h);
		  hprob *= hprob2;
		}
	      if(printDebug() > 16)
		{
		  prDp();
		  cerr << "p(hd) = "<< hprob << endl;
		}
	      float hhprob = (hposprob * hprob);
	      double nextP = hhprob; 
	      h->hd = &subhw;

	      const AnswerTreePair&  
		atp2 = bestParseGivenHead(posInt,subhw,itm,h,(*hi).second); 
	      double bestProbGHead = atp2.first;  
	      AnswerTree* bestTreeGHead = atp2.second;

	      nextP *= bestProbGHead;
	      if(nextP > bestP)
		{
		  //cerr << "setting bestP to " << nextP << endl;
		  bestP = nextP;
		  bestW = subh;
		  bestPos = posInt;
		  bestAT = bestTreeGHead;
		  atp.first = nextP;
		  atp.second = bestAT;
		}
	      //assert(bestP > 0);
	      //  atp.first = bestP;
	      if(printDebug() > 10)
	      {
		prDp();
		int subfv[MAXNUMFS];
		getHt(h, subfv);
		CntxArray ca(subfv);
		cerr << "Bestp for " << *itm << " = " << bestP
		  << " using " << bestW << " " << bestPos << " " << ca 
		  << endl; //<< " " << atp.second << endl;
	      }
	      //atp.second = bestAT;
	    }
	}
    }
    if (printDebug() > 20)
    {
      prDp();
      cerr << "Tag " << ffTag << ", p=" << ffProb << "; bestP=" << bestP 
	   << endl;
    }
    if ((ffProb * bestP) > bestFfP)
    {
      bestFfP = ffProb * bestP;
      bestFf = ffTag;
      bestFfAT = bestAT;
    }
#ifdef INTEGRATED_FUNCTAG
  }
#endif
  return AnswerTreePair(bestFfP, bestFfAT);
}

AnswerTreePair  // when I memoize again this should be a ref
MeChart::
bestFuncGivenHead(int posInt, const Wrd& wd, Item* itm,
		   FullHist* h, ItmGHeadInfo& ighInfo)
{
  return bestParseGivenHead(posInt, wd, itm, h, ighInfo);

  if(printDebug() > 10)
    {
      prDp();
      cerr << "bestFuncGivenHead(" << *itm << ", ...)" << endl;
    }

  int bestFf = -1;
  double bestFfP = 0;
  AnswerTree* bestFfAT = NULL;

#ifdef INTEGRATED_FUNCTAG
  // calculate the possible function tags
  vector<FunctionTagPair> funcTags = tagsForNode(INTEGRATED_FUNCTAG, itm, h, true);

  vector<FunctionTagPair>::iterator fti = funcTags.begin();
  /* Cycle through the possible func tags; for each, set the 
     fullhist's functag, then go through the usual spiel, recursing
     downward.  If it was the best we've seen, keep it. */
  for (; fti != funcTags.end(); ++fti)
  {
    double ffProb = fti->first;
    int ffTag = fti->second;
    h->funcTag[INTEGRATED_FUNCTAG] = ffTag;
#else
    double ffProb = 0.0;
    int ffTag = 0;
#endif

    const AnswerTreePair& 
      atp2 = bestParseGivenHead(posInt, wd, itm, h, ighInfo);

    double atProb = atp2.first;
    AnswerTree* at = atp2.second;

    if (ffProb * atProb > bestFfP) {
      bestFfP = ffProb * atProb;
      bestFf = ffTag;
      bestFfAT = at;
    }
#ifdef INTEGRATED_FUNCTAG
  }
#endif

  return AnswerTreePair(bestFfP, bestFfAT);
}


int
firstPassE(Edge* e)
{
  if(e->pred()) return 1;
  if(e->item()->term()->terminal_p()) return 1;
  return 0;
}
    
AnswerTreePair   // when I re-memoise this should be a &
MeChart::
bestParseGivenHead(int posInt, const Wrd& wd, Item* itm,
		   FullHist* h, ItmGHeadInfo& ighInfo)
{
#ifdef INTEGRATED_FUNCTAG2
  if(!itm->term()->terminal_p())
  {
    int whichFunc = INTEGRATED_FUNCTAG2;
    int basedist = GRCALC;

    double func2Prob = meProb (NOFUNC_NUM, h, basedist + whichFunc);
    
    const AnswerTreePair&  
      atp2 = bestParseGivenHead2(posInt,wd,itm,h,ighInfo); 

    double bestFunc2Prob = func2Prob * atp2.first;
    AnswerTree* bestAT = atp2.second;
    int bestTag = NOFUNC_NUM;

    for (intsetiter it = FuncInfo::functionalTags[whichFunc].begin();
	it != FuncInfo::functionalTags[whichFunc].end();
	++it)
    {
      int tagval = *it;
      func2Prob = meProb (tagval, h, basedist + whichFunc);
      h->setFunctionTag(whichFunc, *it);
      const AnswerTreePair&
	atp3 = bestParseGivenHead2(posInt,wd,itm,h,ighInfo); 
      if (func2Prob * atp3.first > bestFunc2Prob)
      {
	bestFunc2Prob = func2Prob * atp3.first;
	bestAT = atp3.second;
	bestTag = tagval;
      }
    }
    if (bestAT && bestAT->funcTag[whichFunc] != bestTag)
      bestAT->funcTag[whichFunc] = bestTag;
    return AnswerTreePair(bestFunc2Prob, bestAT);
  }
  else
#endif
  {
    return bestParseGivenHead2(posInt,wd,itm,h,ighInfo); 
  }
}

AnswerTreePair&
MeChart::
bestParseGivenHead2(int posInt, const Wrd& wd, Item* itm,
		   FullHist* h, ItmGHeadInfo& ighInfo)
{
  EdgeSet& es = ighInfo.first;  // given a head, these are the possible parses
  AnswerTreeMap&  atm = ighInfo.second;
  int subfVals[MAXNUMFS];
  
  AnswerTreePair& atp = recordedBPGH(itm, atm, h); // memo
  if(atp.first >= 0)
    {
      if(printDebug() > 19)
	{
	  int subfv[MAXNUMFS];
	  getHt(h, subfv);
	  CntxArray ca(subfv);
	  prDp();
	  cerr << "bpknown for " << posInt << ", " << wd
	       << ", " << *itm << ") : " << atp.first << " " << ca <<endl;
	}
      return atp; // return memoized result if possible
    }

  if(itm->term()->terminal_p())
    {
      atp.first = 1;
      atp.second = NULL;
      if (printDebug() > 20)
      {
	prDp();
	cerr << "bpgh bailing because " << *itm << " is a terminal" << endl;
      }
      return atp;
    }
  atp.first = 0;
  if(printDebug() > 10)
    {
      prDp();
      cerr << "bestParseGivenHead(" << posInt << ", " << wd
	   << ", " << *itm  << ")" << endl;
    }
  double bestP = 0;
  AnswerTree* bestAT = NULL;
  EdgeSetIter ei = es.begin();
  /* Two passes to handle unaries properly.  ... */
  for(int passNum = 0 ; passNum < 2 ; passNum++)
    {
      ei = es.begin();
  for( ; ei != es.end() ; ei++) // step through possible parses (edges)
    {
      Edge* e = *ei;
      if (printDebug() > 60) 
      {
	prDp();
	cerr << "Considering edge " << *e << endl;
      }
      int fp = firstPassE(e);
      if(fp && passNum == 1) continue;
      else if(!fp && passNum == 0) continue;
      if(!sufficiently_likely(e))
	{
	  continue;
	}

      float edgePg = 1;
      int finish = e->loc();

      int effVal = effEnd(finish);
      if(itm->term()->name() == "S1") edgePg = 1;
      else if(effVal == 1)
	edgePg = endFactor;
      else if(effVal == 0) edgePg = midFactor;

      h->e = e;
      float prob = meRuleProb(e,h);
      prob *= 1.1; //encourage constits;  //???
      AnswerTree*  nat = new AnswerTree(e);
#ifdef INTEGRATED_FUNCTAG
      nat->funcTag[INTEGRATED_FUNCTAG] = h->funcTag[INTEGRATED_FUNCTAG];
#endif
#ifdef INTEGRATED_FUNCTAG2
      nat->funcTag[INTEGRATED_FUNCTAG2] = h->funcTag[INTEGRATED_FUNCTAG2];
#endif

      int headPos = h->hpos; //set during meRuleProb
      double nextP = prob * edgePg;
      Item* sitm;
      LeftRightGotIter gi(e);  //LRGI is just the kids of the edge
      int pos = 0;
      depth++;
      h = h->extendByEdge(e);
      while( gi.next(sitm) )  // sitm iterates through the kids of e
	{
	  //cerr << "Looking at " << *sitm << endl;
	  if(sitm->term() == Term::stopTerm)
	    {
	      pos++;
	      h = h->extendBySubConstit(); 
	      continue;
	    }
	  /* If sitm refers to the kid that contains the head of this
	     node, then we recurse directly into bPGH.  Otherwise
	     we recurse with bestParse. */
	  if(pos == headPos)
	    {
	      h->preTerm = posInt; 
	      h->hd = &wd;
	      ItmGHeadInfo& ighi = sitm->posAndheads()[posInt][wd]; 

	      AnswerTreePair
		atp = bestFuncGivenHead(posInt,wd,sitm,h,ighi); // ***
	      nextP *= atp.first;
	      nat->extend(atp.second);
	    }
	  else
	    {
	      AnswerTreePair atp = bestParse(sitm, h); // ***
	      nextP *= atp.first;
	      nat->extend(atp.second);
	    }
	  if(printDebug() > 39)
	    {
	      prDp();
	      cerr << "FullHist from " << *h;
	    }
	  h = h->extendBySubConstit(); 
	  if(printDebug() > 39)
	    cerr << " -> " << *h << endl;
	  pos++;
	}
      if(printDebug() > 20)
	{
	  prDp();
	  // The probability of this expansion given the head is...
	  cerr << "P(" << *e << " | " << wd << " ) = " << nextP;
	}
      depth--;
      if(nextP <= bestP)
	{
	  if(printDebug() > 20) cerr << endl;
	  delete nat;
	  nat = NULL;
	}
      else
	{
	  if(bestAT) delete bestAT;
	  bestAT = nat;
	  bestP = nextP;
	  atp.first = bestP;
	  atp.second = bestAT;
	  // New best ??
	  if(printDebug() > 20) cerr << " NB" << endl;
	}
      h->retractByEdge(); 
    }
    }
  //assert(bestP > 0);
  //atp.first = bestP;
  //atp.second = bestAT;
  if(printDebug() > 10)
    {
      int subfv[MAXNUMFS];
      getHt(h, subfv);
      CntxArray ca(subfv);
      prDp();
      cerr << "Bestpgh for " << *itm << ", " << wd << " = " << bestP
	   << " " << ca << endl;//<< " " << atp.second << endl;
    }
  return atp;
}

void
MeChart::
fillInHeads()
{
  // iterate j over possible lengths
  for (int j = 0 ; j <  wrd_count_ ; j++)
    {
      // iterate i over possible start locations
      for (int i = 0 ; i < wrd_count_ - j ; i++)
	{
	  list<Item*> doover;
	  Item* itm;
	  /* Iterate itmitr over all items in cell j,i (starts at i,
	     length j).  Each cell contains three types of items:
	       * Terminals.  Their head is just the corresponding word.
	       * Non-unaries.  Since all smaller constits are already
	           handled, their children already have heads filled in.
	       * Unaries.  We know nothing about the order they are in,
	           so we have to cludge a bit to make sure we have all
		   possible heads recorded for all unaries.
	   */
	  list<Item*>::iterator itmitr =regs[j][i].begin();
	  for( ; itmitr != regs[j][i].end() ; itmitr++)
	    {
	      itm = *itmitr;
//	      if(!sufficiently_likely(itm)) continue;
	      const Term* trm = itm->term();
	      int trmInt = trm->toInt();
	      if(itm->start() == i && itm->finish() == i+j+1) //???always true?
		{
		  if(trm->terminal_p())
		    {
		      if (printDebug() > 60) 
		      {
			prDp();
			cerr << "Recording head for " << *itm << endl;
		      }
		      // this is where posAndheads gets populated for terminals
		      HeadMap& hm = itm->posAndheads()[trmInt]; 
                      hm[*itm->word()];
		      continue;
		    }
		  else
		    doover.push_back(itm);
		}
	      headsFromEdges(itm);
	    }
	  bool cont = true;
	  while(cont)
	    {
	      cont = false;
	      list<Item*>::iterator lii = doover.begin();
	      for( ; lii != doover.end() ; lii++)
		cont = cont || headsFromEdges(*lii);
	    }
	}
    }
}

bool
MeChart::
headsFromEdges(Item* itm)
{
  bool ans = false;
  Edge* e;
  if (printDebug() > 60) 
  {
    prDp();
    cerr << "Calculating heads for " << *itm << endl;
  }
  /* Iterate eli over all the 'ineed' edges, i.e. those edges that 
     correspond to itm. 
   */
  list<Edge*>::iterator eli = itm->ineed().begin();
  for( ; eli != itm->ineed().end() ; eli++)
    {
      e = *eli;
      if(!sufficiently_likely(e)) continue;
      Item* ehd = headItem(e);
      if (printDebug() > 60) 
      {
        prDp();
        cerr << "Edge's head item is " << *ehd 
	  << " and has " << ehd->posAndheads().size() << " poss heads" << endl;
      }
      /* Iterate epi over all the [possible POSses (and corresponding 
	 Headmaps) of heads] of [the child of me that contains my head]. 
       */
      PosIter epi = ehd->posAndheads().begin();
      for( ; epi != ehd->posAndheads().end() ; epi++ )
	{
	  int posInt = (*epi).first;
	  
	  if(itm->posAndheads().find(posInt) == itm->posAndheads().end())
	    ans = true;
	  /* this is where posAndheads gets populated for nonterms: this 
	     creates an entry for posInt in posAndheads if one doesn't
	     already exist.
	   */
	  HeadMap& ihm = itm->posAndheads()[posInt];
	  HeadMap& ehm = (*epi).second;
	  /* Iterate ehi over all head-headinfo pairs corresponding to the
	     current POS.
	   */
	  HeadIter ehi = ehm.begin();
	  for( ; ehi != ehm.end() ; ehi++ )
	    {
	      const Wrd& hd = (*ehi).first;
	      if(ihm.find(hd) == ihm.end())
		{
		  if(printDebug() > 16)
		    {
		      prDp();
		      cerr << "attach hd " << *itm << " " << hd << endl;
		    }
		  ans = true;
		}
	      /* This creates an entry for the current guess at a head
		 (hd), if it doesn't already exist */
	      EdgeSet& se = ihm[hd].first;
	      //cerr << "Insrting " << *e << " for itm = " << *itm << endl;
	      se.insert(e);
	    }
	}
    }
  // Seems to return true iff it had to create an entry for this node's head
  return ans;
}


Item *
MeChart::
headItem(Edge* edge)
{
  if(!sufficiently_likely(edge)) return NULL;
  GotIter gotiter(edge);
  Item* ans;
  Item* next;
  while(gotiter.next(next))  //the head will be the the last thing in gotiter;
    ans = next;
  return next;  
}
     
AnswerTreePair& 
MeChart::
recordedBPGH(Item* itm, AnswerTreeMap& atm, FullHist* h)
{
  int subfv[MAXNUMFS];
  int i;
  for(i = 0 ; i < MAXNUMFS ; i++) subfv[i] = -1;

  if(!itm->term()->terminal_p())
    {
      getHt(h, subfv);
    }
  CntxArray ca(subfv);
  ca.d[CntxArray::sz-1] = -1; //set i field to don't care. 
  return atpFind(ca, atm);
}

AnswerTreePair& 
MeChart::
recordedBP(Item* itm, FullHist* h)
{
  //cerr << "rbp " << *itm << " " << *h << endl;
  int subfv[MAXNUMFS];
  getHt(h, subfv);
  CntxArray ca(subfv);
  return itm->stored(ca);
}

float
MeChart::
meHeadProb(int wInt, FullHist* h)
{
  float ans = meProb(wInt, h, HCALC);
  return ans;
}

float
MeChart::
meRuleProb(Edge* edge, FullHist* h)
{
  if(printDebug() > 30)
    {
      prDp();
      cerr << "In meruleprob " << *h << " " << *edge
	   << " " << edge->headPos() <<endl;
    }
  int i;
  int hpos = edge->headPos(); 
  h->hpos = hpos;
  LeftRightGotIter gi(edge);
  globalGi = &gi;
  Item* got;
  float ans = 1;
  for(i=0 ;  ; i++ )
    {
      if(i >= gi.size()) break;
      got = gi.index(i);
      h->pos = i;
      int cVal = got->term()->toInt();
      int whichInt = LCALC;
      if(h->pos == hpos) whichInt = MCALC;
      else if(h->pos > hpos) whichInt = RCALC;
      ans *= meProb(cVal, h, whichInt);
      if(ans == 0) break;
    }
  if(printDebug() > 30)
    {
      prDp();
      cerr << "merp = " << ans << endl;
    }
  globalGi = NULL;
  return ans;
}

/** Calculates the probability of a certain value, given a tree,
  * according to a certain distribution.
  * @param cVal the target value
  * @param h the 'current node'
  * @param whichInt the distribution to use
  * @return the probability of cVal given h according to the whichInt 
  *   distribution */
float
MeChart::
meProb(int cVal, FullHist* h, int whichInt)
{
  if(printDebug() > 38)
    {
      prDp();
      cerr << "meP" << whichInt << "(" << cVal << " | " << *h << ")" <<endl;
    }
  int subfVals[MAXNUMFS];
  FeatureTree* ginfo[MAXNUMFS];  
  ginfo[0] = FeatureTree::roots(whichInt);
  float smoothedPs[MAXNUMFS];

  float ans = 1;

  for(int i = 1 ; i <= Feature::total[whichInt] ; i++)
    {
      ginfo[i] = NULL;
      Feature* feat = Feature::fromInt(i, whichInt); 
      /* e.g., g(rtlu) starts from where g(rtl) left off (after tl)*/
      int searchStartInd = feat->startPos;

      FeatureTree* strt = ginfo[searchStartInd];
      if(!strt)
	{
	  continue;
	}
      SubFeature* sf = SubFeature::fromInt(feat->subFeat, whichInt);
      int nfeatV = (*(sf->fun))(h);
      FeatureTree* histPt = strt->follow(nfeatV, feat->auxCnt); 
      ginfo[i] = histPt;
      if(i == 1)
	{
          assert (histPt);
	  smoothedPs[0] = 1;
	  Feat* f = histPt->getFeat(cVal);
	  if(!f)
	    {
	      if(printDebug() > 30)
		{
		  prDp();
		  cerr << "Zero p" << feat->name << " " << nfeatV << endl;
		}
	      return 0.0;
	    }
	  smoothedPs[1] = f->g();
	  if(printDebug() > 38)
	    {
	      prDp();
	      cerr << i << " " << nfeatV << " " << smoothedPs[1] << endl;
	    }
	  for(int j = 2; j <= Feature::total[whichInt] ; j++)
	    smoothedPs[j] = 0;
	  ans = smoothedPs[1];
	  continue;
	}
      if(!histPt)
	{
	  continue;
	}
      float estm = histPt->getCount() * smoothedPs[1];
      int b = bucket(estm);

      Feat* ft = histPt->getFeat(cVal);
      float unsmoothedVal;
      if(!ft) unsmoothedVal = 0;
      else unsmoothedVal = ft->g();
      float lam = Feature::getLambda(whichInt, i, b);
      float uspathprob = lam*unsmoothedVal;
      float osmoothedVal = smoothedPs[searchStartInd];
      //float osmoothedVal = smoothedPs[i-1]; //for deleted interp.
      float smpathprob = (1-lam)*osmoothedVal;
      float nsmoothedVal = uspathprob+smpathprob;
      if(printDebug() > 38)
	{
	  prDp();
	  cerr << i << " " << nfeatV << " "
	       << estm << " " << b <<" "<<unsmoothedVal << " " << lam << " " 
	       << nsmoothedVal <<  endl;
	}
      smoothedPs[i] = nsmoothedVal;
      ans *= (nsmoothedVal/osmoothedVal);
    }
  if(whichInt == HCALC) ans *= 600;
  if(printDebug() > 30)
    {
      prDp();
      cerr<<"p"<<whichInt<< "(" << cVal << "|" << *h << ") = " << ans << endl;
    }
  return ans;
}


void
MeChart::
getHt(FullHist* h, int* subfVals)
{
  int i;
  int whichTree = UCALC;
  //cerr << "getHt(" << *h << ", " << whichTree << ")" << endl;
  for(i = 1 ; i < MAXNUMFS ; i++) subfVals[i] = -1;
  for(i = 1 ; i <= Feature::total[whichTree] ; i++)
    {
      Feature* ft = Feature::fromInt(i, whichTree); 
      int sfInt = ft->subFeat;
      SubFeature* sf = SubFeature::fromInt(sfInt, whichTree);
      int val = (*sf->fun)(h);
      subfVals[sfInt] = val;
    }
  //cerr << "done getHt" << endl;
}
    
bool
sufficiently_likely(const Item* itm)
{
  double pout = itm->poutside();
  double pin = itm->prob();
  double factor = .0008;
  if((pout * pin) > factor) return true;  
  return false;
}

bool
sufficiently_likely(Edge* edge)
{
  Item* fp = edge->finishedParent();
  if(!fp) return false;
  if(!sufficiently_likely(fp)) return false;
  GotIter gotIter(edge);
  Item* got;
  while( gotIter.next(got) )
    {
      if(got->term() == Term::stopTerm) continue;
      if(!sufficiently_likely(got)) return false;
    }
  return true;
}

  /* the general rule for the order in which to process constits
     of the same length, since there may be rules of the form
     SBAR -> S, etc, is:
     z) terminals processed first
     a) S1 is processed last;
     b) SBAR is processed after S;
     bc) S ans SQ are processed after VP;
     c) otherwise in order of decreasing probability.
     */

bool
MeChart::
okDecendent(Item* chld, FullHist* h)
{
  //cerr << "okd1 " << *h << endl;
  FullHist* parHist = h->back;
  while(parHist)
    {
      Item* par = parHist->e->finishedParent();
      assert(par);
      if(par->start() != chld->start()) return true;
      if(par->finish() != chld->finish()) return true;
      const Term* ptrm  = par->term();
      if(ptrm->name() == "S1") return true;
      if(par == chld)
	{
	  if(printDebug() > 28)
	    cerr << "Rejecting " << *par << " over " << *chld << endl;
	  return false;
	}
      parHist = parHist->back;
    }
  error("never get here");
  return false;
}

vector<FunctionTagPair>
MeChart::
tagsForNode(int whichTag, Item* itm, FullHist* h, bool knownhead)
{
    vector<FunctionTagPair> retval;

    if(itm->term()->terminal_p())
    {
      retval.push_back(FunctionTagPair(1.0, NOFUNC_NUM));
      return retval;
    }

    int basedist = knownhead? GRHCALC: GRCALC;

    float funcProb = meProb (NOFUNC_NUM, h, basedist + whichTag);
    retval.push_back(FunctionTagPair(funcProb, NOFUNC_NUM));

    if (printDebug() > 30)
    {
      cerr << "----------------------------------------------" << endl;
      cerr << *this << endl;
      if (printDebug() > 50)
	cerr << "no tag : " << funcProb << endl;
    }

    for (intsetiter it = FuncInfo::functionalTags[whichTag].begin();
	 it != FuncInfo::functionalTags[whichTag].end();
	 ++it)
    {
      funcProb = meProb (*it, h, basedist + whichTag);
      if (printDebug() > 50)
	cerr << *it << " : " << funcProb << endl;
      if (funcProb > 0)
	retval.push_back(FunctionTagPair(funcProb, *it));
    }

    return retval;
}
