
#include "Bchart.h"
#include "headFinder.h"

Wrd*
Bchart::
add_word(int i, string wrdStr)
{
  Wrd* ans = &sentence_[i]; 
  assert(ans);
  if(ans->toInt() >= 0) return ans;
  char temp[128];
  string headL(toLower(wrdStr.c_str(),temp));
  int wint = wtoInt(headL); 
  ans->toInt() = wint;
  return ans;
}

bool
Bchart::
edgeFromIEdge(istream& str)
{
  int b, e;
  b = -1;
  string trmNm;
  readItem(str, b, e, trmNm);
  if(b == -1) return false;
  string tmp;
  str >> tmp;
  assert(tmp == "->");
  const Term* trm = Term::get(trmNm);
  assert(trm);
  //cerr << "eFE " << *trm << endl;
  /* If this is a terminal node, the rhs will be a word; otherwise it
     will be a rule expansion consisting of several Item s.
   */
  if(trm->terminal_p())
    {
      str >> tmp;
      Wrd* wrd = add_word(b, tmp); 
      assert(wrd);
      add_item(b, e, trmNm, wrd); 
      return true;
    }
  else
    {
      Item* lhs = add_item(b, e, trmNm, NULL); 
      Items subItems;
      subItems.push_back(stops[b]);
      for( ; str ; )
      {
        readItem(str, b, e, trmNm);
        Item* itm = add_item(b, e, trmNm, NULL);
        //cerr << "Saw item " << *itm << endl;
        subItems.push_back(itm);
        char nxt;
        str.get(nxt);
        nxt = str.peek();
        if(nxt == '\n') break;
      }
      subItems.push_back(stops[e]);
      Edge* edg = add_edge(lhs, subItems); 
      if (printDebug() > 5)
      {
	cerr << "Reading edge " << *edg << ": p=" << edg->prob() << endl;
      }
      //cerr << "endeFE " << *edg << endl;
      return true;
    }
}
  
void
Bchart::
readItem(istream& str, int& b, int& e, string& trm)
{
  string tmp;
  trm = "";
  string numStr = "";
  str >> tmp;
  //cerr << "rI " << tmp << endl;
  if(tmp == "->") return;
  int i;
  bool doneTerm = false;
  
  for(i = 0 ; ; i++)
    {
      char c1 = tmp[i];
      if(c1 == '(') doneTerm = true;
      else if(c1 == ',' && i != 0) break;
      else if(!doneTerm) trm += c1;
      else numStr += c1;
    }
  b = atoi(numStr.c_str());
  //cerr << "rI2 " << b << trm  << endl;
  str >> tmp;
  numStr = "";
  for(i = 0 ; ; i++)
    {
      char c1 = tmp[i];
      if(c1 == ')') break;
      else numStr += c1;
    }
  e = atoi(numStr.c_str());
}

Item*
Bchart::
add_item(int b, int e, string trmNm, Wrd* wrd)
{
  const Term* trm = Term::get(trmNm);
  Item* ans = in_chart(NULL, trm, b, e);
  if(wrd) ans->word() = wrd;
  if(ans) return ans;
  ans = new Item(trm, b, e);
  //cerr << "AI " << *ans << endl;
  ans->poutside() = 1;  // just make sure > cutoff;
  ans->prob() = 1;
  put_in_reg(ans);
  return ans;
}


Edge*
Bchart::
add_edge(Item* lhs, Items& rhs)
{
  int headPos = headPosFromItems(lhs, rhs);
  //cerr << "In add_edge " << *lhs << " "
    //   << rhs.size() << " " << headPos << endl;
  Items::iterator ii = rhs.begin();
  int pos = 0;
  for( ; ii != rhs.end() ; ii++)
    {
      if(pos == headPos) break;
      pos++;
    }
  Items::iterator ii2 = ii;
  ii2++;
  Edge  dummyedge(lhs->term());
  Edge* prevEdge = &dummyedge;
  /* This is some crazy stuff.  Apparently we cycle through bits of the edge,
     starting to left of head and working left, then going to right of
     head and working right.  For each, an Edge is newed (!) though it is
     later garbage collected (!!) by the alreadyPoped mechanism.
   */
  for( ; pos >= 0 ; pos--)
    {
      Item* item = (*ii);
      //cerr << "bne " << *item << endl;
      Edge*   newEdge = new Edge(*prevEdge, *item, 0);
      prevEdge  = newEdge;
      //cerr << "ae1 " << *item << " " << *newEdge << endl;
      alreadyPoped[alreadyPopedNum++] = newEdge;  //so it will be gced.;
      if(pos > 0) ii--;
    }
  for( ; ii2 != rhs.end() ; ii2++)
    {
      Item* item = (*ii2);
      Edge*   newEdge = new Edge(*prevEdge, *item, 1);
      prevEdge  = newEdge;
      //cerr << "ae2 " << *item << " " << *newEdge << endl;
      alreadyPoped[alreadyPopedNum++] = newEdge;  //so it will be gced.;
    }
  prevEdge->setFinishedParent( lhs );    
  /* Since the chart itself only stores Items, the Edges in the chart are
     stored in the 'ineed' variable in the corresponding Item.
   */
  lhs->ineed().push_back(prevEdge);
  return prevEdge;
}

int
Bchart::
headPosFromItems(Item* lhs, Items& rhs)
{
  int   ansPriority = 10;
  string lhsString(lhs->term()->name());
  if(lhsString == "") lhsString = "S1";
  int   pos = -1;
  int   ans = -1;

  Items::iterator ii = rhs.begin();
  Item   *subi;
  for( ; ii != rhs.end() ; ii++ )
    {
      subi = (*ii);
      pos++;
      string rhsString(subi->term()->name());
      if(subi->term() == Term::stopTerm) continue;
      int nextPriority = headPriority(lhsString, rhsString, ansPriority);
      if(nextPriority <= ansPriority)
	{
	  ans = pos;
	  ansPriority = nextPriority;
	}
    }
  return ans;
}
