#include "ParseStats.h"
#include <string>
#include "utils.h"

using namespace std;

istream&
operator >>( istream& is, ParseStats& ps )
{
  if( ps.totalMap != !! ps.totalPrecision != 0 )
    error( "Reading into non-empty parse stats." );
  ps.readInput( is );
  return is;
}

void
ParseStats::
readInput( istream& is )
{
  string bracket;
  is >> bracket;
  if( !is ) return;
  if( bracket != "<" ) error( "No open bracket for ParseStats" );
  is >> totalMap;
  is >> obviousMap;
  is >> correctMap;
  is >> totalPrecision;
  is >> obviousPrecision;
  is >> correctPrecision;
  is >> totalExact;
  is >> ignoredExact;
  is >> exactMap;
  is >> exactPrecision;
  is >> otherWrong;
  is >> bracket;
  if ( bracket == ">" ) 
    return;
  complementPrecision = atoi(bracket.c_str());
  for (int func = 0; func < 4; ++func)
  {
    is >> bracket;
    if ( bracket == ">" ) 
      return;
    funcTagPrecisionA[func] = atoi(bracket.c_str());
    is >> bracket;
    if ( bracket == ">" ) 
      return;
    funcTagPrecisionB[func] = atoi(bracket.c_str());
    is >> funcTagPrecisionC[func];
    is >> funcTagPrecisionD[func];
    is >> funcTagPrecisionE[func];
  }
  is >> bracket;
  if( bracket != ">" ) error( "No close bracket for ParseStats" );
}

ParseStats&
ParseStats::
operator+= ( const ParseStats& src )
{
  totalMap += src.totalMap;
  obviousMap += src.obviousMap;
  correctMap += src.correctMap;
  totalPrecision += src.totalPrecision;
  obviousPrecision += src.obviousPrecision;
  correctPrecision += src.correctPrecision;
  totalExact += src.totalExact;
  ignoredExact += src.ignoredExact;
  exactMap += src.exactMap;
  exactPrecision += src.exactPrecision;
  otherWrong += src.otherWrong;
  complementPrecision += src.complementPrecision;
  for (int func = 0; func < 4; ++func)
  {
    funcTagPrecisionA[func] += src.funcTagPrecisionA[func];
    funcTagPrecisionB[func] += src.funcTagPrecisionB[func];
    funcTagPrecisionC[func] += src.funcTagPrecisionC[func];
    funcTagPrecisionD[func] += src.funcTagPrecisionD[func];
    funcTagPrecisionE[func] += src.funcTagPrecisionE[func];
  }
  return *this;
}

ostream&
operator <<( ostream& os, const ParseStats& ps )
{
  os << "< " << ps.totalMap << " " << ps.obviousMap << " "
     << ps.correctMap << " " << ps.totalPrecision << " "
     << ps.obviousPrecision << " " << ps.correctPrecision
     << " " << ps.totalExact
     << " " << ps.ignoredExact
     << " " << ps.exactMap
     << " " << ps.exactPrecision
     << " " << ps.otherWrong 
     << " " << ps.complementPrecision;
     for (int func = 0; func < 4; ++func)
     {
       os << " " << ps.funcTagPrecisionA[func]
          << " " << ps.funcTagPrecisionB[func]
          << " " << ps.funcTagPrecisionC[func]
          << " " << ps.funcTagPrecisionD[func]
          << " " << ps.funcTagPrecisionE[func];
     }
  os << " >\n";
  return os;
}
