/*
 * Copyright 1997, Brown University, Providence, RI.
 * 
 *                         All Rights Reserved
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose other than its incorporation into a
 * commercial product is hereby granted without fee, provided that the
 * above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Brown University not be used in
 * advertising or publicity pertaining to distribution of the software
 * without specific, written prior permission.
 * 
 * BROWN UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR ANY
 * PARTICULAR PURPOSE.  IN NO EVENT SHALL BROWN UNIVERSITY BE LIABLE FOR
 * ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#ifndef PST_H
#define PST_H
#include <list>
#include "WordInfo.h"
#include "Phegt.h"

class Pst
{
 public:
  Pst(std::string& path);
  float&  pHugt(int i) { return pHugt_[i]; }
  float&  pHhypgt(int i) { return pHhypgt_[i]; }
  float&  pHcapgt(int i) { return pHcapgt_[i]; }
  float&  pHtgnt(int i, int j);
  float   pHegt(std::string& es, int t);
  float   pegt(std::string& sh, int t);
  void    readTermProbs(std::string& pTString, std::string& pUString);
  float           pHst( const std::string& s, int trm) const
    {
      const Phsgt* sh = useSubTermC( s, trm ); 
      return( sh ? sh->prob : 0 );
    }

  int             ch(std::string& s)
    {
      const WordInfo* sh = useHeadC( s ); 
      return( sh ? sh->c() : 0 );
    }  
  static int& classNum(int wnum, int pos);
  static int classNm(int wnum, int pos);
  std::list<double> wordPlistConstruct(const std::string& head, int word_num);
  static int  hSize() { return hSize_; }
  static const WordInfo*  get(const std::string& wd)
    { return useHeadC(wd); }
  static const std::string& fromInt(int i);
  static void classReader(const std::string& path);

  static bool isInitialized() { return initialized; }

  protected:
  void  readPhsgt(std::string& path);
  const Phsgt*         useSubTermC(const std::string& head,
				   const int trm) const;
  static const WordInfo*      useHeadC(const std::string& head);
  double  pstt(std::string& shU, int t, int word_num);
  double  psutt(const std::string& shU, int t, int word_num);
  double  psktt(const std::string& shU, int t, int word_num);
  double  pCapgt(const std::string& shU, int t, int word_num);
  double  pHypgt(const std::string& shU, int t);

  int   egtSize_;
  static int   numClasses_[50];
  static int   classOffsets_[50];
  int   nttCounts_[100];  //???
  float pHugt_[100];
  float pHcapgt_[100];
  float pHhypgt_[100];
  float pHtgnt_[60][30]; 
  Phegt *pHegt_;
  static int       hSize_;
  static WordInfo* h_;

  static bool initialized;
  };

#endif /* ! PST_H */



