#include "Rule.h"

#include <fstream>
#include "Term.h"

using namespace std;

int Rule::nextId = 0;
map<int, map<vector<int>, const Rule*> > Rule::rules;

string Rule::toString() const
{
  string retval = Term::fromInt(parent)->name();
  for (vector<int>::const_iterator it = children.begin();
       it != children.end();
       ++it)
  {
    int child = *it;
    retval += ' ';
    retval += Term::fromInt(child)->name();
  }
  return retval;
}

void Rule::init(string path)
{
  path += "rules.txt";
  ifstream is(path.c_str());
  init(is);
}

void Rule::init(istream& is)
{
  while (is)
  {
    string line;
    bool success = getline(is, line);
    if (!success) break;
    int space = line.find(' ');
    string item(line, 0, space);
    int parent = Term::get(item)->toInt();
    vector<int> children;
    while (1)
    {
      line.erase(0, space + 1);
      space = line.find(' ');
      if (space == string::npos && line.length() == 0) break;
      item = string(line, 0, space);
      int child = Term::get(item)->toInt();
      children.push_back(child);

      if (space == string::npos) break;
    }



    /*
    strstream linestr;
    linestr << line;
    string item;
    linestr >> item;
    int parent = Term::get(item)->toInt();
    vector<int> children;
    while (linestr)
    {

      linestr >> item;
      int child = Term::get(item)->toInt();
      children.push_back(child);
    }
    */
    registerRule(parent, children);
  }
}

void Rule::dumpRules(ostream& os)
{
  for (ParChildMap::const_iterator it = rules.begin();
       it != rules.end();
       ++it)
  {
    const int& par = (*it).first;
    const ChildMap& kids = (*it).second;
    for (ChildMap::const_iterator it = kids.begin();
	 it != kids.end();
	 ++it)
    {
      const Rule* rule = (*it).second;
      os << rule->toString() << endl;
    }
  }
}
