#ifndef _RULE_H_
#define _RULE_H_

#include <vector>
#include <map>
#include <iostream>
#include <string>

using namespace std;

class Rule 
{
  public:
    Rule (int p, const vector<int>& c) : parent(p), children(c) 
      { id = nextId++; rules[p][c] = this; }

    typedef map<vector<int>, const Rule*> ChildMap;
    typedef map<int, ChildMap> ParChildMap;
    static ParChildMap rules;

    int toInt() const { return id; }
    string toString() const;

    static void init(string path);
    static void init(istream& is);
    static void dumpRules(ostream& os);

    static const Rule* getRule(int p, const vector<int>& c)
    { if (rules.count(p) && rules[p].count(c)) return rules[p][c]; 
      else return NULL;
    }

    static const Rule* registerRule(int p, const vector<int>& c)
    { 
      const Rule* rule = getRule(p, c);
      if (rule) return rule;
      else return new Rule(p, c);
    }

  private:
    static int nextId;

    int id;
    int parent;
    vector<int> children;
};

#endif
