
#include "SentRep.h"
#include <iostream>
#include "utils.h"

const string	SentRep::sentence_closer_ = ".";

SentRep::
SentRep()
: length_( 0 )
{
}

SentRep::
SentRep( istream& istr )
: length_( 0 )
{
  sentRepCreator(istr, SentRep::STD);
}

SentRep::
SentRep(istream& istr, SentRep::SentLayout layout)
: length_( 0 )
{
  sentRepCreator(istr, layout);
}

SentRep::
SentRep(list<string> wtList) 
       : length_( 0 )
{
  int len =  wtList.size();
  length_ = len;
  //assert(len < MAXSENTLEN);
  list<string>::iterator wi = wtList.begin();
  for(int i = 0 ; i < len ; i++)
    {
      vecRef(words_,i).lexeme_ = *wi;
      vecRef(words_,i).loc_ = i;
      wordSet[lemmatise(*wi)] = i;
      wi++;
    }
}

SentRep::
SentRep(list<pair<string, string> >& wtList) 
       : length_( 0 )
{
  int len =  wtList.size();
  length_ = len;
  //assert(len < MAXSENTLEN);
  list<pair<string, string> >::const_iterator wi = wtList.begin();
  for(int i = 0 ; i < len ; i++)
    {
      vecRef(words_,i).lexeme_ = (*wi).second;
      vecRef(words_,i).loc_ = i;
      wordSet[lemmatise((*wi).second)] = i;
      wi++;
    }
}

void
SentRep::
sentRepCreator( istream& istr, SentRep::SentLayout layout)
{
    // SGML layout introduces sentence with <s> and ends it with </s>.
    if( layout == SGML )
    {
	for(; istr;)
	{
	    Wrd	temp;
	    istr >> temp;
	    if( temp.lexeme() == "<s>" ) break;
	    else if( temp.lexeme() == "</s>" )
		warn( "found sentence end before intro; " );
	}
    }
    for( ; ; )
    {
       //assert(length_ < MAXSENTLEN);

	istr >> vecRef(words_, length_ );
        words_[length_].loc_ = length_;
	// EOF or error.  Note that because we are looking ahead, istr
	// can be finished while we still have words left to output;
	if( !istr && words_[length_].lexeme() == "" )
	{
	    // For asci text be somewhat looser;
	    if(layout != ASCI) length_ = 0;
	    return;
	}
	// end of sentence
	if(layout == STD && isSentCloser(words_[length_++]))
	    break;
	else if( layout == SGML && words_[ length_++ ].lexeme() == "</s>" )
	{
	    length_--;
	    break;
	}
      }
}
/*
SentRep::
SentRep( TokStrm& istr, SentRep::SentLayout layout )
: length_( 0 )
{
    // SGML layout introduces sentence with <s> and ends it with </s>.
    if( layout == SGML || layout == ASCI )
    {
	for(; istr;)
	{
	    Wrd	temp;
	    istr >> temp;
	    if( !strcmp( temp.lexeme(), "<s>" ) )
		break;
	    else if( !strcmp( temp.lexeme(), "</s>" ) )
		warn( "found sentence end before intro; " );
	}
    }
    for( ; ; )
    {
        //assert(length_ < MAXSENTLEN);

	istr >> vecRef(words_, length_ );
	// EOF or error.  Note that because we are looking ahead, istr
	// can be finished while we still have words left to output;
	if( !istr && strlen(words_[length_].lexeme()) == 0 )
	{
	    // For asci text be somewhat looser;
	    if(layout != ASCI) length_ = 0;
	    return;
	}
	// end of sentence
	if( layout == STD && isSentCloser(words_[ length_++ ] ) )
	    break;
	else if( layout != STD && !strcmp( words_[ length_++ ].lexeme(), "</s>" ) )
	{
	    length_--;
	    break;
	}
      }
}
*/


SentRep::
SentRep( const SentRep& sr )
: length_( sr.length_ )
{
    for( int i = 0; i < length_; i++ )
	vecRef(words_, i ) = sr.words_[ i ];
}


int
SentRep::
read( istream& istr )
{
    length_ = 0;
    for( ; ; )
    {
	istr >> vecRef(words_, length_);
	// EOF or error 
	if( !istr )
	{
	    length_ = 0;
	    return 1;
	}
	// end of sentence
	if( isSentCloser(words_[ length_++ ] ) )
	    break;
    }
    return 0;
}

int
SentRep::length() const
{
    return length_;
}

SentRep&
SentRep::
operator= (const SentRep& sr)
{
    length_ = sr.length_;
    for( int i = 0; i < length_; i++ )
	vecRef(words_, i ) = sr.words_[ i ];
    return *this;
}

int
SentRep::
operator== (const SentRep& sr) const
{
#if 0
    if( this == &sr )
	return true;
    if( length_ != sr.length_ )
	return false;
    for( int i = 0; i < length_; i++ )
    {
	if( !( vecRef(words_, i ) == sr.words_[ i ] ) )
	    return false;
    }
    return true;
#endif
    return false;
}

ostream&
operator<< (ostream& os, const SentRep& sr)
{
    for( int i = 0; i < sr.length_; i++ )
	os << sr.words_[ i ] << " ";
    return os;
}

bool
SentRep::
isSentCloser(const Wrd& wrd)
{
  const string& s = wrd.lexeme();
  if(s == "." || s == "?" || s == "!" ) return true;
  else return false;
}
					  
