
#ifndef SENTREP_H
#define SENTREP_H

#include <string>
#include <list>
#include <set>
#include <vector>
#include <map>
#include "Wrd.h"
#include "utils.h"

using namespace std;

//#define MAXSENTLEN 200

class SentRep
{
public:
    enum	SentLayout { STD, SGML, ASCI };

    SentRep();
    SentRep( istream& istr );
    SentRep( istream& istr, SentRep::SentLayout layout );
    //SentRep( TokStrm& istr, SentRep::SentLayout layout = STD );
    SentRep( const SentRep& sr );
    int	    	read( istream& istr );
    int	    	length() const;
    SentRep(list<string> wtList);
    SentRep(list<pair<string, string> >& wtList);
    SentRep&	operator= (const SentRep& sr );
    int	    	operator== ( const SentRep& sr ) const;
    const Wrd& operator[] ( int index ) const
	    	{   return words_[ index ];   }
    Wrd&       operator[] ( int index ) 
	    	{   return words_[ index ];   }
    bool contains (string word) const
      		{  return wordSet.count(word) > 0; }
    bool contains (string word, int st, int fn) const
      		{ if (wordSet.count(word))
		  { 
		    int pos = getByKey(wordSet, word);
		    return pos >= st && pos < fn;
		  }
		  else return false;
		}

    bool       isSentCloser(const Wrd& wrd);

friend ostream& operator<< (ostream& os, const SentRep& sr);
friend class SentRepIter;

  protected:
    void        sentRepCreator( istream& istr, SentRep::SentLayout layout);
    int		length_;
    vector<Wrd>	words_;
    map<string, int> wordSet;

    static const string sentence_closer_;
};

class SentRepIter
{
public:
    SentRepIter( const SentRep& sr )
    	: srhandle_( sr ), index_( 0 )
    	{}
    void reset()
	{
	    index_ = 0;
	}
    const Wrd * nextWrd()
	{
	    if( index_ < srhandle_.length_ )
	        return &( srhandle_.words_[ index_++ ] );
	    else
		return NULL;
	}
private:
    const SentRep& srhandle_;
    int	index_;
};
#endif /* ! SENTREP_H */
