
#include "Term.h"
#include "utils.h"

Term*  Term::array_[400];
map<string, Term*,less<string> > Term::termMap_;
int    Term::lastTagInt_ = 0;
int    Term::lastNTInt_ = 0;
Term*  Term::stopTerm;
Term*  Term::startTerm;
Term*  Term::rootTerm;
bool   Term::initialized = false;

Term::
Term()
: num_( -1 ), terminal_p_( 0 )
{}

Term::
Term(const string s, int terminal, int num, const string b)
: name_( s ), num_( num ), terminal_p_( terminal )
{
  if (b == "~")
    backoff_ = this;
  else
    backoff_ = get(b);
  assert (backoff_);
}

Term::
Term( const Term& src )
: name_( src.name() ), 
  //num_(src.toInt()),
  terminal_p_( src.terminal_p() )
{}

ostream&
operator<< ( ostream& os, const Term& t )
{
    os << t.name();
    return os;
}

int
Term::
operator== ( const Term& rhs ) const
{
    if( this == &rhs || name_ == rhs.name() )
	return 1;
    return 0;
}


void
Term::
init(string & prefix)
{
  string          fileName(prefix);
  fileName += "terms.txt";
  ifstream           stream(fileName.c_str(), ios::in);
  if (!stream)
    {
      cerr << "Can't open terms file " << fileName << endl;;
      return;
    }
  
  string          termName, backoffStr;
  int ind, n;
  n = 0;
  bool seenNTs = false;
  while (stream >> termName)
    {
      stream >> ind;
      stream >> backoffStr;
      Term* nextTerm = new Term(termName, ind, n, backoffStr);
      termMap_[nextTerm->name()] = nextTerm;
      if(termName == "STOP") Term::stopTerm = nextTerm;
      else if(termName == "G4") Term::startTerm = nextTerm;
      else if(termName == "S1") Term::rootTerm = nextTerm;
      array_[n] = nextTerm;
      if(!ind && !seenNTs)
	{
	  assert(n > 0);
	  lastTagInt_ = n-1;
	  seenNTs = true;
	}
      n++;
      assert(n < 400);
    }
  //assert(!ind); //Not valid now that -NONE- is at the end
  lastNTInt_ = n-1;
  //lastNTInt_ = n-4;  //??? hack to ignore G1 and G2 and G3;
  stream.close();

  initialized = true;
}

Const_Term_p
Term::
get(const string& nm)
{
  TermMap::iterator ti = termMap_.find(nm);
  if( ti == termMap_.end()) return NULL;
  return (*ti).second;
}

