
#ifndef TERM_H
#define TERM_H

#include <string>
#include <list>
#include <map>
#include <cassert>
#include <iostream>
#include <fstream>

using namespace std;

class Term;

typedef Term *		Term_p;
typedef const Term *	Const_Term_p;
typedef const Term     ConstTerm;

#define Terms list<ConstTerm*>
#define ConstTerms const list<ConstTerm*>
#define TermsIter list<Term*>::iterator
#define ConstTermsIter list<ConstTerm*>::const_iterator
typedef map<string, Term*, less<string> >  TermMap;

class Term 
{
public:
    Term();			// provided only for maps.
    Term( const string s, int terminal, int n, const string b );
    Term( const Term& src );
    int              toInt() const { return num_; }
    const string& name() const {  return name_; }
    const Term* getSupertag() const { return backoff_; }
    int	  terminal_p() const { return (terminal_p_ != 0); }
    friend ostream& operator<< ( ostream& os, const Term& t );
    friend ostream& operator>> ( istream& os, const Term& t );
    int		operator== (const Term& rhs ) const;
    bool   isPunc() const { return (terminal_p_ == 2) ? true : false ; }
    bool   openClass() const { return (terminal_p_ == 3) ? true : false ; }
    static Const_Term_p get(const string& getname);
    static void  init(string & prefix);
    static const Term* fromInt(int i) 
      { assert(i < 400); return array_[i]; }
    static int  lastTagInt() { return lastTagInt_; }
    static int  lastNTInt() { return lastNTInt_; }
    static Term* stopTerm;
    static Term* startTerm;
    static Term* rootTerm;

    static bool isInitialized() { return initialized; }

private:
    string* namePtr() { return (string*)&name_; }
    int    	terminal_p_;
    int		num_;
    const string name_;
    const Term* backoff_;
    static Term*  array_[400];
    static TermMap termMap_ ;
    static int    lastTagInt_;
    static int    lastNTInt_;
    static bool initialized;
};
  

#endif /* ! TERM_H */
