#include "TreeHist.h"
#include "InputTree.h"
#include "Term.h"
#include "WordInfo.h"
#include "Pst.h"
#include "Feat.h"

int 
TreeHist::getLabel() const
{
  const string& trmStr  = tree->term();
  const Term* trm = Term::get(trmStr);
  assert(trm);
  return trm->toInt();
}

int
TreeHist::getStart() const
{
  return tree->start();
}

int
TreeHist::getFinish() const
{
  return tree->finish();
}

int
TreeHist::getHeadPOS() const
{
  const string& trmStr  = tree->hTag();
  const Term* trm = Term::get(trmStr);
  assert(trm);
  assert(trm->terminal_p());
  assert(trm->toInt() != 47 /*STOP*/);
  return trm->toInt();
}

int
TreeHist::getHead() const
{
  char temp[128];
  string wrdStr(toLower(tree->head().c_str(), temp));
  const WordInfo* wi = Pst::get(wrdStr);
  if(!wi)
    {
      if(Feat::Usage == PARSE || Feat::Usage == OTHER) return -1;
      assert(wi);
    }
  int ans = wi->toInt();
  assert(ans >= 0);
  return ans;
}

string
TreeHist::getHeadString() const
{
  return tree->head();
}

int 
TreeHist::getFunctionTag(int i) const
{
  return tree->getFuncTag(i);
}

int
TreeHist::getNumChildren() const
{
  return getChildren().size();
}

int
TreeHist::getChildLabel (int targetPosition) const
{
  assert (targetPosition >= 0);
  myassert (targetPosition < getChildren().size(), targetPosition);

  int i;
  list<Hist*>::const_iterator it;

  for (i = 0, it = getChildren().begin();
       it != getChildren().end();
       ++i, ++it)
    if (i == targetPosition) return (*it)->getLabel();

  error("TreeHist::getChild: shouldn't get here");
  return -1;
}
