//#include <time.h>
#include <sys/resource.h>
//#include <fstream.h>
//#include <iostream.h>
//#include <unistd.h>
//#include <math.h>
//#include "Field.h"
//#include "GotIter.h"
//#include "Wrd.h"
#include "InputTree.h"
//#include "Bchart.h"
#include "ECArgs.h"
//#include "MeChart.h"
#include "headFinder.h"
#include "Term.h"
#include "FuncInfo.h"
//#include "Cluster.h"
//#include "Pst.h"
#include "CLRInfo.h"
#include <math.h>
#include "utils.h"

map <string, int> vbCounts;
map <string, int> vbWithSibCounts;
int vbTotal;
map <string, int> vbSibCounts;
int vbSibTotal;
map <pair<string, string>, int> vbSibPairs;

void observePair (const string& verb, const string& sibHead, int ct=1)
{
  vbSibTotal += ct;
  vbSibCounts[sibHead] += ct;
  vbWithSibCounts[verb] += ct;
  vbSibPairs[pair<string, string>(verb, sibHead)] += ct;
}

void observeFrame (const string& verb, const vector<string>& sibs, int ct=1)
{
  string sibstring;
  if (sibs.size())
  {
    sibstring = sibs[0];
    for (vector<string>::const_iterator it = sibs.begin()+1;
	 it != sibs.end();
	 ++it)
      sibstring += " " + *it;
  }
  vbSibPairs[pair<string, string>(verb, sibstring)] += ct;
}

void countVerbs (InputTree* tree)
{
//cerr << "Counting for " << tree->getLabel() << "/" << tree->getHead() << endl;
  if (string(tree->getLabel(), 0, 2) == "VB")
  {
    string verb = (tree->getHead());
  //cerr << "Found verb " << verb << endl;;

    InputTree* parent = tree->getParent();
    assert(parent);

    ++vbTotal;
    ++vbCounts[verb];

    vector<string> sibs;

    for (InputTreesIter it = parent->subTrees().begin();
	 it != parent->subTrees().end();
	 ++it)
    {
      InputTree* sib = *it;
      if (sib == tree) continue;
      //if (sib->getLabel() != "PP") continue;
      //if (sib->getLabel() != "PRT") continue;
      if (sib->getLabel() != "PRT" && sib->getLabel() != "PP") continue;
      string sibHead = sib->getHead();

      sibs.push_back(sibHead);
    }
    observeFrame(verb, sibs);
  }
  else for (InputTreesIter it = tree->subTrees().begin();
            it != tree->subTrees().end();
            ++it)
  {
    InputTree* child = *it;
    countVerbs(child);
  }
}

multimap <double, pair<string, string> > vbSibMerit;
void procVerbs()
{
    for (map<pair<string, string>, int>::const_iterator it = vbSibPairs.begin();
	 it != vbSibPairs.end();
	 ++it)
    {
      const pair<string, string>& vspair = (*it).first;
      const string& verb = (*it).first.first;
      const string& sib = (*it).first.second;
      const int& count = (*it).second;
      const double total = vbSibTotal;

      double expected = (vbWithSibCounts[verb] * vbSibCounts[sib])
			/ (total * total);
      double actual = count / (double)vbWithSibCounts[verb];
      double merit = actual * log(count);

      vbSibMerit.insert(pair<double, pair<string, string> >(merit, vspair));
    }
}

void dumpVerbs(ostream& os)
{
  os << "Total sibling pairs: " << vbSibTotal << endl;
    for (map<double, pair<string, string> >::const_iterator it 
						= vbSibMerit.begin();
	 it != vbSibMerit.end();
	 ++it)
    {
      const double& merit = (*it).first;
      const pair<string, string>& vspair = (*it).second;
      const string& verb = (*it).second.first;
      const string& sib = (*it).second.second;
      int count = vbSibPairs[vspair];

      os << count << "/" << vbWithSibCounts[verb]
        << "\t" << verb << " " << sib << ":"
	<< " V=" << vbWithSibCounts[verb] 
		<< "(" << vbWithSibCounts[verb]/(double)vbSibTotal << ")"
	<< " S=" << vbSibCounts[sib]  
		<< "(" << vbSibCounts[sib]/(double)vbSibTotal << ")"
	<< " S|V=" << count 
		<< "(" << count / (double)vbWithSibCounts[verb] << ")"
	<< endl;
    }

}

void dumpCounts(ostream& os)
{
  for (map<pair<string, string>, int>::const_iterator it = vbSibPairs.begin();
       it != vbSibPairs.end();
       ++it)
  {
    const string& verb = (*it).first.first;
    const string& sib = (*it).first.second;
    const int& count = (*it).second;
    os << (toLower(verb)) << "\t\t" << sib << "\t\t" << count << endl;
  }
}

int
main(int argc, char *argv[])
{
   ECArgs args( argc, argv );
   /* o = basic, but not debugging, output.
      l = length of sentence to be proceeds 0-40 is default
      n = work on each #'th line.
      D = data directory (for terms.txt et al)
      d = print out debugging info at level #
      W = use wwclasses
      R = use rwclasses
      t = report timings (requires o)
      s = maximum sleep time
      f = f# says multiply ctl2 counts by #
      p = p# use prepFactor #
      P = which types of prob models to use */

   // prevent core file creation;
   struct rlimit 	core_limits;
   core_limits.rlim_cur = 0;
   core_limits.rlim_max = 0;
   setrlimit( RLIMIT_CORE, &core_limits );

   cerr << "Starting countVerbs" << endl;
   //params.init( args );
	 
   //if( args.nargs() > 2 || args.nargs() == 0 )	// require path name 
   if( args.nargs() < 1)
     error( "Need data dir." );
   string datapath;
   string inputpath;
   if (args.isset('D')) {
     datapath = args.value('D');
     inputpath = args.arg( 0 );
   } else {
     inputpath = datapath = args.arg( 0 );
   }
   readHeadInfo( datapath );
   FuncInfo::init( datapath );
   Term::init( datapath );
   //InputTree::init( datapath );
   //Cluster::init( datapath );
   Pst pst( datapath );
   init_lemmatiser(datapath);

   vector<string> files;
   if (args.nargs() == 1)
     files.push_back(string("text.txt"));
   else 
   {
     for (int i = 1; i < args.nargs(); ++i)
       files.push_back(inputpath + args.arg(i));
   }

   int      sentenceCount = 0;  //counts all sentences so we can use 1/50;

   if (args.isset('P')) {
     for (int fnum=0; fnum < files.size(); ++fnum)
     {
       ifstream testSStream(files[fnum].c_str());
       if( !testSStream ) error( "No testSStream" );

       string verb;
       string sib;
       int count;

       while (testSStream)
       {
         testSStream >> verb >> sib >> count;
         observePair ((toLower(verb)), sib, count);
       }
     }
     procVerbs();
     dumpVerbs(cout);
   }
   else
   {
     for (int fnum = 0; fnum < files.size(); ++fnum)
     {
       ifstream testSStream(files[fnum].c_str());
       if( !testSStream ) error( "No testSstream" );

       for( ; testSStream ; )
       {
	 sentenceCount++;

	 //InputTree     correct;  
	 InputTree*    goldparse;
	 
	 //testSStream >> correct;
	 //goldparse = &correct;
	 goldparse = new InputTree();
	 testSStream >> *goldparse;
	 countVerbs(goldparse);
	 delete goldparse;
	 goldparse = NULL;

	 //cout << sentenceCount << ": " << CLRInfo::getSize() << " entries\n";
       }
     }

     cerr << "Dumping counts...." << endl;
     dumpCounts(cout);
   }

   return 0;
}
