#include <stdlib.h>
#include <math.h>
#include <iostream.h>
#include <stdiostream.h>
#include <fstream.h>
#include <iomanip.h>
#include <signal.h>
#include <assert.h>
#include "ECArgs.h"
#include "utils.h"
//#include <ngram.h>
#include "ParseStats.h"
#include <string>


// Local types


// Global objects
//
//		(or object wannabes)
//

#define MAXN 10


// Misc utility functions.
//

void
usage( char * program )
{
    cerr << program << " usage: " << program
      << " [-w Wordtype] [ <file_prefix> ]" << endl;
    exit( 1 );
}


double lambda_ngram;
double lambda_grammar;
int train_lambdas( ifstream& sentenceStream);
void compute_lambdas(double probs[][2], int num_sents, ofstream& ostrm);

int
  main(int argc, char *argv[])
{
   ECArgs args( argc, argv );
   // L says allow all sents for longest length;
    // l says find values for lower sentence lengths
   // n says that there will be n different measurements for each sent.
   if( args.nargs() > 1 )	// max one filename allowed.
     usage( *argv );
   
   string path = args.arg(0);

   cout.precision(5);
   int lens[3];
    if(args.isset('l'))
      {
	lens[0] = 12;
	lens[1] = 21;
	lens[2] = 31;
      }
    else
      {
	lens[0] = 17;
	lens[1] = 26;
	lens[2] = 41;
      }
   
   int         sOfLen[3];
   int         wOfLen[3];
   for(int i = 0 ; i < 3 ; i++)
     {
       sOfLen[i] = 0;
       wOfLen[i] = 0;
     }
   int         numMeasures = 2;

   int         cbStats[MAXN][3];
   int         j;
   for(i = 0 ; i < MAXN ; i++)
     for(j = 0 ; j < 3 ; j++)
       cbStats[i][j] = 0;

   ParseStats  pStatsAll[3][MAXN];
   if(args.isset('n'))
     numMeasures = atoi(args.value('n').c_str());
   assert( numMeasures <= MAXN);
   
   string fileString(path);
   fileString += "PStats.txt";
   string      parseStatsString( fileString );
   ifstream    parseStatsStream( parseStatsString.c_str() );
   if( !parseStatsStream ) error( "unable to open parseStats stream." );

   cout << "Av sent lens =\t";

   int totSents = 0;
   int totUnparsed = 0;

   while( parseStatsStream )
     {
       int len;
       parseStatsStream >> len;
       if(!parseStatsStream) break;
       if(len == 9999)
	 {
	   int tot;
	   parseStatsStream >> tot;
	   totSents += tot;
	   parseStatsStream >> tot;
	   totUnparsed += tot;
	   continue;
	 }
       
       if(!parseStatsStream) break;
       for(int m = 0 ; m < numMeasures ; m++)
	 {
	   ParseStats temp;
	   parseStatsStream >> temp;
	   int numxb = temp.totalMap - temp.correctMap;
	   cbStats[m][0] += numxb;
	   if(numxb == 0) cbStats[m][1]++;
	   if(numxb < 3) cbStats[m][2]++;
	   for(int ls = 0 ; ls < 3 ; ls++)
	     {
	       if(len < lens[ls] || args.isset('L'))
		 {
		   pStatsAll[ls][m] += temp;
		   wOfLen[ls] += len;
		   sOfLen[ls]++;
		 }
	     }
	 }
     }
   for(int l = 0 ; l < 3 ; l++)
     cout << lens[l]-1 << ": "<< (float)wOfLen[l]/(float)sOfLen[l] << "\t";
   cout << "\n";
   cout << "Tot sentences: " << totSents << "\t";
   cout << "Tot unparsed: " << totUnparsed << "\n";
   for(l = 2 ; l < 3 ; l++)
     {
       cout << "\nAt Length " << lens[l] << endl;
       for(int m = 0 ; m < numMeasures ; m++)
	 {
	   cout << "\nFor model " << m << endl;
	   double rec =
	     (double)pStatsAll[l][m].correctPrecision/pStatsAll[l][m].totalExact;
	   double prec =
	     (double)pStatsAll[l][m].correctPrecision
	       /pStatsAll[l][m].totalPrecision;
	   double ac =
	     (double)(pStatsAll[l][m].correctMap - pStatsAll[l][m].obviousMap)/
	       (pStatsAll[l][m].totalMap - pStatsAll[l][m].obviousMap);
	   double erec =
	     (double)pStatsAll[l][m].exactPrecision/pStatsAll[l][m].totalExact;
	   double eprec =
	     (double)pStatsAll[l][m].exactPrecision
	       /pStatsAll[l][m].totalPrecision;
	   double lrec2 =
	     (double)pStatsAll[l][m].otherWrong/pStatsAll[l][m].totalExact;
	   double lprec2 =
	     (double)pStatsAll[l][m].otherWrong //standing in for lprec2;
	       /pStatsAll[l][m].totalPrecision;
	   cout << "pStats = " << pStatsAll[l][m];
	   cout << "Acc: " << ac << "\t";
	   cout << "Rec: " << rec << "\t";
	   cout << "Prec: " << prec << "\t";
	   cout << "Lr: " << erec << "\t";
	   cout << "Lp: " << eprec << "\n";
	   cout << "\t\t" << "Lr2: " << lrec2 << "\t";
	   cout << "Lp2: " << lprec2 << "\n";
	   cout << "(p+r)/2 = " << (rec+prec)/2 << "\t";
	   cout << "(lp+lr)/2 = " << (erec+eprec)/2 << "\t";
	   cout << "(lp2+lr2)/2 = " << (lrec2+lprec2)/2 << "\n";
	   if(l != 2) continue;
	   /* for full lengths, print out cb info */
	   cout << "CBs: " << (double)cbStats[m][0]/(double)totSents << "\t ";
	   cout << "0 CBs: " << (double)cbStats[m][1]/(double)totSents << "\t ";
	   cout << "<=2CBs: " << (double)cbStats[m][2]/(double)totSents << "\n";
	 }
     }
   return 0;  
 }

