#include <time.h>
//#include <sys/resource.h>
#include <fstream>
#include <iostream>
#include <unistd.h>
#include <cmath>
#include "Field.h"
#include "GotIter.h"
#include "Wrd.h"
#include "InputTree.h"
#include "Bchart.h"
#include "ECArgs.h"
//#include "fixSent.h"
#include "MeChart.h"
#include "headFinder.h"
#include "Term.h"
#include "FuncInfo.h"
#include "Cluster.h"
#include "Pst.h"
#include "CLRInfo.h"
#include "Rule.h"
#include "FeatureFunctions.h"
#include "edgeSubFns.h"
#include "ArrayFeatureTree.h"

using namespace std;

extern SentRep* CurrentSentence;
extern int globalDebugLevel;

string usage();

/* If VERIFY_ANSWER is defined, it will evaluate the parse and 
   function tagging that it outputs; thus the gold parse is necessary
   even if MAP_INPUT is defined (for evaluation purposes). */
//#define VERIFY_ANSWER
/* Thus:
                U_I                    M_I                   F_I

   V_A=0  Input is gold std       Input is MAP in     Input is gold then
          in list form; will      list form           parse forest; will
	  be parsed                                   be mini-parsed

   V_A=1  Input is gold std       Input is gold+MAP,  Input is gold then
          in list form; will      both in list form;  parse forest; will
	  be parsed and verified  will be verified    be mini-parsed and
                                                      verified
*/

/* Should normally be false */
#define RESTORE_FUNC false

float endFactor = 1.2;
float midFactor = 1;

class Params 
{
public:
    Params(): 	    file(0),
		    fileString_(),
		    whichSent_( 0 ),
		    ofTotal_( 1 ),
		    field_( 0 ),
                    outputData_(false),
                    stdInput_(false),
		    numString_()
		    {}
    void	    init( ECArgs& args );
    const char *    file;
    const string&   fileString()
		    {  return fileString_;  }
    const string&   numString()
		    {   return numString_;  }
    const int	    whichSent()
		    {   return whichSent_;   }
    const int	    ofTotal()
		    {   return ofTotal_;   }
    const Field&    field() const
		    {   return *field_;   }
    bool&      stdInput() { return stdInput_; }
    bool&      outputData() { return outputData_; }
private:
    bool       stdInput_;
    bool       outputData_;
    string	    fileString_;
    string          numString_;
    int             whichSent_;
    int             ofTotal_;
    Field *         field_;
};

InputTree* inputTreeFromAnswerTree(AnswerTree* at, Item* itm, int pos=0);

void
Params::
init( ECArgs& args ) 
{  
  fileString_ = args.arg(0);
  if( args.isset('n') )
    {
      char etemp[16];
      strcpy(etemp,args.value('n').c_str());
      char *	temp = strchr( etemp, '/' );
      if( !temp )
	error( "No terminal '/' found in '-n' argument" );
      *temp = '\0';
      ofTotal_ = atoi( ++temp );
      char *	mask = new char[ ofTotal_ ];
      for( int i = 0; i < ofTotal_; i++ )
	mask[ i ] = 0;
      // fill in mask with valid numbers;
      string tmp2 = etemp;
      numString_ = tmp2;		// meaningful id for this process;
      whichSent_ = atoi(tmp2.c_str());
      mask[ whichSent_ ] = 1;
      field_ = new Field( ofTotal_, mask );
    }
    else
    {
	static char mask[1] = { 1 };
	field_ = new Field( 1, mask );
    }
}

// Global objects
//
//		(or object wannabes)
//

Params 		    params;

int
main(int argc, char *argv[])
{
   ECArgs args( argc, argv );
   /* o = basic, but not debugging, output.
      l = length of sentence to be proceeds 0-40 is default
      N = start with Nth sentence (numbered from 0)
      n = work on each #'th line.
      D = data directory (for terms.txt et al)
      d = print out debugging info at level #
      W = use wwclasses
      R = use rwclasses
      t = report timings (requires o)
      s = maximum sleep time
      f = f# says multiply ctl2 counts by #
      p = p# use prepFactor #
      P = pretty print
      F = print only function tags (no parse)
      q = 'quiet'---only output trees, to stdout */

   if (args.isset('h'))
   {
     cout << usage();
     exit(0);
   }

   if (! args.isset('q'))
     cerr << "Reading data files...." << endl;

   params.init( args );
   if(args.isset('s'))
     {
       int  maxDelay = atoi(args.value('s').c_str());
       srand(params.whichSent());
       int randN = rand();
       int delay = randN%maxDelay;
       sleep(delay);
     }
   if( args.isset('d') )
     {
       int lev = atoi(args.value('d').c_str());
       Bchart::printDebug() = lev;
       InputTree::printDebug = lev;
       globalDebugLevel = lev;
     }
	 
   int maxSentLen = 0;
   if(args.isset('l'))
     {
       maxSentLen = atoi(args.value('l').c_str());
     }
   int startSentence = 0;
   if (args.isset('N'))
   {
     startSentence = atoi(args.value('N').c_str());
   }
   int    totPopedEdges = 0;
   clock_t lastTime, currTime;
   double lastTimeSec, currTimeSec, elapsedTime;

   endFactor = 1.2;
   midFactor = (1.0 - (.3684 * endFactor))/(1.0 - .3684);

   if( args.nargs() > 2 || args.nargs() == 0 )	// require path name 
     error( "Need data dir and input file." );
   string datapath = args.arg( 0 );
   readHeadInfo( datapath );
   Term::init( datapath );
   FuncInfo::init( datapath );
   Cluster::init( datapath );

#ifdef VERIFY_ANSWER  
   InputTree::init( datapath );
#endif

   Pst pst( datapath );
   CLRInfo::init( datapath );
   Rule::init ( datapath );
   Feat::Usage = PARSE;
   addEdgeSubFeatureFns();
   addSubFeatureFunctions();
   Feature::init ( datapath );
   ArrayFeatureTree::init ( datapath );

   istream* instream;

   if(args.nargs() == 2)
   {
     string testSString(args.arg(1));
     instream = new ifstream(testSString.c_str());
     if( !*instream ) error( "Can't open testSstream " + testSString );
   }
   else
   {
     instream = &cin;
   }


#ifdef VERIFY_ANSWER
    string      pstatStreamName( params.fileString());
    pstatStreamName  += "PStatInfo/pStat";
    pstatStreamName += params.numString();
    if(maxSentLen > 40)
      pstatStreamName += "lg.txt";
    else pstatStreamName += ".txt";
    ofstream    pstatStream( pstatStreamName.c_str(), ios::out);
    if( !pstatStream )
      {
	cerr << "Looking to output to " << pstatStreamName << endl;
	error( "unable to open pstat stream" );
      }
#endif

   int numProbs = 1;
   int whichArray[10] = {7};

   int      sentenceCount = 0;  //counts all sentences so we can use 1/50;

   string  probSumString( datapath );
   probSumString += "pSgT.txt";
   ifstream    probSumStream( probSumString.c_str() );
   if( !probSumStream ) error( "Failed to find probSum file" );

   Bchart::readTermProbs(datapath);

   int totSents = 0;
   int totUnparsed = 0;
   FuncStats fSt;

   if (! args.isset ('q'))
     cerr << "Tagging sentences...." << endl;
   for( ; *instream ; )
     {
       InputTree     correct;  
       InputTree     MAP;  
       InputTree*    goldparse;
       InputTree*    mapparse;
       ParseStats pSt;
       string buffer;

       
       /* First, input the main parse tree.  This will be the gold-standard
	  parse in all cases except MAP_INPUT && !VERIFY_ANSWER, but in
	  such case we won't *need* a gold-standard parse, so I'm using
	  that variable anyway.  Bad style, but convenient. */
       *instream >> correct;

       goldparse = &correct;
       if (!args.isset('q'))
         cerr << "Sentence #" << sentenceCount << ":\n";
       if (maxSentLen != 0 && correct.length() > maxSentLen)
       {
	 if (!args.isset('q'))
	   cout << "Too long, skipped.\n";
	 continue;
       }

       if (sentenceCount < startSentence)
       {
	 ++sentenceCount;
	 continue;
       }

#ifdef VERIFY_ANSWER  
       /* Now we need to read in the MAP parse. */
       *instream >> MAP;
       mapparse = &MAP;
       assert (mapparse->length() == goldparse->length());
#else //!VERIFY_ANSWER
       /* Now we need to set the MAP parse to be the same as the
	  gold-standard parse, because of my laziness above. */
       mapparse = goldparse;
#endif //VERIFY_ANSWER

       if( !*instream ) break;
       int len = goldparse->length();
       if( !params.field().in(sentenceCount) )
       {
	 ++sentenceCount;
	 continue;
       }
       sentenceCount++;

       /* Thanks to my craftiness, there is now guaranteed to be some
	  parse in goldparse, which at least has the same words as the
	  correct parse.  'make' puts these words into the list passed
	  in as an argument. */
       list<string>  wtList;
       goldparse->make(wtList);

       int endRemoved = 0;
       int numRemoved = 0;

       SentRep orgsr( wtList );  // used in precision calc

       SentRep* sr = new SentRep( wtList );  
       CurrentSentence = sr;
       totSents++;

       // Strip existing function tags out of the tree
       mapparse->stripFuncTags();
       // for each node in tree, add appropriate function tags
       mapparse->postProcess(0);

#ifdef VERIFY_ANSWER
       mapparse->mapCompat( *goldparse, pSt, fSt, orgsr );
       goldparse->mapExact( *mapparse, pSt, fSt );
       mapparse->precision( *goldparse, pSt, fSt, orgsr, RESTORE_FUNC );
#endif // VERIFY_ANSWER


       //cout << "GOLD: \n" << *goldparse << endl;
       cout << mapparse->toString(args.isset('P'), !args.isset('F'))
	    << endl;

#ifdef VERIFY_ANSWER
       pstatStream << len << endl;
       pstatStream << pSt;
       cerr << pSt << endl;

       pstatStream << endl;
       pstatStream.flush();
#endif // VERIFY_ANSWER

       delete sr;
       sr = NULL;
     }

#ifdef VERIFY_ANSWER
   pstatStream << fSt << endl;
   pstatStream << "9999" << endl;
   pstatStream << totSents << "\t" << totUnparsed << endl;
#endif

   return 0;
}


InputTree*
inputTreeFromAnswerTree(AnswerTree* at, Item* itm, int position)
{
  //cerr << "itfat " << *itm << endl;
  int strt = itm->start();
  int fin = itm->finish();
  const Term* trm = itm->term();
  bool cmpl = itm->isComplement();
  string trmString = trm->name();
  string wrdString;
  string ntString;
  list<InputTree*> subtrs;
  InputTree* ans;
  if(trm->terminal_p())
  {
    wrdString = itm->word()->lexeme();
  }
  else
  {
    assert(at);
    Edge* e = at->e;
    list<AnswerTree*>::iterator ati = at->subtrees.begin();
    Item* subi;
    LeftRightGotIter gi(e);
    for(int i=0 ; ati != at->subtrees.end() ; ++i, ++ati)
    {
      gi.next(subi);
      //cerr << "SIT " << *subi->term() << endl;
      //stop terms will now appear in edges.
      if(subi->term() == Term::stopTerm)
      {
	ati--;
	continue;
      }
      AnswerTree* sab = *ati;
      InputTree* sit = inputTreeFromAnswerTree(sab,subi,i);
      subtrs.push_back(sit);
    }
  }
  //cerr << "C itfat " << *itm << endl;
  ans = new InputTree(strt, fin, wrdString, trmString, ntString,
		      subtrs, NULL, NULL, true, cmpl);
  if (trm->terminal_p())
    ans->setHeadTree(ans);
  else
  {
    InputTrees::iterator headSubTree = subtrs.begin();
    int headPos = headPosFromTree(ans);
    int i;
    for (i = 0; i < headPos; ++i)
      ++headSubTree;
    ans->setHeadTree((*headSubTree)->headTree());
    for (i = 0; i < 4; ++i)
      ans->setFuncTag(i, at->funcTag[i]);
  }
  for( InputTrees::iterator it = ans->subTrees().begin();
       it != ans->subTrees().end();
       ++it)
  {
    (*it)->setParent(ans);
  }
  //cerr << *ans << endl;
  return ans;
}

string usage()
{
  return
    "funcTest [options] datadir [datafile]\n"
    "\tdatadir is directory containing featureTree.ff, pSgt.txt, etc\n"
    "\tdatafile (or stdin) contains parsed input "
    	"in Penn Treebank format\n"
    "\n"
    "\t-h\tThis text\n"
    "\t-P\tPretty print (line wrap, indent on s-exps)\n"
    "\t-F\tPrint function tags only (no parse)\n"
    "\t-q\tQuiet: only output trees/tags, to stdout\n"
    "\n"
    "\t-d#\tPrint debugging info\n"
    "\t-l#\tMaximum sentence length (default: no maximum)\n"
    "\t-N#\tIgnore the first N sentences in input\n"
    "\t-nM/N\tTreat input as N-sentence blocks, proc Mth in ea block\n"
    "\t-s#\tAt start, sleep random time up to # seconds\n"
    ;
}
