//#include <time.h>
#include <sys/resource.h>
//#include <fstream.h>
//#include <iostream.h>
//#include <unistd.h>
//#include <math.h>
//#include "Field.h"
//#include "GotIter.h"
//#include "Wrd.h"
#include "InputTree.h"
//#include "Bchart.h"
#include "ECArgs.h"
//#include "MeChart.h"
#include "headFinder.h"
#include "Term.h"
#include "FuncInfo.h"
//#include "Cluster.h"
//#include "Pst.h"
#include "CLRInfo.h"

#define FOREACH(X,Y,Z) for (Y::iterator X = (Z).begin(); X != (Z).end(); ++it)
void handleCLR (InputTree* tree) 
{
  static int CLR = 0;
  if (!CLR) CLR = FuncInfo::get("CLR")->toInt();

  if (FuncInfo::getFuncInt(tree->ntInfo(), FuncInfo::MISC) == CLR)
  {
    //cout << "Adding pattern\n";
    CLRInfo::addCLRPattern(tree);
    /*
    //cout << "Parent's head: " << tree->parent()->head() << endl;
    //cout << "My head: " << tree->head() << endl;
    int pat = CLRInfo::getCLRPattern (tree);
    if (pat == 7 || pat == 2)
      cout << pat << " " 
        << tree->parent()->head() << " " << tree->getLeftSibling()->head() 
	<< " " << tree->head() << ":" << endl;
    else if (pat == 4)
      cout << pat << " " 
        << tree->parent()->head() << " " << tree->head() 
	<< " " << tree->getRightSibling()->head() << ":" << endl;
    else 
      cout << pat << " " 
        << tree->parent()->head() << " " << tree->head() << ":" << endl;
    cout << *tree->parent() << endl;
    */
  }
  //foreach child in tree->subTrees
  for (InputTreesIter it = tree->subTrees().begin();
       it != tree->subTrees().end();
       ++it)
  {
    InputTree* child = *it;
    handleCLR(child);
  }
}

map <string, map<string, int> > clrcount[2];
map <string, int> ntcount;
void countCLR (InputTree* tree)
{
  static int CLR = 0;
  if (!CLR) CLR = FuncInfo::get("CLR")->toInt();
  static int PP = 0;
  if (!PP) PP = Term::get("PP")->toInt();
  static int VP = 0;
  if (!VP) VP = Term::get("VP")->toInt();

  for (InputTreesIter it = tree->subTrees().begin();
       it != tree->subTrees().end();
       ++it)
  {
    InputTree* child = *it;
    countCLR(child);
  }

  //cerr << "-----------\nNew item " << tree->getLabel() << endl;
  InputTree* parent = tree->getParent();
  if (!parent) return;
  //cerr << "Has parent" << endl;
  if (tree->getLabel() != "PP") return;
  //cerr << "PP" << endl;
  if (parent->getLabel() != "VP") return;
  //cerr << "parent is VP" << endl;
  
  bool isCLR = 
    (FuncInfo::getFuncInt(tree->ntInfo(), FuncInfo::MISC) == CLR);

  ++clrcount[isCLR][parent->getHead()][tree->getHead()];
  ++ntcount[tree->ntInfo()];
}

void dumpNT(ostream& os)
{
    for (map<string, int>::const_iterator it = ntcount.begin();
	 it != ntcount.end();
	 ++it)
    {
      const string& ntinfo = (*it).first;
      const int& count = (*it).second;
	os << count << " " 
	   << ntinfo << endl;
    }

}

void dumpCLR(ostream& os)
{
  for (int isCLR = 0; isCLR <= 1; ++isCLR)
    for (map<string, map<string, int> >::const_iterator it = clrcount[isCLR].begin();
	 it != clrcount[isCLR].end();
	 ++it)
    {
      const string& verb = (*it).first;
      const map<string, int>& preps = (*it).second;
      for (map<string, int>::const_iterator it = preps.begin();
	   it != preps.end();
	   ++it)
      {
	const string& prep = (*it).first;
	int count = (*it).second;
	os << count << " " 
	   << verb << " "
	   << prep << " "
	   << isCLR << endl;
      }
    }

}

int
main(int argc, char *argv[])
{
   ECArgs args( argc, argv );
   /* o = basic, but not debugging, output.
      l = length of sentence to be proceeds 0-40 is default
      n = work on each #'th line.
      D = data directory (for terms.txt et al)
      d = print out debugging info at level #
      W = use wwclasses
      R = use rwclasses
      t = report timings (requires o)
      s = maximum sleep time
      f = f# says multiply ctl2 counts by #
      p = p# use prepFactor #
      P = which types of prob models to use */

   // prevent core file creation;
   struct rlimit 	core_limits;
   core_limits.rlim_cur = 0;
   core_limits.rlim_max = 0;
   setrlimit( RLIMIT_CORE, &core_limits );

   cerr << "Starting getPhV" << endl;
   //params.init( args );
	 
   //if( args.nargs() > 2 || args.nargs() == 0 )	// require path name 
   if( args.nargs() < 1)
     error( "Need data dir." );
   string datapath;
   string inputpath;
   if (args.isset('D')) {
     datapath = args.value('D');
     inputpath = args.arg( 0 );
   } else {
     inputpath = datapath = args.arg( 0 );
   }
   readHeadInfo( datapath );
   FuncInfo::init( datapath );
   Term::init( datapath );
   //InputTree::init( datapath );
   //Cluster::init( datapath );
   Pst pst( datapath );

   vector<string> files;
   if (args.nargs() == 1)
     files.push_back(string("text.txt"));
   else 
   {
     for (int i = 1; i < args.nargs(); ++i)
       files.push_back(inputpath + args.arg(i));
   }

   int      sentenceCount = 0;  //counts all sentences so we can use 1/50;

 for (int fnum = 0; fnum < files.size(); ++fnum)
 {
   ifstream testSStream(files[fnum].c_str());
   if( !testSStream ) error( "No testSstream" );

   for( ; testSStream ; )
     {
       sentenceCount++;

       //InputTree     correct;  
       InputTree*    goldparse;
       
       //testSStream >> correct;
       //goldparse = &correct;
       goldparse = new InputTree();
       testSStream >> *goldparse;
       countCLR(goldparse);
       delete goldparse;
       goldparse = NULL;

       //cout << sentenceCount << ": " << CLRInfo::getSize() << " entries\n";
     }
 }

 cerr << "Dumping counts...." << endl;
   //CLRInfo::dump(cout);
 //dumpCLR(cout);
 dumpNT(cout);

   return 0;
}
