#include <math.h>
#include <fstream>
#include <sys/resource.h>
#include <iostream>
#include <unistd.h>
#include "ECArgs.h"
#include <string>
#include "utils.h"
#include "FeatIter.h"
#include "Feature.h"
#include "Term.h"
#include "headFinder.h"
#include "fixSent.h"

using namespace std;

#define NUMPASSES 4
#define NUMNEWTPASSES 4

int whichInt;

FeatureTree* features;

string conditionedType;
int passN;
int hic = 0;

static bool
badPreReqs(MapFeatureTree* ft, MapFeatureTree* parft)
{
  int fInt = ft->getFeatureInt();
  //cerr << "ft to feat gives " << *ft << " " << fInt << endl;
  int cprf = Feature::fromInt(fInt, whichInt)->condPR;
  int pfInt = parft->getFeatureInt();
  //cerr << "ft to feat gives " << *parft << " " << pfInt << endl;
  int cprpf = Feature::fromInt(pfInt, whichInt)->condPR;
  if(cprpf < 0) return false;
  if(cprpf == cprf) return false;
  //cerr << "badPr for " << *ft << " " << *parft << " " << fInt << " "
    //<< pfInt << endl;
  return true;
}

Feat*
parentFeat(Feat* f)
{
  FeatureTree* ft = f->toTree();
  assert(ft);
  FeatureTree* parft = ft->getParent();
  if(!parft) return NULL; 
  if(parft->getInd() == ROOTIND) return NULL;
  while(parft->getInd() == AUXIND) parft = parft->getParent();
  //if(badPreReqs(ft,parft)) return NULL;
  return parft->getFeat(f->ind());
}

void
initFeatVals()
{
  FeatIter fi((MapFeatureTree*) features); 
  int m, i;
  Feat* f;
  for( ; fi.alive() ; fi.next() )
    {
      f = fi.curr;
      int fhij = f->cnt();
      assert(fhij > 0);
      int hij = f->toTree()->getCount();
      assert(hij > 0);
      assert(hij >= fhij);
      Feat* fj = parentFeat(f);
      int hj, fhj;
      if(fj)
	{
	  hj = fj->toTree()->getCount();
	  fhj = fj->cnt();
	}
      else
	{
	  fhj = 1;
	  hj = 1; // this sets val to fhij/hij;
	  //hj=FeatureTree::totCaboveMin[whichInt][Feature::assumedFeatVal]+1;
	}
      assert(hj > 0);
      assert(fhj > 0);
      assert(hj >= fhj);
      //float val = (float)(fhij * hj)/(float)(fhj * hij);
      float val = ((float)fhij/(float)hij);
      fi.curr->g() = val;
      //cerr << *(f->toTree()) << " " << f->ind()
	//   << " " << val << endl;
      if(!(val > 0))
	{
	  cerr << fhij << " " << hj << " " << fhj << " " << hij << endl;
	  assert(val > 0);
	}
    }
}

int
main(int argc, char *argv[])
{
   struct rlimit 	core_limits;
   core_limits.rlim_cur = 0;
   core_limits.rlim_max = 0;
   setrlimit( RLIMIT_CORE, &core_limits );

   ECArgs args( argc, argv );
   Feat::Usage = ISCALE;
   string  path( args.arg( 1 ) );
   Term::init(path);
   readHeadInfo(path);

   conditionedType = args.arg(0);
   cerr << "start iScale: " << conditionedType << endl;
   //Pst pst(path);

   Feature::init(path, conditionedType); 
   whichInt = Feature::whichInt;
   string fHp(path);
   fHp += conditionedType;
   fHp += ".f";
   ifstream fHps(fHp.c_str());
   assert(fHps);

   features = new MapFeatureTree(fHps);

   //features->subtree;
   
   initFeatVals();
   
   string gt(path);
   gt += conditionedType;
   gt += ".g";
   fstream gtstream(gt.c_str(), ios::out);
   assert(gtstream);
   gtstream.precision(3);

   FeatTreeIterator ftmi = features->childrenBegin();
   for( ; ftmi != features->childrenEnd() ; ++ftmi)
     {
       int afv = (*ftmi)->getInd();
       ((MapFeatureTree*) (*ftmi))->printFTree(afv, gtstream);
     }
   return 1;
}
