
#include <fstream>
#include <sys/resource.h>
#include <iostream>
#include <unistd.h>
#include <set>
#include "ECArgs.h"
#include "Feature.h"
#include "MapFeatureTree.h"
#include "InputTree.h"
#include "headFinder.h"
#include "FeatureFunctions.h"
#include "TreeHist.h"
#include "Term.h"
#include "FuncInfo.h"
#include "Cluster.h"
#include "Pst.h"
#include "CLRInfo.h"
#include <string>
#include "parseIntegration.h"
#include "SentRep.h"
#include "Rule.h"

using namespace std;

extern SentRep* CurrentSentence;

/* for a given history, as specified by a tree, for each feature f_i record
   how often it was used. */
bool prune = false;
InputTree* curS = NULL;

typedef set<string, less<string> > StringSet;
StringSet wordSet;
bool
inWordSet(int i)
{
  const string& w = Pst::fromInt(i);
  if(wordSet.find(w) == wordSet.end()) return false;
  return true;
}

extern InputTree* sentence[256];
extern int endPos;
void wordsFromTree(InputTree* tree);

void
makeSent(InputTree* tree)
{
  endPos = 0;
  for(int i = 0 ; i < 128 ; i++) sentence[i] = NULL;
  wordsFromTree(tree);  // sets the global setnence to the leaf nodes of tree
  assert(endPos == tree->finish());
}

void
wordsFromTree(InputTree* tree)
{
  if(tree->word() != "")
    {
      sentence[endPos++] = tree;
      return;
    }
  InputTreesIter subTreeIter = tree->subTrees().begin();
  for( ; subTreeIter != tree->subTrees().end() ; subTreeIter++ )
    {
      InputTree* subTree = *subTreeIter;
      wordsFromTree(subTree);
    }
}

int nfeatVs[20];

void
processG(int i, MapFeatureTree* ginfo[], TreeHist* treeh, int cVal)
{
  Feature* feat = Feature::fromInt(i, Feature::whichInt); 

  /* e.g., g(rtlu) starts from where g(rtl) left off (after tl)*/
  int searchStartInd = feat->startPos;
  MapFeatureTree* strt = ginfo[searchStartInd];
  if(prune && !strt) return;
  assert(strt);
  SubFeature* sf = SubFeature::fromInt(feat->subFeat, Feature::whichInt);
  //cerr << "pg " << sf->name << endl;
  int nfeatV = (*(sf->fun))(treeh);
  //nfeatVs[i] = nfeatV; 
  int ufi = sf->usf;
  //if(prune && ((ufi == 3 || ufi == 4 || ufi == 11)) && !inWordSet(nfeatV))
    //return;
  MapFeatureTree* histPt = strt->next(nfeatV, feat->auxCnt); 
  assert(histPt);
  ginfo[i] = histPt;
  histPt->incCount();
  //cerr << "pg " << i << " " << searchStartInd << " " << nfeatV
    // << " " << cVal << " " << treeh->pos <<  endl;
  //histPt->getFeat(cVal)->cnt()++;
  histPt->incFeat(cVal);
  /* ### Eventually we will put in a class of features that are keyed
     only when certain regular subfeatures are present, e.g., looking
     at the grandparent if we have seen */
  if(histPt->getSpecFeatures())
    error("Spec Features not yet implemented");
}

void
callProcG(TreeHist* treeh)
{
  int i;
  //for(i = 0 ; i < 20 ; i++) nfeatVs[i] = -1;
  MapFeatureTree* ginfo[MAXNUMFS];
  ginfo[0] = MapFeatureTree::root(); 
  int cVal = (*Feature::conditionedEvent)(treeh);
  if(cVal < 0) return;
  for(i = 1 ; i <= Feature::total[Feature::whichInt] ; i++)
    processG(i, ginfo, treeh, cVal);
}

void
gatherFfCounts(InputTree* tree, int inHpos)
{
  InputTrees& st = tree->subTrees();
  InputTrees::iterator  subTreeIter= st.begin();
  InputTree  *subTree;
  int hpos = 0;
  if(st.size() != 0) hpos = headPosFromTree(tree); // rationalist heuristic
  //cerr << hpos << *tree << endl;
  int pos = 0;
  for( ; subTreeIter != st.end() ; subTreeIter++ )
    {
      subTree = *subTreeIter;
      gatherFfCounts(subTree, pos==hpos ? 1 : 0);  // recurse!
      pos++;  // at end, pos = number of subtrees
    }
  //cerr << "g " << *tree << endl;
  //TreeHist treeh(tree, 0);  // just a struct
  TreeHist* treeh = tree->getTreeHist();
  treeh->setPosition(pos);
  treeh->setHeadPosition(hpos);
  //treeh.hpos = hpos; // this is the index of the subtree the head is in
  const Term* lhsTerm = Term::get(tree->term());
  if(Feature::whichInt == HCALC || Feature::whichInt == UCALC
#ifdef INTEGRATED_FUNCTAG
  	|| Feature::whichInt == GRCALC + INTEGRATED_FUNCTAG)
#else
	)
#endif
    {
      if(!inHpos) callProcG(treeh);
      return;
    }
  if(inHpos && (Feature::whichInt == GRHCALC || Feature::whichInt == FFHCALC
	|| Feature::whichInt == TPHCALC || Feature::whichInt == MFHCALC))
  {
    callProcG(treeh);
    return;
  }
  if(lhsTerm->terminal_p())
    {
      if(Feature::whichInt == TTCALC) callProcG(treeh);
      return;
    }
  if(st.size() == 1 && st.front()->term() == tree->term()) return;
  //cerr << "gff " << *tree << endl;
  subTreeIter = st.begin();
  int cVal;
  treeh->setPosition(-1);
  if(Feature::whichInt == LMCALC) callProcG(treeh);
  if(Feature::whichInt == LCALC) callProcG(treeh);
  if(Feature::whichInt == FCALC) callProcG(treeh);
  if(Feature::whichInt == GRCALC) callProcG(treeh);
  if(Feature::whichInt == FFCALC) callProcG(treeh);
  if(Feature::whichInt == TPCALC) callProcG(treeh);
  if(Feature::whichInt == MFCALC) callProcG(treeh);
  if(Feature::whichInt == HCCALC) callProcG(treeh);
  pos = 0;
  for( ; subTreeIter != st.end() ; subTreeIter++)
    {
      treeh->setPosition(pos);
      if(pos == hpos && Feature::whichInt == MCALC) callProcG(treeh);
      if(pos < hpos && Feature::whichInt == LCALC) callProcG(treeh);
      if(pos > hpos && Feature::whichInt == RCALC) callProcG(treeh);
      if(pos == hpos && Feature::whichInt == RUCALC) callProcG(treeh);
      if(pos >= hpos && Feature::whichInt == RMCALC) callProcG(treeh);
      if(pos <= hpos && Feature::whichInt == LMCALC) callProcG(treeh);
      pos++;
    }
  //cerr << "gg " << *tree << endl;
  treeh->setPosition(pos);
  if(Feature::whichInt == RCALC) callProcG(treeh);
  if(Feature::whichInt == RMCALC) callProcG(treeh);
}

int
main(int argc, char *argv[])
{
   struct rlimit 	core_limits;
   core_limits.rlim_cur = 0;
   core_limits.rlim_max = 0;
   setrlimit( RLIMIT_CORE, &core_limits );

   ECArgs args( argc, argv );
   assert(args.nargs() == 2);

   string  conditionedType( args.arg(0) );
   cerr << "start rCounts " <<  conditionedType << endl;
   string  path( args.arg( 1 ) );
   int minCount = 1;
   if(args.isset('m')) minCount = atoi(args.value('m').c_str());
   MapFeatureTree::setMinCount(minCount);
   if(args.isset('P')) prune = true;

   Term::init(path);
   FuncInfo::init(path);
   Cluster::init(path);
   readHeadInfo(path);
   addSubFeatureFunctions();
   Pst pst(path);
   CLRInfo::init(path);
   Rule::init(path);

   Feature::assignCalc(conditionedType);
       
   MapFeatureTree::root() = new MapFeatureTree();
   Feature::init(path, conditionedType);
   int ceFunInt = Feature::conditionedFeatureInt[Feature::whichInt];

   Feature::conditionedEvent
     = SubFeature::Funs[ceFunInt];
   string trainingString( path );
   trainingString += "train.txt";
   //trainingString += "test.pp";
   ifstream trainingStream(trainingString.c_str());
   if( !trainingStream ) error( "No training stream" );
   if(prune)
     {
       string wlString(path);
       wlString += "wordList.txt";
       ifstream wlStream(wlString.c_str());
       myassert(wlStream, wlString);
       while(true)
	 {
	   string nxtw;
	   wlStream >> nxtw;
	   if(!wlStream) break;
	   wordSet.insert(nxtw);
	 }
     }
       
   string resS(path);
   resS += conditionedType;
   resS += ".ff";
   ofstream res(resS.c_str());
   if (!res) error ("Could not open output file " + resS);

   int sentenceCount = 0;
   for( ; trainingStream ; sentenceCount++)
     {
       //if(sentenceCount >= 1) break;
       //if(sentenceCount >= 1000) break;
       if(sentenceCount%10000 == 1)
	 {
	   cerr << "rCounts "
	     << sentenceCount << endl;
	 }
       InputTree     correct;  
       trainingStream >> correct;
       if(correct.length() == 0) continue;
       EcSPairs wtList;
       correct.make(wtList);  // wtList is now list of word-tag pairs
       CurrentSentence = new SentRep(wtList);
       InputTree* par;
       par = &correct;

       makeSent(par);  // sentence is now array of InputTree leaf nodes
       curS = par;  // global
       gatherFfCounts(par, 0);
       if(Feature::whichInt == TTCALC)
	 {
	   list<InputTree*> dummy2;
	   InputTree stopInputTree(par->finish(),par->finish(),"","STOP","",
				   dummy2,NULL,NULL);
	   TreeHist* treeh = stopInputTree.getTreeHist();
	   treeh->setPosition(0);
	   treeh->setHeadPosition(0);
	   callProcG(treeh);
	 }

       delete CurrentSentence;
       CurrentSentence = NULL;
     }

   FeatTreeIterator fti = MapFeatureTree::root()->childrenBegin();
   for( ; fti != MapFeatureTree::root()->childrenEnd() ; ++fti)
     {
       int asVal = (*fti)->getInd();
       ((MapFeatureTree*) (*fti))->printFTree(asVal, res);
     }
   cout << "Total params for " << conditionedType << " = "
	<< MapFeatureTree::getTotParams() << endl;
   cout << "Done: " << (int)sbrk(0) << endl;
}
