#include <fstream>
#include <sys/resource.h>
#include <unistd.h>
#include <iostream>
#include "Feat.h"
#include "FeatIter.h"
#include "math.h"
#include "utils.h"
#include "ECArgs.h"
#include <map>
#include "MapFeatureTree.h"

using namespace std;

multimap<float, MapFeatureTree*, less<float> > featRank;

float percentDesiredFeatures = 0;
int totStates = 0;
int totSelectedStates = 0;
int totDesiredFeatures = 0;
int whichInt;

static float logLcomp(int fhi, int hi, int fhj, int hj);
static float logLikely(FeatureTree* f);
static void markFeat(MapFeatureTree*& f, float fr);
static Feat* parentFeat(Feat* f);
static FeatureTree* parentTree(FeatureTree* ft);
static void markFeats();

static float
logLcomp(int fhi, int hi, int fhj, int hj)
{
  if(fhi == 0)
    {
      return 0.0;
    }
  double lfhi = log((double)fhi);
  double lhi = log((double)hi);
  double lfhj = log((double)fhj);
  double lhj = log((double)hj);
  double ans = fhi * (lfhi - lhi - lfhj + lhj);
  //cerr << "logLcomp(" << fhi << ", " << hi << ", " << fhj
	// << ", " << hj << ") = "
	  //<< fhi << " * (" << lfhi << " - " << lhi << " - "
	 // << lfhj << " + " << lhj << ") = " << ans << endl;
  return (float)ans;
}

static float
logLikely(FeatureTree* ft)
{
  //cerr << "LL " << *ft << endl;
  int ftc = ft->getCount();
  assert(ftc > 0);
  FeatureTree* ftp = parentTree(ft);
  assert(ftp);
  int ftpc = ftp->getCount();
  assert(ftpc > 0);
  float ans = 0;
  /* we compute the sum for each feature f on ft of
     |f|(log(p(f|ft))-log(p(f|ftb)))  = logLcomp(fc, ftc, fpc, ftpc) */
  
  FeatIterator fmi = ft->featsBegin();
  for( ; fmi != ft->featsEnd() ; ++fmi)
    {
      Feat* f = *fmi;
      Feat* fp = parentFeat(f);
      assert(fp);
      int fc = f->cnt();
      int fpc = fp->cnt();
      float val = logLcomp(fc, ftc, fpc, ftpc);
      //cerr << "lcomp " << *f << " = " << val << endl;
      ans += val;
    }
  return ans;
}
/*
float
logLikely(FeatureTree* f)
{
  int fc = f->cnt();
  assert(fc > 0);
  int hij = f->toTree()->count;
  assert(hij > 0);
  assert(hij >= fhij);
  Feat* fj = parentFeat(f);
  assert(fj);
  int hj = fj->toTree()->count;
  assert(hj > 0);
  int fhj = fj->cnt();
  assert(fhj > 0);
  assert(hj >= fhj);
  int fhnij = fhj - fhij;
  if(!(fhnij >= 0))
    {
      cerr << fhnij << " not >= 0 for " << *f << " " << *fj << endl;
      cerr << *(f->toTree()) << endl;
      cerr << *(fj->toTree()) << endl;
      assert(fhnij >= 0);
    }
  int hnij = hj - hij;
  assert(hnij >= 0);
  if(!(hnij >= fhnij))
    {
      cerr << hnij << " " << fhnij << " for " << *f << endl;
      assert(hnij >= fhnij);
    }
  int nfhj = hj - fhj;
  int nfhij = hij - fhij;
  assert(nfhj >= 0);
  assert(nfhij >= 0);
  if(!(hij >= nfhij))
    {
      cerr << nfhij << " " << hij << " for " << *f << endl;
      assert(hij >= nfhij);
    }
  int nfhnij = hnij - fhnij;
  assert(nfhnij >= 0);
  assert(hnij >= nfhnij);
  float ans;
  ans = logLcomp(fhij, hij, fhj, hj);
  ans += logLcomp(fhnij, hnij, fhj, hj);
  ans += logLcomp(nfhij, hij, nfhj, hj);
  ans += logLcomp(nfhnij, hnij, nfhj, hj);
  if(ans < -0.001)
    {
      cerr << ans << "\t" << *(f->toTree()) 
	   << f->getInd() << endl;
      cerr << fhij << " " << hij << ",  " << fhj << " " << hj << ",  " << fhnij
	<< " " << hnij << ",  " << nfhij << "  " << nfhj << " " << nfhnij << endl;
      assert(ans >= -0.001);
    }
  return ans;
}
*/
static bool
badPreReqs(FeatureTree* f, FeatureTree* pf)
{
  MapFeatureTree* ft = (MapFeatureTree*) f;
  MapFeatureTree* parft = (MapFeatureTree*) pf;
  int fInt = ft->getFeatureInt();
  //cerr << "ft to feat gives " << *ft << " " << fInt << endl;
  int cprf = Feature::fromInt(fInt, whichInt)->condPR;
  int pfInt = parft->getFeatureInt();
  //cerr << "ft to feat gives " << *parft << " " << pfInt << endl;
  int cprpf = Feature::fromInt(pfInt, whichInt)->condPR;
  if(cprpf < 0) return false;
  if(cprpf == cprf) return false;
  //cerr << "badPr for " << *ft << " " << *parft << " " << fInt << " "
    //<< pfInt << endl;
  return true;
}

static Feat*
parentFeat(Feat* f)
{
  FeatureTree* ft = f->toTree();
  assert(ft);
  FeatureTree* parft = ft->getParent();
  if(!parft) return NULL; 
  if(parft->getInd() == ROOTIND) return NULL;
  while(parft->getInd() == AUXIND)
    {
      FeatureTree* parft2 = parft->getParent();
      assert(parft2);
      parft = parft2;
    }
  if(f->ind() == -1 && badPreReqs(ft,parft)) return NULL;
  return parft->getFeat(f->ind());
}

static FeatureTree*
parentTree(FeatureTree* ft)
{
  assert(ft);
  FeatureTree* parft = ft->getParent();
  if(!parft) return NULL; 
  if(parft->getInd() == ROOTIND) return parft;
  while(parft->getInd() == AUXIND)
    {
      FeatureTree* parft2 = parft->getParent();
      assert(parft2);
      parft = parft2;
    }
  return parft;
}

void
doRanking()
{
  MapFeatureTree* root = (MapFeatureTree*)FeatureTree::roots(whichInt);
  assert(root);
  FeatTreeIter fi(root);
  MapFeatureTree* f;
  for( ; fi.alive() ; fi.next() )
    {
      f = fi.curr;
      //cerr << "Looking at " << *f << endl;
      /* features at level 1 are automatically used */
      FeatureTree* parFt = parentTree(f);
      assert(parFt);
      assert(parFt->getInd() != AUXIND);
      if(parFt->getInd() == ROOTIND) markFeat(f, 1);
      else
	{
	  float ll = logLikely(f);
	  pair<const float, MapFeatureTree*> ipair(ll, f);
	  if(ll > 0) featRank.insert(ipair);
	}
      /*
	int cnt = f->cnt();
	if(cnt >= 20) markFeat(f, 1);
	else if(!parFt) markFeat(f, 1);
	else
	{
	  float ll = logLikely(f);
	  if(cnt >= 10 && ll > 0.1) markFeat(f,1);
	  else if(cnt >= 5 && ll > 2.0) markFeat(f,1);
	  else if(cnt >= 3 && ll > 8.0) markFeat(f, 1);
	}
	*/

      totStates++;
    }
}

static void
markFeats()
{
  multimap<float, MapFeatureTree*, less<float> >::reverse_iterator
    fRankIter = featRank.rbegin();
  int i = 0 ;
  bool justCut = false;
  for( ; fRankIter != featRank.rend() ; fRankIter++)
    {
      float fRank = (*fRankIter).first;
      MapFeatureTree* ft = (*fRankIter).second;
      if(i++%500 == 2) cerr << i << " " << fRank 
			    << " " << *ft << endl;
      if(totSelectedStates < totDesiredFeatures)
	{
	  markFeat(ft, fRank);
	  /*
	  if(f->cnt() <= 4)
	    {
	      cerr << fRank << "\t" << *(f->toTree()) 
		<< f->getInd() << endl;
	    }
	    */
	}
      else if(!justCut)
	{
	  justCut = true;
	  cerr << "Just cut off at " << i << " " << fRank << endl;
	}
    }
}

static void
markFeat(MapFeatureTree*& f, float fr)
{
  //assert(!(f->marked == 0)); 
  if(f->getMarked() > -1) return; 
  f->setMarked(fr);
  if(f->getInd() != AUXIND) totSelectedStates++;
  MapFeatureTree* pf = (MapFeatureTree*) f->getParent();
  if(!pf) return;
  markFeat(pf,fr);
}


int
main(int argc, char *argv[])
{
  struct rlimit 	core_limits;
  core_limits.rlim_cur = 0;
  core_limits.rlim_max = 0;
  setrlimit( RLIMIT_CORE, &core_limits );

  ECArgs args( argc, argv );
  assert(args.nargs() == 3);
  string conditionedType = args.arg(0);
  Feat::Usage = SEL;

  percentDesiredFeatures = (float)atoi(args.arg(1).c_str())/100.0;
  cerr << "start selFeats: " << conditionedType
       << " " << percentDesiredFeatures  << endl;
  string  path( args.arg( 2 ) );
  string fHp(path);
  fHp += conditionedType;
  fHp += ".ff";
  Feature::init(path, conditionedType);
  ifstream fHps(fHp.c_str());
  new MapFeatureTree(fHps);

  whichInt = Feature::whichInt;

  cerr << "Before doRanking" << endl;
  doRanking();
  totDesiredFeatures = (int) (percentDesiredFeatures * totStates);
  cerr << "Before markFeats" << endl;
  markFeats();

  string resS(path);
  resS += conditionedType;
  resS += ".f";
  ofstream res(resS.c_str());
  if (!res) error ("Can't open output file " + resS);

  FeatureTree* root = FeatureTree::roots(whichInt);
  FeatTreeIterator ftmIter = root->childrenBegin();
  cerr << "About to print featuretree" << endl;
  for( ; ftmIter != root->childrenEnd() ; ++ftmIter)
    {
      int asVal = (*ftmIter)->getInd();
      MapFeatureTree* subRoot = (MapFeatureTree*) root->getChild(asVal);
      subRoot->printFTree(asVal, res);
    }
  res << "\n\nSelected " << totSelectedStates << " of "
       << totStates << endl;
  return 1;
}
