#include <sys/resource.h>
#include "InputTree.h"
#include "ECArgs.h"
#include "headFinder.h"
#include "Term.h"
#include "FuncInfo.h"
#include "CLRInfo.h"

#define FOREACH(X,Y,Z) for (Y::iterator X = (Z).begin(); X != (Z).end(); ++it)

void testCLR (InputTree* tree)
{
  FOREACH (it, InputTrees, tree->getSubTrees())
  {
    testCLR(*it);
  }
  if (!tree->getParent()) return;
  if (CLRInfo::isCLRContext(tree->getTreeHist()))
    cout << "CLR Context: " << endl;
  else 
    cout << "Not a context: " << endl;
  cout << *tree << endl << endl;
}

int
main(int argc, char *argv[])
{
   ECArgs args( argc, argv );
   /* o = basic, but not debugging, output.
      l = length of sentence to be proceeds 0-40 is default
      n = work on each #'th line.
      D = data directory (for terms.txt et al)
      d = print out debugging info at level #
      W = use wwclasses
      R = use rwclasses
      t = report timings (requires o)
      s = maximum sleep time
      f = f# says multiply ctl2 counts by #
      p = p# use prepFactor #
      P = which types of prob models to use */

   // prevent core file creation;
   struct rlimit 	core_limits;
   core_limits.rlim_cur = 0;
   core_limits.rlim_max = 0;
   setrlimit( RLIMIT_CORE, &core_limits );

   cerr << "Starting testPhV" << endl;
	 
   if( args.nargs() < 1)
     error( "Need data dir." );
   string datapath;
   string inputpath;
   if (args.isset('D')) {
     datapath = args.value('D');
     inputpath = args.arg( 0 );
   } else {
     inputpath = datapath = args.arg( 0 );
   }
   readHeadInfo( datapath );
   FuncInfo::init( datapath );
   Term::init( datapath );
   //InputTree::init( datapath );
   //Cluster::init( datapath );
   Pst pst( datapath );
   CLRInfo::init( datapath );

   string testSString( inputpath );
   if(args.nargs() == 2)
     testSString += args.arg(1);
   else testSString += "test.txt"; 

   ifstream testSStream(testSString.c_str());
   if( !testSStream ) error( "No testSstream" );

   for (int i = 0; i < 5 && testSStream; ++i)
   {
     InputTree* goldparse;

     goldparse = new InputTree();
     testSStream >> *goldparse;
     testCLR (goldparse);
     delete goldparse;
     goldparse = NULL;
   }

   //CLRInfo::dump(cout);

   return 0;
}
