
#include <fstream>
#include <sys/resource.h>
#include <iostream>
#include <unistd.h>
#include <set>
#include "ECArgs.h"
#include "Feature.h"
#include "FeatureTree.h"
#include "InputTree.h"
#include "headFinder.h"
#include "FeatureFunctions.h"
#include "Pst.h"
#include "Smoother.h"
#include "TreeHist.h"
#include "Term.h"
#include "FuncInfo.h"
#include "Cluster.h"
#include "MapFeatureTree.h"
#include "CLRInfo.h"
#include "parseIntegration.h"
#include "SentRep.h"
#include "Rule.h"

using namespace std;

extern SentRep* CurrentSentence;

int pass;
int whichInt;
int sentenceCount;

/* for a given history, as specified by a tree, for each feature f_i record
   how often it was used. */

FeatureTree* tRoot = NULL;
string conditionedType;
InputTree* curTree = NULL;

struct TrData 
{
  int c;
  float pm;
};

TrData trData[MAXNUMFS][15];

float smoothedPs[MAXNUMFS];
bool prune = false;
typedef set<string, less<string> > StringSet;
StringSet wordSet;
bool
inWordSet(int i)
{
  const string& w = Pst::fromInt(i);
  if(wordSet.find(w) == wordSet.end()) return false;
  return true;
}

extern InputTree* sentence[256];
extern int endPos;
void wordsFromTree(InputTree* tree);

void
makeSent(InputTree* tree)
{
  endPos = 0;
  for(int i = 0 ; i < 128 ; i++) sentence[i] = NULL;
  wordsFromTree(tree);
  assert(endPos == tree->finish());
}

void
wordsFromTree(InputTree* tree)
{
  if(tree->word() != "")
    {
      sentence[endPos++] = tree;
      return;
    }
  InputTreesIter subTreeIter = tree->subTrees().begin();
  for( ; subTreeIter != tree->subTrees().end() ; subTreeIter++ )
    {
      InputTree* subTree = *subTreeIter;
      wordsFromTree(subTree);
    }
}

void
processG(int i, FeatureTree* ginfo[], TreeHist* treeh, int cVal)
{
  /* This is going to be re-set later.  I'm not sure why it's
   * nulled here.  Hygiene, I guess. */
  ginfo[i] = NULL;

  /* Use the static function fromInt to look up the ith member
     of the feature tree for distribution whichInt */
  Feature* feat = Feature::fromInt(i, whichInt); 

  /* e.g., g(rtlu) starts from where g(rtl) left off (after tl)*/
  int searchStartInd = feat->startPos;
  FeatureTree* strt = ginfo[searchStartInd];
  if(!strt)
    {
      if(i == 1) cerr << "no start for i = 1\n" << *curTree << endl;
      return;
    }

  /* We figure out which subfeature to use */
  SubFeature* sf = SubFeature::fromInt(feat->subFeat, whichInt);

  /* We now run one of the feature functions (from FeatureFunctions.C)
   * on the current node (treeh) to get the value of the feature, to
   * be stored in nfeatV. */
  int nfeatV = (*(sf->fun))(treeh);
  //cerr << nfeatV << endl;

  //cerr << "pg " << i << ", " << cVal << ", " << nfeatV << endl;
  /* Selects child feature to use, based on value of this feature. */
  FeatureTree* histPt = strt->follow(nfeatV, feat->auxCnt); 

  /* Store this feature tree in ginfo, so that in the next iteration
   * we can start where we left off. */
  ginfo[i] = histPt;
  if(i == 1)
    {
      smoothedPs[0] = 1;

      assert(histPt);
      Feat* feat1 = histPt->getFeat(cVal);
      float unsmoothedVal1 = feat1? feat1->g(): 0;

      smoothedPs[1] = unsmoothedVal1;
      if(smoothedPs[1] == 0)
	{
	  if(pass == 0)
	    cerr << "Zero at level 1 " << treeh->pos
		 << "\n" << *curTree << endl;
	  smoothedPs[1] = .0001;
	}
      //cerr <<i<< " " << nfeatV << " " << cVal << " "<< smoothedPs[1] << endl;
      for(int j = 2; j <= Feature::total[whichInt] ; j++)
	smoothedPs[j] = 0;
      return;
    }
  if(!histPt)
    {
      return;
    }
  float estm = histPt->getCount() * smoothedPs[1];
  //float estm = histPt->count * 0.1;
  int b = Smoother::bucket(estm);

  Feat* feat1 = histPt->getFeat(cVal);
  float unsmoothedVal = feat1? feat1->g(): 0;

  //if(unsmoothedVal == 0)
    //cerr << "Zero at level 2 " << treeh->pos << "\n" << *curTree << endl;
  float lam = Feature::getLambda(whichInt, i, b);
  //cerr << i << " " << nfeatV << " " << histPt->featureInt << " "
    // << estm << " " << b << " " << unsmoothedVal << endl;
  float uspathprob = lam*unsmoothedVal;
  float osmoothedVal = smoothedPs[searchStartInd];
  float smpathprob = (1-lam)*osmoothedVal;
  float nsmoothedVal = uspathprob+smpathprob;
  smoothedPs[i] = nsmoothedVal;
  float pm = uspathprob/nsmoothedVal;
  //cerr << "\t" << uspathprob << " " << smpathprob << " " << pm << endl;
  trData[i][b].c++;
  trData[i][b].pm += pm;  
}

void
callProcG(TreeHist* treeh)
{
  FeatureTree* ginfo[MAXNUMFS];
  ginfo[0] = MapFeatureTree::root(); 
  int cVal = (*Feature::conditionedEvent)(treeh);
  if(cVal < 0) return;
  for(int i = 1 ; i <= Feature::total[Feature::whichInt] ; i++)
    processG(i, ginfo, treeh, cVal);
}

void
gatherFfCounts(InputTree* tree, int inHpos)
{
  InputTrees& st = tree->subTrees();;
  InputTrees::iterator  subTreeIter= st.begin();
  InputTree  *subTree;
  int hpos = 0;
  if(st.size() != 0) hpos = headPosFromTree(tree);
  //cerr << hpos << *tree << endl;
  int pos = 0;
  for( ; subTreeIter != st.end() ; subTreeIter++ )
    {
      subTree = *subTreeIter;
      gatherFfCounts(subTree, pos==hpos ? 1 : 0);
      pos++;
    }
  //cerr << "g " << *tree << endl;
  curTree = tree;
  TreeHist* treeh = tree->getTreeHist();
  treeh->setPosition(pos);
  treeh->setHeadPosition(hpos);
  const Term* lhsTerm = Term::get(tree->term());
  if(lhsTerm->terminal_p())
    {
      if(Feature::whichInt == TTCALC) callProcG(treeh);
      return;
    }
  if(Feature::whichInt == HCALC || Feature::whichInt == UCALC
#ifdef INTEGRATED_FUNCTAG
  	|| Feature::whichInt == GRCALC + INTEGRATED_FUNCTAG)
#else
	)
#endif
    {
      if(!inHpos) callProcG(treeh);
      return;
    }
  if(inHpos && (Feature::whichInt == GRHCALC || Feature::whichInt == FFHCALC
	|| Feature::whichInt == TPHCALC || Feature::whichInt == MFHCALC))
  {
    callProcG(treeh);
    return;
  }
  if(st.size() == 1 && st.front()->term() == tree->term()) return;
  //cerr << "gff " << *tree << endl;
  subTreeIter = st.begin();
  int cVal;
  treeh->setPosition(-1);
  if(Feature::whichInt == LMCALC) callProcG(treeh);
  if(Feature::whichInt == LCALC) callProcG(treeh);
  pos = 0;
  for( ; subTreeIter != st.end() ; subTreeIter++)
    {
      treeh->setPosition(pos);
      if(pos == hpos && Feature::whichInt == MCALC) callProcG(treeh);
      if(pos < hpos && Feature::whichInt == LCALC) callProcG(treeh);
      if(pos > hpos && Feature::whichInt == RCALC) callProcG(treeh);
      if(pos == hpos && Feature::whichInt == RUCALC) callProcG(treeh);
      if(pos >= hpos && Feature::whichInt == RMCALC) callProcG(treeh);
      if(pos <= hpos && Feature::whichInt == LMCALC) callProcG(treeh);
      pos++;
    }
  //cerr << "gg " << *tree << endl;
  treeh->setPosition(pos);
  if(Feature::whichInt == RCALC) callProcG(treeh);
  if(Feature::whichInt == FCALC) callProcG(treeh);
  if(Feature::whichInt == GRCALC) callProcG(treeh);
  if(Feature::whichInt == FFCALC) callProcG(treeh);
  if(Feature::whichInt == TPCALC) callProcG(treeh);
  if(Feature::whichInt == MFCALC) callProcG(treeh);
  if(Feature::whichInt == HCCALC) callProcG(treeh);
  if(Feature::whichInt == RMCALC) callProcG(treeh);
}

void
printLambdas(ostream& res)
{
   int b,f;
   for(b = 1; b < 15 ; b++)
     {
       res << b ;
       for(f = 2; f <= Feature::total[whichInt] ; f++)
	 {
	   res << "\t";
	   float val = Feature::getLambda(whichInt, f, b);
	   if(val > .998) val = .998;  //nothing too close to 1;
	   res << val;
	 }
       res << "\n";
     }
}

void
updateLambdas()
{
   int b,f;
   for(b = 1; b < 15 ; b++)
     {
       for(f = 2; f <= Feature::total[whichInt] ; f++)
	 {
	   int denom = trData[f][b].c;
	   float num = trData[f][b].pm;
	   float ans = num/(float)denom;
	   if(denom == 0) ans = ((float)b/15.0);
	   else if(denom < 5 && trData[f][b+1].pm == 0 &&
		   (b == 0 ||  trData[f][b-1].pm == 0))
	     ans = 0;
	   Feature::setLambda(whichInt,f,b,ans);
	 }
     }
}

void
zeroData()
{
   int b,f;
   for(b = 0; b < 15 ; b++)
     {
       for(f = 1; f <= Feature::total[whichInt] ; f++)
	 {
	   trData[f][b].c = 0;
	   trData[f][b].pm = 0.0;
	 }
     }
}

void
lamInit()
{
   int b,f;
   float factor = 1.0/15.0;
   int numFs = Feature::total[whichInt];
   for(b = 0; b < 15 ; b++)
     {
       for(f = 1; f <= numFs ; f++)
	 {
	   Feature* feat = Feature::fromInt(f, whichInt); 
	   int strtpos = feat->startPos;
	   /* we are trying b/15 */
	   float num = factor* b;
	   Feature::setLambda(whichInt,f,b,num);
	   trData[f][b].c = 0;
	   trData[f][b].pm = 0.0;
	 }
     }
   printLambdas(cout);
}

int
main(int argc, char *argv[])
{
   struct rlimit 	core_limits;
   core_limits.rlim_cur = 0;
   core_limits.rlim_max = 0;
   setrlimit( RLIMIT_CORE, &core_limits );

   ECArgs args( argc, argv );
   assert(args.nargs() == 2);
   conditionedType = args.arg(0);
   cerr << "start trainRs: " << conditionedType << endl;

   string  path( args.arg( 1 ) );
   Term::init(path);
   FuncInfo::init(path);
   Cluster::init(path);
   readHeadInfo(path);

   Pst pst(path);
   CLRInfo::init(path);
   Rule::init(path);

   addSubFeatureFunctions();
   Feature::init(path, conditionedType); 

   whichInt = Feature::whichInt;
   int ceFunInt = Feature::conditionedFeatureInt[Feature::whichInt];
   Feature::conditionedEvent
     = SubFeature::Funs[ceFunInt];

   Feat::Usage = PARSE;
   string ftstr(path);
   ftstr += conditionedType;
   ftstr += ".g";
   ifstream fts(ftstr.c_str());
   assert(fts);
   tRoot = new MapFeatureTree(fts); //puts it in root;

   cout.precision(3);
   cerr.precision(3);

   lamInit();

   string trainingString( path );
   trainingString += "test.main";
   ifstream trainingStream(trainingString.c_str());
   if( !trainingStream ) error( "No training stream" );

   sentenceCount = 0;
   InputTree* trainingData[1001];
   for( ; sentenceCount <= 1000 && trainingStream ; sentenceCount++)
     {
       if(sentenceCount > 1000) break;
       if(sentenceCount%10000 == 1)
	 {
	   cerr << conditionedType << ".tr "
	     << sentenceCount << endl;
	 }
       InputTree*     correct = new InputTree;  
       trainingStream >> (*correct);
       if(correct->length() == 0) continue;
       EcSPairs wtList;
       correct->make(wtList); 
       CurrentSentence = new SentRep(wtList);
       InputTree* par;
       par = correct;
       trainingData[sentenceCount] = par;
     }

   for(pass = 0 ; pass < 10 ; pass++)
     {
       cout << "Pass " << pass << endl;
       for(sentenceCount = 0 ; sentenceCount < 1000 ; sentenceCount++)
	 {
	   InputTree* par = trainingData[sentenceCount];
	   if(sentenceCount%100 == 1)
	     cerr << sentenceCount << endl;
	   makeSent(par);
	   gatherFfCounts(par,0);
	   if(Feature::whichInt == TTCALC)
	     {
	       list<InputTree*> dummy2;
	       InputTree stopInputTree(par->finish(),par->finish(),"","STOP",
				       "", dummy2,NULL,NULL);
	       TreeHist* treeh = stopInputTree.getTreeHist();
	       treeh->setPosition(0);
	       treeh->setHeadPosition(0);
	       callProcG(treeh);
	     }
	 }
       updateLambdas();
       printLambdas(cout);
       zeroData();
     }
   string resS(path);
   resS += conditionedType;
   resS += ".lambdas";
   ofstream res(resS.c_str());
   if (!res) error ("Can't open output file " + resS);
   res.precision(3);
   printLambdas(res);
   cout << "Total params = " << MapFeatureTree::getTotParams() << endl;
   cout << "Done: " << (int)sbrk(0) << endl;
}

