/*
 * Copyright 1997, Brown University, Providence, RI.
 * 
 *                         All Rights Reserved
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose other than its incorporation into a
 * commercial product is hereby granted without fee, provided that the
 * above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Brown University not be used in
 * advertising or publicity pertaining to distribution of the software
 * without specific, written prior permission.
 * 
 * BROWN UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR ANY
 * PARTICULAR PURPOSE.  IN NO EVENT SHALL BROWN UNIVERSITY BE LIABLE FOR
 * ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
#include "utils.h"
#include <cmath>
#include <cstring>
#include <cassert>
#include <cctype>
#include "stemmer.h"

using namespace std;

void
ignoreComment(ifstream& inpt)
{
  string nxt;
  char a;
  inpt.get(a);
  if(a == '/')
    {
      char b = inpt.peek();
      if(b == '*')
	{
	  while(inpt)
	    {
	      inpt >> nxt;
	      if(nxt == "*/") break;
	    }
	  return;
	}
    }
  inpt.putback(a);
  return;
}
	  

double 
ran()
{
    return (rand() * 4.656612875245796E-10);
}

double
mylog2(double n)
{
  static double	nat_log_2 = log( 2.0 );
  return log( n )/ nat_log_2;
}


char*
toLower(const char* str, char* temp)
{
  int l = strlen(str);
  assert(l < 128);
  for(int i = 0 ; i <= l ; i++)
    {
      char n = str[i];
      int ni = (int)n;
      if(ni >= 65 && ni <= 90)
	{
	  temp[i] = (char)(ni+32);
	}
      else temp[i] = n;
    }
  return temp;
}

string
toLower(string old_str)
{
  string str(old_str);
  int l = str.size();
  for(int i = 0 ; i <= l ; i++)
    {
      str[i] = tolower(old_str[i]);
      /*
      char n = str[i];
      int ni = (int)n;
      if(ni >= 'A' && ni <= 'Z')
	{
	  str[i] = (char)(ni+('a' - 'A'));
	}
	*/
    }
  return str;
}

map<string, string> stemMapping;

string
lemmatise(const string& old_str)
{
  string str = toLower(old_str);
  //return str;
  //return stem(str);
  string retval = stemMapping[str];

  if (retval == "")
    retval = stemMapping[str] = stem(str);

  //cerr << retval.length() << ": " << retval.c_str() << endl;

  return retval;
}

void 
init_lemmatiser(string datadir)
{
  datadir += "irreg.txt";
  ifstream lemStream(datadir.c_str());

  string base;
  string form;
  while (lemStream)
  {
    lemStream >> form;
    if (!lemStream) break;
    myassert (form == "(", form);
    lemStream >> base;
    base = stem(base);
    while (lemStream)
    {
      lemStream >> form;
      if (form == ")") break;
      stemMapping[form] = base;
    }
  }
}
